/*-----------------------------------------------------------------------------

	ST-Sound ( YM files player library )

	Copyright (C) 1995-1999 Arnaud Carre ( http://leonard.oxg.free.fr )

	Extended YM-2149 Emulator, with ATARI music demos effects.
	(SID-Like, Digidrum, Sync Buzzer, Sinus SID and Pattern SID)

-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------

	This file is part of ST-Sound

	ST-Sound is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	ST-Sound is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ST-Sound; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-----------------------------------------------------------------------------*/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "Ym2149Ex.h"

extern	int	bSoundFilter;

//-------------------------------------------------------------------
// env shapes.
//-------------------------------------------------------------------
#ifdef __ARM_ADS
static const int EnvWave[16][8] =
{
{ 1,0,0,0,0,0,0,0 },
{ 0,1,0,0,0,0,0,0 },
{ 1,0,1,0,1,0,1,0 },
{ 1,0,0,0,0,0,0,0 },
{ 1,0,0,1,1,0,0,1 },
{ 1,0,1,1,1,1,1,1 },
{ 0,1,0,1,0,1,0,1 },
{ 0,1,1,1,1,1,1,1 },
{ 0,1,1,0,0,1,1,0 },
{ 0,1,0,0,0,0,0,0 }
};
#else
static const int Env00xx[8]={ 1,0,0,0,0,0,0,0 };
static const int Env01xx[8]={ 0,1,0,0,0,0,0,0 };
static const int Env1000[8]={ 1,0,1,0,1,0,1,0 };
static const int Env1001[8]={ 1,0,0,0,0,0,0,0 };
static const int Env1010[8]={ 1,0,0,1,1,0,0,1 };
static const int Env1011[8]={ 1,0,1,1,1,1,1,1 };
static const int Env1100[8]={ 0,1,0,1,0,1,0,1 };
static const int Env1101[8]={ 0,1,1,1,1,1,1,1 };
static const int Env1110[8]={ 0,1,1,0,0,1,1,0 };
static const int Env1111[8]={ 0,1,0,0,0,0,0,0 };
static const int *EnvWave[16] = {Env00xx,Env00xx,Env00xx,Env00xx,
							Env01xx,Env01xx,Env01xx,Env01xx,
							Env1000,Env1001,Env1010,Env1011,
							Env1100,Env1101,Env1110,Env1111};
#endif

#ifdef __ARM_ADS
#define T(a) (a*2)/6
static const 
#else
#define T(a) a
#endif

unsigned short ymVolumeTable[16] =
{	T(62),T(161),T(265),T(377),T(580),T(774),T(1155),T(1575),T(2260),T(3088),T(4570),T(6233),T(9330),T(13187),T(21220),T(32767)};


//----------------------------------------------------------------------
// Very cool and fast DC Adjuster ! This is the *new* stuff of that
// package coz I get that idea working on SainT in 2004 !
// ( almost everything here is from 1995 !)
//----------------------------------------------------------------------
void CDcAdjusterInit(CDcAdjuster* a)
{
	Reset(a);
}

int	GetDcLevel(CDcAdjuster* a)
{
	return (a->m_sum>>DC_ADJUST_BUFFERLEN_SHIFT);
}

void Reset(CDcAdjuster*a)
{
	int i;
	for (i=0;i<DC_ADJUST_BUFFERLEN;i++)
		a->m_buffer[i] = 0;

	a->m_pos = 0;
	a->m_sum = 0;
}

void AddSample(CDcAdjuster*a,int sample)
{
	a->m_sum -= a->m_buffer[a->m_pos];
	a->m_sum += sample;

	a->m_buffer[a->m_pos] = sample;
	a->m_pos = (a->m_pos+1)&(DC_ADJUST_BUFFERLEN-1);
}

//----------------------------------------------------------------------
// Very simple low pass filter.
// Filter coefs are 0.25,0.5,0.25
//----------------------------------------------------------------------
static	signed short	*pInternalInput = NULL;
static	long			internalInputLen = 0;
static	signed short	oldFilter[2] = {0,0};

signed short *getBufferCopy(signed short *pIn,int len)
{
		if (len>internalInputLen)
		{
			if (pInternalInput) free(pInternalInput);
			pInternalInput = (signed short *)malloc(len*sizeof(short));
			internalInputLen = len;
		}
		//memcpy(pInternalInput,pIn,len*sizeof(short));
		return pInternalInput;
}

#define	DSP_FILTER(a,b,c)	((((int)a)+((int)b)+((int)b)+((int)c))>>3)

// filter: out -> out
void	lowpFilterProcess(signed short *pOut,int len)		// private
{
signed short *pIn;
int i;
signed short s;

		pIn = pInternalInput;//getBufferCopy(pOut,len);
		if (len>0)
		{
			s = DSP_FILTER(oldFilter[0],oldFilter[1],pIn[0]);
			*pOut++ = s;
			*pOut++ = s;
		}
		if (len>1)
		{
			s = DSP_FILTER(oldFilter[1],pIn[0],pIn[1]);
			*pOut++ = s;
			*pOut++ = s;
		}
		oldFilter[1] = pIn[len-1];
		oldFilter[0] = pIn[len-2];
		for (i=2;i<len;i++)
		{
			s = DSP_FILTER(pIn[0],pIn[1],pIn[2]);
			*pOut++ = s;
			*pOut++ = s;
			pIn++;
		}
}

static	unsigned char *ym2149EnvInit(unsigned char *pEnv,int a,int b)
{
	int i;
	int d;

		d = b-a;
		a *= 15;
		for (i=0;i<16;i++)
		{
			*pEnv++ = (unsigned char)a;
			a += d;
		}
		return pEnv;
}

void CYm2149ExInit(CYm2149Ex* ym,unsigned long masterClock,int prediv,unsigned long playRate)
{
	int env;
	unsigned char *pEnv;

		ym->frameCycle = 0;
#ifndef __ARM_ADS
		if (ymVolumeTable[15]==32767)		// excuse me for that bad trick ;-)
		{
			int i;
			for (i=0;i<16;i++)
			{
				ymVolumeTable[i] = (ymVolumeTable[i]*2)/6;
			}
		}
#endif

	//--------------------------------------------------------
	// build env shapes.
	//--------------------------------------------------------
		pEnv = &ym->envData[0][0][0];
		for (env=0;env<16;env++)
		{
			int phase;
#ifdef __ARM_ADS
			for (phase=0;phase<4;phase++)
				pEnv = ym2149EnvInit(pEnv,EnvWave[env][phase*2+0],EnvWave[env][phase*2+1]);
#else
			const int *pse = EnvWave[env];
			for (phase=0;phase<4;phase++)
			{
				pEnv = ym2149EnvInit(pEnv,pse[phase*2+0],pse[phase*2+1]);
			}
#endif
		}

		ym->internalClock = masterClock/prediv;		// YM at 2Mhz on ATARI ST
		ym->replayFrequency = playRate;				// DAC at 44.1Khz on PC
		ym->cycleSample = 0;

		ym->freqStep = ym->internalClock;
		ym->freqStep /= ym->replayFrequency;
		ym->freqStep *= 4096;
		ym->freqSid = ((double)(1<<16))/ym->replayFrequency;

	// Set volume voice pointers.
		ym->pVolA = &ym->volA;
		ym->pVolB = &ym->volB;
		ym->pVolC = &ym->volC;

	// Reset YM2149
		CYm2149ExReset(ym);
}

void CYm2149ExDestroy(CYm2149Ex* ym)
{
}

void	setClock(CYm2149Ex* ym,unsigned long _clock)
{
		ym->internalClock = _clock;
}


unsigned long toneStepCompute(CYm2149Ex* ym,int rHigh,int rLow)
{
double step;
unsigned long istep;
int per;

	per = rHigh&15;
	per = (per<<8)+rLow;
	if (per<=5) 
	{
		return 0;
	}
/*
	step = ym->internalClock;
	step /= ((double)per*(double)ym->replayFrequency);
	step *= 4096.0 * 65536.0;
*/
	step = ym->freqStep/per;
	step *= 65536;
	istep = (unsigned long)step;
	return istep;
}

unsigned long noiseStepCompute(CYm2149Ex* ym,int rNoise)
{
double step;
unsigned long istep;
int per;

	per = (rNoise&0x1f);
	if (per<3)
		return 0;
/*
	step = ym->internalClock;
	step /= ((double)per*(double)ym->replayFrequency);
	step *= 4096.0;
*/
	step = ym->freqStep/per;
	istep = (unsigned long)step;
	return istep;
}

unsigned long envStepCompute(CYm2149Ex* ym,int rHigh,int rLow)
{
double step;
unsigned long istep;
int per;

	per = rHigh;
	per = (per<<8)+rLow;
/*
	step = ym->internalClock;
	step /= ((double)per*(double)ym->replayFrequency);
	step *= 128.0*65536.0;
*/
	step = ym->freqStep/per;
	step *= 2048;
	istep = (unsigned long)step;
	return istep;
}

unsigned	int	rndCompute(CYm2149Ex* ym)
{
		int	rBit = (ym->rndRack&1) ^ ((ym->rndRack>>2)&1);
		ym->rndRack = (ym->rndRack>>1) | (rBit<<16);
		return (rBit ? 0 : 0xffff);
}

void	CYm2149ExReset(CYm2149Ex* ym)
{
		writeRegister(ym,7,0x3f);
		writeRegister(ym,8,0);
		writeRegister(ym,9,0);
		writeRegister(ym,10,0);
		ym->currentNoise = 0xffff;
		ym->rndRack = 1;
		sidStop(ym,0);
		sidStop(ym,1);
		sidStop(ym,2);

		ym->envShape = 0;
		ym->envPhase = 0;
		ym->envPos = 0;

		Reset(&ym->m_dcAdjust);


		memset(ym->specialEffect,0,sizeof(ym->specialEffect));

		ym->bSyncBuzzer = YM_FALSE;

}


void	sidVolumeCompute(CYm2149Ex* ym,int voice,int *pVol)
{

		YmSpecialEffect	*pVoice = &ym->specialEffect[voice];

		if (pVoice->bSid)
		{
			if (pVoice->sidPos & (1<<31))
				writeRegister(ym,8+voice,pVoice->sidVol);
			else
				writeRegister(ym,8+voice,0);
		}
		else if (pVoice->bDrum)
		{
//			writeRegister(8+voice,pVoice->drumData[pVoice->drumPos>>DRUM_PREC]>>4);

			*pVol = (pVoice->drumData[pVoice->drumPos>>DRUM_PREC] * 255) / 6;
			switch (voice)
			{
				case 0:
					ym->pVolA = &ym->volA;
					ym->mixerTA = 0xffff;
					ym->mixerNA = 0xffff;
					break;
				case 1:
					ym->pVolB = &ym->volB;
					ym->mixerTB = 0xffff;
					ym->mixerNB = 0xffff;
					break;
				case 2:
					ym->pVolC = &ym->volC;
					ym->mixerTC = 0xffff;
					ym->mixerNC = 0xffff;
					break;
			}

			pVoice->drumPos += pVoice->drumStep;
			if ((pVoice->drumPos>>DRUM_PREC) >= pVoice->drumSize)
			{
				pVoice->bDrum = 0;
			}

		}
}

signed short nextSample(CYm2149Ex* ym)
{
int vol;
int bt,bn;

		if (ym->noisePos&0xffff0000)
		{
			ym->currentNoise ^= rndCompute(ym);
			ym->noisePos &= 0xffff;
		}
		bn = ym->currentNoise;

		ym->volE = ymVolumeTable[ym->envData[ym->envShape][ym->envPhase][ym->envPos>>(32-5)]];

		sidVolumeCompute(ym,0,&ym->volA);
		sidVolumeCompute(ym,1,&ym->volB);
		sidVolumeCompute(ym,2,&ym->volC);

	//---------------------------------------------------
	// Tone+noise+env+DAC for three voices !
	//---------------------------------------------------
		bt = ((((signed long)ym->posA)>>31) | ym->mixerTA) & (bn | ym->mixerNA);
		vol  = (*ym->pVolA)&bt;
		bt = ((((signed long)ym->posB)>>31) | ym->mixerTB) & (bn | ym->mixerNB);
		vol += (*ym->pVolB)&bt;
		bt = ((((signed long)ym->posC)>>31) | ym->mixerTC) & (bn | ym->mixerNC);
		vol += (*ym->pVolC)&bt;

	//---------------------------------------------------
	// Inc
	//---------------------------------------------------
		ym->posA += ym->stepA;
		ym->posB += ym->stepB;
		ym->posC += ym->stepC;
		ym->noisePos += ym->noiseStep;
		ym->envPos += ym->envStep;
		if (ym->envPhase==0)
		{
			if (ym->envPos<ym->envStep)
			{
				ym->envPhase = 1;
			}
		}

		ym->specialEffect[0].sidPos += ym->specialEffect[0].sidStep;
		ym->specialEffect[1].sidPos += ym->specialEffect[1].sidStep;
		ym->specialEffect[2].sidPos += ym->specialEffect[2].sidStep;

	//---------------------------------------------------
	// Normalize process
	//---------------------------------------------------
		AddSample(&ym->m_dcAdjust,vol);
		return (vol - GetDcLevel(&ym->m_dcAdjust));
}


int		readRegister(CYm2149Ex* ym,int reg)
{
		if ((reg>=0) && (reg<=13)) return ym->registers[reg];
		else return -1;
}

void	writeRegister(CYm2149Ex* ym,int reg,int data)
{

		switch (reg)
		{
		case 0:
			ym->registers[0] = data&255;
			ym->stepA = toneStepCompute(ym,ym->registers[1],ym->registers[0]);
			if (!ym->stepA) ym->posA = (1<<31);		// Assume output always 1 if 0 period (for Digi-sample !)
			break;

		case 2:
			ym->registers[2] = data&255;
			ym->stepB = toneStepCompute(ym,ym->registers[3],ym->registers[2]);
			if (!ym->stepB) ym->posB = (1<<31);		// Assume output always 1 if 0 period (for Digi-sample !)
			break;

		case 4:
			ym->registers[4] = data&255;
			ym->stepC = toneStepCompute(ym,ym->registers[5],ym->registers[4]);
			if (!ym->stepC) ym->posC = (1<<31);		// Assume output always 1 if 0 period (for Digi-sample !)
			break;

		case 1:
			ym->registers[1] = data&15;
			ym->stepA = toneStepCompute(ym,ym->registers[1],ym->registers[0]);
			if (!ym->stepA) ym->posA = (1<<31);		// Assume output always 1 if 0 period (for Digi-sample !)
			break;

		case 3:
			ym->registers[3] = data&15;
			ym->stepB = toneStepCompute(ym,ym->registers[3],ym->registers[2]);
			if (!ym->stepB) ym->posB = (1<<31);		// Assume output always 1 if 0 period (for Digi-sample !)
			break;

		case 5:
			ym->registers[5] = data&15;
			ym->stepC = toneStepCompute(ym,ym->registers[5],ym->registers[4]);
			if (!ym->stepC) ym->posC = (1<<31);		// Assume output always 1 if 0 period (for Digi-sample !)
			break;

		case 6:
			ym->registers[6] = data&0x1f;
			ym->noiseStep = noiseStepCompute(ym,ym->registers[6]);
			if (!ym->noiseStep)
			{
				ym->noisePos = 0;
				ym->currentNoise = 0xffff;
			}
			break;

		case 7:
			ym->registers[7] = data&255;
			ym->mixerTA = (data&(1<<0)) ? 0xffff : 0;
			ym->mixerTB = (data&(1<<1)) ? 0xffff : 0;
			ym->mixerTC = (data&(1<<2)) ? 0xffff : 0;
			ym->mixerNA = (data&(1<<3)) ? 0xffff : 0;
			ym->mixerNB = (data&(1<<4)) ? 0xffff : 0;
			ym->mixerNC = (data&(1<<5)) ? 0xffff : 0;
			break;

		case 8:
			ym->registers[8] = data&31;
			ym->volA = ymVolumeTable[data&15];
			if (data&0x10)
				ym->pVolA = &ym->volE;
			else
				ym->pVolA = &ym->volA;
			break;
		
		case 9:
			ym->registers[9] = data&31;
			ym->volB = ymVolumeTable[data&15];
			if (data&0x10)
				ym->pVolB = &ym->volE;
			else
				ym->pVolB = &ym->volB;
			break;
		
		case 10:
			ym->registers[10] = data&31;
			ym->volC = ymVolumeTable[data&15];
			if (data&0x10)
				ym->pVolC = &ym->volE;
			else
				ym->pVolC = &ym->volC;
			break;

		case 11:
			ym->registers[11] = data&255;
			ym->envStep = envStepCompute(ym,ym->registers[12],ym->registers[11]);
			break;

		case 12:
			ym->registers[12] = data&255;
			ym->envStep = envStepCompute(ym,ym->registers[12],ym->registers[11]);
			break;

		case 13:
			ym->registers[13] = data&0xf;
			ym->envPos = 0;
			ym->envPhase = 0;
			ym->envShape = data&0xf;
			break;

		}
}

signed short *getBufferCopy(signed short *pIn,int len);

//#define LOW_PASS_FILTER
#define LOW_PASS_FILTER2

void	CYm2149ExUpdate(CYm2149Ex* ym,signed short *pSampleBuffer,int nbSample)
{
#ifdef LOW_PASS_FILTER
		signed short *pBuffer = getBufferCopy(pSampleBuffer,nbSample);//pSampleBuffer;
		int	nbs = nbSample;

		for (;nbSample>0;--nbSample)
			*pBuffer++ = nextSample(ym);

		lowpFilterProcess((signed short*)pSampleBuffer,nbs);
#else
#ifdef LOW_PASS_FILTER2
		static int last0 = 0, last1 = 0;
		int last2;
		signed short s;
		for (;nbSample>0;--nbSample)
		{
			last2 = nextSample(ym);
			s = (signed short)((last0+(last1<<2)+last2)>>3);
			*pSampleBuffer++ = s;
			*pSampleBuffer++ = s;
			last0 = last1;
			last1 = last2;
		}
#else
		signed short s;
		for (;nbSample>0;--nbSample)
		{
			*pSampleBuffer++ = s = nextSample(ym);
			*pSampleBuffer++ = s;
		}
#endif
#endif
}

void	drumStart(CYm2149Ex* ym,int voice,unsigned char *pDrumBuffer,unsigned long drumSize,int drumFreq)
{
	ym->specialEffect[voice].drumData = pDrumBuffer;
	ym->specialEffect[voice].drumPos = 0;
	ym->specialEffect[voice].drumSize = drumSize;
	ym->specialEffect[voice].drumStep = (drumFreq<<DRUM_PREC)/ym->replayFrequency;
	ym->specialEffect[voice].bDrum = 1;
}

void	drumStop(CYm2149Ex* ym,int voice)
{
		ym->specialEffect[voice].bDrum = 0;
}

void	sidStart(CYm2149Ex* ym,int voice,int timerFreq,int vol)
{
		//double tmp = (double)timerFreq*((double)(1<<31))/(double)ym->replayFrequency;
		double tmp = ym->freqSid*timerFreq;
		tmp*=32768;
		ym->specialEffect[voice].sidStep = (unsigned long)tmp;
		ym->specialEffect[voice].sidVol = vol&15;
		ym->specialEffect[voice].bSid = 1;
}

void	sidSinStart(CYm2149Ex* ym,int voice,int timerFreq,int vol)
{
/*
		double tmp = (double)timerFreq*((double)(1<<(32-3)))/(double)replayFrequency;
		ym->ymVoice[voice].sidStep = (unsigned long)tmp;
		ym->ymVoice[voice].sidVol = vol&15;
		ym->ymVoice[voice].bSid = YM_FALSE;
		ym->ymVoice[voice].bSidSin = YM_TRUE;
*/
}

void	sidStop(CYm2149Ex* ym,int voice)
{
		ym->specialEffect[voice].bSid = 0;
}

void	syncBuzzerStart(CYm2149Ex* ym,int timerFreq,int _envShape)
{
/*
		double tmp = (double)timerFreq*((double)(1<<31))/(double)ym->replayFrequency;
		ym->syncBuzzerShape = _envShape&15;
		ym->syncBuzzerStep = (unsigned long)tmp;
		ym->bSyncBuzzer = 1;
*/
}

void	syncBuzzerStop(CYm2149Ex* ym)
{
		ym->bSyncBuzzer = 0;
}
