/*-----------------------------------------------------------------------------

	ST-Sound ( YM files player library )

	Copyright (C) 1995-1999 Arnaud Carre ( http://leonard.oxg.free.fr )

	Manage YM file depacking and parsing

-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------

	This file is part of ST-Sound

	ST-Sound is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	ST-Sound is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ST-Sound; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-----------------------------------------------------------------------------*/

#include "fileio.h"
#include <stdlib.h>
#include <string.h>
#include "YmMusic.h"
#include "Depacker.h"

static const unsigned short ymVolumeTable0[16] =
 {	62,161,265,377,580,774,1155,1575,2260,3088,4570,6233,9330,13187,21220,32767};


static	void	signeSample(unsigned char *ptr,long size)
{

		if (size>0)
		{
			do
			{
				*ptr++ ^= 0x80;
			}
			while (--size);
		}
}

char	*mstrdup(char *in)
{
		char *out = (char*)malloc(strlen(in)+1);
		if (out) strcpy(out,in);
		return out;
}

UD      readMotorolaDword(UB **ptr)
{
UD n;
UB *p = *ptr;

        n = (p[0]<<24)|(p[1]<<16)|(p[2]<<8)|p[3];
        p+=4;
        *ptr = p;
        return n;
}

UW      readMotorolaWord(UB **ptr)
{
UW n;
UB *p = *ptr;

        n = (p[0]<<8)|p[1];
        p+=2;
        *ptr = p;
        return n;
}

char    *readNtString(char **ptr)
{
char *p;

		p = mstrdup(*ptr);
		(*ptr) += strlen(*ptr)+1;
        return p;
}

UD      readDword(UB **ptr)
{
UD n;
UB *p = *ptr;

        n = (p[3]<<24)|(p[2]<<16)|(p[1]<<8)|p[0];
        p+=4;
        *ptr = p;
        return n;
}

unsigned char	*depackFile(CYmMusic* ym)
 {
 lzhHeader_t pHeader;
 UB	*pNew;
 UB	*pSrc;
 UB	*p;
 int i;

		p = ym->pBigMalloc;
		pHeader.size = *p++;
		pHeader.sum = *p++;
		for (i=0;i<5;i++)
			pHeader.id[i] = *p++;
		pHeader.packed = readDword(&p);
		pHeader.original = readDword(&p);
		for (i=0;i<5;i++)
			pHeader.reserved[i] = *p++;
		pHeader.level = *p++;;
		pHeader.name_length = *p++;

		if ((pHeader.size==0) ||
			(strncmp(pHeader.id,"-lh5-",5)))
		{ // Le fichier n'est pas compresse, on retourne l'original.
			return ym->pBigMalloc;
		}

		ym->fileSize = (UD)-1;

		if (pHeader.level != 0)
		{ // Compression LH5, header !=0 : Error.
			free(ym->pBigMalloc);
			ym->pBigMalloc = NULL;
			setLastError(ym,"LHARC Header must be 0 !");
			return NULL;
		}

		ym->fileSize = pHeader.original;
		pNew = (UB*)malloc(pHeader.original);
		if (!pNew)
		{
			setLastError(ym,"MALLOC Failed !");
			free(ym->pBigMalloc);
			ym->pBigMalloc = NULL;
			return NULL;
		}

		pSrc = p+pHeader.name_length;

		// Gros patch: PAS BO DU TOUT mais tant pis !!
//		*((UD*)pSrc) = pHeader->original;

		pSrc += 2;		// skip CRC16

		if (!LzhDepackBlock(pSrc,pNew,pHeader.original))
		{
			setLastError(ym,"LH5 Depacking Error !");
			free(pNew);
			free(ym->pBigMalloc);
			pNew = NULL;
			ym->pBigMalloc = NULL;
			return NULL;
		}

		// Tout est bon, le fichier est depack, on free le bloque
		// pack et on retourne le nouveau...
		free(ym->pBigMalloc);
		return pNew;
 }





static size_t	fileSizeGet(GPFILE *h)
 {
		return h->size;
 }


YM_BOOL	deInterleave(CYmMusic* ym)
 {
 long	nextPlane[32];
 UB	*pW,*tmpBuff;
 long	j,k;


		if (ym->attrib&A_STREAMINTERLEAVED)
		{

			tmpBuff = (UB*)malloc(ym->nbFrame*ym->streamInc);
			if (!tmpBuff)
			{
				setLastError(ym,"Malloc error in deInterleave()\n");
				return YM_FALSE;
			}

			// Precalcul les offsets.
			for (j=0;j<ym->streamInc;j++) nextPlane[j] = ym->nbFrame*j;

			pW = tmpBuff;
			for (j=0;j<nextPlane[1];j++)
			{
				for (k=0;k<ym->streamInc;k++)
				{
					pW[k] = ym->pDataStream[j + nextPlane[k]];
				}
				pW += ym->streamInc;
			}

			free(ym->pBigMalloc);
			ym->pBigMalloc = tmpBuff;
			ym->pDataStream = tmpBuff;

			ym->attrib &= (~A_STREAMINTERLEAVED);
		}
		return TRUE;
 }



YM_BOOL	ymDecode(CYmMusic* ym)
 {
 ymHeader_t *pHeader;
 UD *pUD;
 UB	*ptr;
 int skip;
 int i;
 unsigned long sampleSize;
 long tmp;


		pHeader = (ymHeader_t*)ym->pBigMalloc;
		switch (pHeader->id)
		{
			case 0x21324D59:	//'!2MY':		// MADMAX specific.
				ym->songType = YM_V2;
				ym->nbFrame = (ym->fileSize-4)/14;
				ym->loopFrame = 0;
				setClock(&ym->ymChip,ATARI_CLOCK);
				setPlayerRate(ym,50);
				ym->pDataStream = ym->pBigMalloc+4;
				ym->streamInc = 14;
				ym->nbDrum = 0;
				setAttrib(ym,A_STREAMINTERLEAVED|A_TIMECONTROL);
				ym->pSongName = mstrdup("");
				ym->pSongAuthor = mstrdup("Unknown");
				ym->pSongComment = mstrdup("Converted by Leonard.");
				ym->pSongType = mstrdup("YM 2");
				ym->pSongPlayer = mstrdup("YM-Chip driver.");
				break;

			case 0x21334D59:	//'!3MY':		// Standart YM-Atari format.
				ym->songType = YM_V3;
				ym->nbFrame = (ym->fileSize-4)/14;
				ym->loopFrame = 0;
				setClock(&ym->ymChip,ATARI_CLOCK);
				setPlayerRate(ym,50);
				ym->pDataStream = ym->pBigMalloc+4;
				ym->streamInc = 14;
				ym->nbDrum = 0;
				setAttrib(ym,A_STREAMINTERLEAVED|A_TIMECONTROL);
				ym->pSongName = mstrdup(""); //mstrdup(tmpFName);
				ym->pSongAuthor = mstrdup("Unknown");
				ym->pSongComment = mstrdup("");
				ym->pSongType = mstrdup("YM 3");
				ym->pSongPlayer = mstrdup("YM-Chip driver.");
				break;

			case 0x62334D59:	//'b3MY':		// Standart YM-Atari format + Loop info.
				pUD = (UD*)(ym->pBigMalloc+ym->fileSize-4);
				ym->songType = YM_V3;
				ym->nbFrame = (ym->fileSize-4)/14;
				ym->loopFrame = *pUD;
				setClock(&ym->ymChip,ATARI_CLOCK);
				setPlayerRate(ym,50);
				ym->pDataStream = ym->pBigMalloc+4;
				ym->streamInc = 14;
				ym->nbDrum = 0;
				setAttrib(ym,A_STREAMINTERLEAVED|A_TIMECONTROL);
				ym->pSongName = mstrdup(""); //mstrdup(tmpFName);
				ym->pSongAuthor = mstrdup("Unknown");
				ym->pSongComment = mstrdup("");
				ym->pSongType = mstrdup("YM 3b (loop)");
				ym->pSongPlayer = mstrdup("YM-Chip driver.");
				break;

			case 0x21344D59:	//'!4MY':		// Extended ATARI format.
				setLastError(ym,"No more YM4! support. Use YM5! format.");
				return YM_FALSE;
				break;

			case 0x21354D59: 	//'!5MY':		// Extended YM2149 format, all machines.
			case 0x21364D59:	//'!6MY':		// Extended YM2149 format, all machines.
				if (strncmp((const char*)(ym->pBigMalloc+4),"LeOnArD!",8))
				{
					setLastError(ym,"Not a valid YM format !");
					return YM_FALSE;
				}
				ptr = ym->pBigMalloc+12;
				ym->nbFrame = readMotorolaDword(&ptr);
				setAttrib(ym,readMotorolaDword(&ptr));
				ym->nbDrum = readMotorolaWord(&ptr);
				setClock(&ym->ymChip,readMotorolaDword(&ptr));
				setPlayerRate(ym,readMotorolaWord(&ptr));
				ym->loopFrame = readMotorolaDword(&ptr);
				skip = readMotorolaWord(&ptr);
				ptr += skip;
				if (ym->nbDrum>0)
				{
					ym->pDrumTab=(digiDrum_t*)malloc(ym->nbDrum*sizeof(digiDrum_t));
					for (i=0;i<ym->nbDrum;i++)
					{
						ym->pDrumTab[i].size = readMotorolaDword(&ptr);
						if (ym->pDrumTab[i].size)
						{
							ym->pDrumTab[i].pData = (UB*)malloc(ym->pDrumTab[i].size);
							memcpy(ym->pDrumTab[i].pData,ptr,ym->pDrumTab[i].size);
							if (ym->attrib&A_DRUM4BITS)
							{
								UD j;
								UB *pw = ym->pDrumTab[i].pData;
								for (j=0;j<ym->pDrumTab[i].size;j++)
								{
									*pw++ = ymVolumeTable0[(*pw)&15]>>7;
								}
							}
							ptr += ym->pDrumTab[i].size;
						}
						else
						{
							ym->pDrumTab[i].pData = NULL;
						}
					}
					ym->attrib &= (~A_DRUM4BITS);
				}
				ym->pSongName = readNtString((char**)&ptr);
				ym->pSongAuthor = readNtString((char**)&ptr);
				ym->pSongComment = readNtString((char**)&ptr);
				ym->songType = YM_V5;
				if (pHeader->id==0x21364D59)//'!6MY')
				{
					ym->songType = YM_V6;
					ym->pSongType = mstrdup("YM 6");
				}
				else
				{
					ym->pSongType = mstrdup("YM 5");
				}
				ym->pDataStream = ptr;
				ym->streamInc = 16;
				setAttrib(ym,A_STREAMINTERLEAVED|A_TIMECONTROL);
				ym->pSongPlayer = mstrdup("YM-Chip driver.");
				break;

			case 0x3158494D:	//'1XIM':		// ATARI Remix digit format.

				if (strncmp((const char*)(ym->pBigMalloc+4),"LeOnArD!",8))
				{
					setLastError(ym,"Not a valid YM format !");
					return YM_FALSE;
				}
				ptr = ym->pBigMalloc+12;
				ym->songType = YM_MIX1;
				tmp = readMotorolaDword(&ptr);
				setAttrib(ym,0);
				if (tmp&1) setAttrib(ym,A_DRUMSIGNED);
				sampleSize = readMotorolaDword(&ptr);
				ym->nbMixBlock = readMotorolaDword(&ptr);
				ym->pMixBlock = (mixBlock_t*)malloc(ym->nbMixBlock*sizeof(mixBlock_t));
				for (i=0;i<ym->nbMixBlock;i++)
				{	// Lecture des block-infos.
					ym->pMixBlock[i].sampleStart = readMotorolaDword(&ptr);
					ym->pMixBlock[i].sampleLength = readMotorolaDword(&ptr);
					ym->pMixBlock[i].nbRepeat = readMotorolaWord(&ptr);
					ym->pMixBlock[i].replayFreq = readMotorolaWord(&ptr);
				}
				ym->pSongName = readNtString((char**)&ptr);
				ym->pSongAuthor = readNtString((char**)&ptr);
				ym->pSongComment = readNtString((char**)&ptr);

				ym->pBigSampleBuffer = (unsigned char*)malloc(sampleSize);
				memcpy(ym->pBigSampleBuffer,ptr,sampleSize);

				if (!(ym->attrib&A_DRUMSIGNED))
				{
					signeSample(ym->pBigSampleBuffer,sampleSize);
					setAttrib(ym,A_DRUMSIGNED);
				}

				ym->mixPos = -1;		// numero du block info.
				ym->pSongType = mstrdup("MIX1");
				ym->pSongPlayer = mstrdup("Digi-Mix driver.");

				break;

			case 0x31544D59:	//'1TMY':		// YM-Tracker
			case 0x32544D59:	//'2TMY':		// YM-Tracker
/*;
; Format du YM-Tracker-1
;
; 4  YMT1
; 8  LeOnArD!
; 2  Nb voice
; 2  Player rate
; 4  Music lenght
; 4  Music loop
; 2  Nb digidrum
; 4  Flags		; Interlace, signed, 8 bits, etc...
; NT Music Name
; NT Music author
; NT Music comment
; nb digi *
*/
				if (strncmp((const char*)(ym->pBigMalloc+4),"LeOnArD!",8))
				{
					setLastError(ym,"Not a valid YM format !");
					return YM_FALSE;
				}
				ptr = ym->pBigMalloc+12;
				ym->songType = YM_TRACKER1;
				ym->nbVoice = readMotorolaWord(&ptr);
				setPlayerRate(ym,readMotorolaWord(&ptr));
				ym->nbFrame= readMotorolaDword(&ptr);
				ym->loopFrame = readMotorolaDword(&ptr);
				ym->nbDrum = readMotorolaWord(&ptr);
				ym->attrib = readMotorolaDword(&ptr);
				ym->pSongName = readNtString((char**)&ptr);
				ym->pSongAuthor = readNtString((char**)&ptr);
				ym->pSongComment = readNtString((char**)&ptr);
				if (ym->nbDrum>0)
				{
					ym->pDrumTab=(digiDrum_t*)malloc(ym->nbDrum*sizeof(digiDrum_t));
					for (i=0;i<(int)ym->nbDrum;i++)
					{
						ym->pDrumTab[i].size = readMotorolaWord(&ptr);
						ym->pDrumTab[i].repLen = ym->pDrumTab[i].size;
						if (pHeader->id==0x32544D59) //'2TMY')
						{
							ym->pDrumTab[i].repLen = readMotorolaWord(&ptr);	// repLen
							readMotorolaWord(&ptr);		// flag
						}
						if (ym->pDrumTab[i].repLen>ym->pDrumTab[i].size)
						{
							ym->pDrumTab[i].repLen = ym->pDrumTab[i].size;
						}

						if (ym->pDrumTab[i].size)
						{
							ym->pDrumTab[i].pData = (UB*)malloc(ym->pDrumTab[i].size);
							memcpy(ym->pDrumTab[i].pData,ptr,ym->pDrumTab[i].size);
							ptr += ym->pDrumTab[i].size;
						}
						else
						{
							ym->pDrumTab[i].pData = NULL;
						}
					}
				}

				ym->ymTrackerFreqShift = 0;
				if (pHeader->id==0x32544D59) //=='2TMY')
				{
					ym->ymTrackerFreqShift = (ym->attrib>>28)&15;
					ym->attrib &= 0x0fffffff;
					ym->pSongType = mstrdup("YM-T2");
				}
				else
				{
					ym->pSongType = mstrdup("YM-T1");
				}


				ym->pDataStream = ptr;
				setClock(&ym->ymChip,ATARI_CLOCK);

				ymTrackerInit(ym,100);		// 80% de volume maxi.
				ym->streamInc = 16;
				setTimeControl(ym,YM_TRUE);
				ym->pSongPlayer = mstrdup("Universal Tracker");
				break;

			default:
				setLastError(ym,"Unknow YM format !");
				return YM_FALSE;
				break;
		}

		if (!deInterleave(ym))
		{
			return YM_FALSE;
		}

		return YM_TRUE;
 }

YM_BOOL	load(CYmMusic* ym,char *fileName)
{
GPFILE	*in;


		stop(ym);
		unLoad(ym);

		in = smc_fopen(fileName,"rb");
		if (!in)
		{
			setLastError(ym,"File not Found");
			return YM_FALSE;
		}

		//---------------------------------------------------
		// Allocation d'un buffer pour lire le fichier.
		//---------------------------------------------------
		ym->fileSize = fileSizeGet(in);
		ym->pBigMalloc = (unsigned char*)malloc(ym->fileSize);
		if (!ym->pBigMalloc)
		{
			setLastError(ym,"MALLOC Error");
			smc_fclose(in);
			return YM_FALSE;
		}

		//---------------------------------------------------
		// Chargement du fichier complet.
		//---------------------------------------------------
		if (smc_fread(ym->pBigMalloc,1,ym->fileSize,in)!=ym->fileSize)
		{
			free(ym->pBigMalloc);
			setLastError(ym,"File is corrupted.");
			smc_fclose(in);
			return YM_FALSE;
		}
		smc_fclose(in);

		//---------------------------------------------------
		// Transforme les donnes en donnes valides.
		//---------------------------------------------------
//		pMus->fileSize = fileSize;
		ym->pBigMalloc = depackFile(ym);
		if (!ym->pBigMalloc)
		{
			return YM_FALSE;
		}

		//---------------------------------------------------
		// Lecture des donnes YM:
		//---------------------------------------------------
		if (!ymDecode(ym))
		{
			free(ym->pBigMalloc);
			ym->pBigMalloc = NULL;
			return YM_FALSE;
		}

		CYm2149ExReset(&ym->ymChip);
		ym->bMusicOk = YM_TRUE;
		ym->bPause = YM_FALSE;
		return YM_TRUE;
 }

YM_BOOL	loadMemory(CYmMusic* ym,void *pBlock,unsigned long size)
{

		stop(ym);
		unLoad(ym);

		//---------------------------------------------------
		// Allocation d'un buffer pour lire le fichier.
		//---------------------------------------------------
		ym->fileSize = size;
		ym->pBigMalloc = (unsigned char*)malloc(ym->fileSize);
		if (!ym->pBigMalloc)
		{
			setLastError(ym,"MALLOC Error");
			return YM_FALSE;
		}

		//---------------------------------------------------
		// Chargement du fichier complet.
		//---------------------------------------------------
		memcpy(ym->pBigMalloc,pBlock,size);

		//---------------------------------------------------
		// Transforme les donnes en donnes valides.
		//---------------------------------------------------
		ym->pBigMalloc = depackFile(ym);
		if (!ym->pBigMalloc)
		{
			return YM_FALSE;
		}

		//---------------------------------------------------
		// Lecture des donnes YM:
		//---------------------------------------------------
		if (!ymDecode(ym))
		{
			free(ym->pBigMalloc);
			ym->pBigMalloc = NULL;
			return YM_FALSE;
		}

		CYm2149ExReset(&ym->ymChip);
		ym->bMusicOk = YM_TRUE;
		ym->bPause = YM_FALSE;
		return YM_TRUE;
 }

void	myFree(void **pPtr)
{
		if (*pPtr) free(*pPtr);
		*pPtr = NULL;
}

void	unLoad(CYmMusic* ym)
{

		ym->bMusicOk = YM_FALSE;
		ym->bPause = YM_TRUE;
		ym->bMusicOver = YM_FALSE;
		myFree((void**)&ym->pSongName);
		myFree((void**)&ym->pSongAuthor);
		myFree((void**)&ym->pSongComment);
		myFree((void**)&ym->pSongType);
		myFree((void**)&ym->pSongPlayer);
		myFree((void**)&ym->pBigMalloc);
		if (ym->nbDrum>0)
		{
			int i;
			for (i=0;i<ym->nbDrum;i++)
			{
				myFree((void**)&ym->pDrumTab[i].pData);
			}
			ym->nbDrum = 0;
			myFree((void**)&ym->pDrumTab);
		}
		myFree((void**)&ym->pBigSampleBuffer);
		myFree((void**)&ym->pMixBlock);

}

void	stop(CYmMusic* ym)
{
	ym->bPause = YM_TRUE;
	ym->currentFrame = 0;
	ym->mixPos = -1;
}

void	play(CYmMusic* ym)
{
	ym->bPause = YM_FALSE;
}

void	pause(CYmMusic* ym)
{
	ym->bPause = YM_TRUE;
}
