#include "gp32.h"
#include "fileio.h"
#include "YmMusic.h"
#include "smf_fat.h"

#include <stdlib.h>
#include <string.h>

unsigned short *framebuffer;

int py = 10;
void println(char*s)
{
	gp_drawString(20,py,strlen(s),s,0x0000,framebuffer);
	py+=10;
}

#define SEGMENTSIZE (SMP_SIZE<<2)
#define SMP_SIZE 992

static YMMUSIC *pMusic;

static void MyTIMER4IRQ(void) __attribute__ ((interrupt ("IRQ")));

static void MyTIMER4IRQ(void)
{
	rTCON = BITCLEAR(rTCON,20); // timer4 off
	ymMusicCompute(pMusic,gp_getRingSegment(),SMP_SIZE);
	rTCON = BITSET(rTCON,20); // timer4 on
}

static int initsound=1;

void gp_startSound()
{
	int pclk;
	pclk = gp_getPCLK();
	pclk/= 16;
	pclk/= 256;
	pclk/= 48;  // 51Hz timer, tested 33 66 133 Mhz

	if (initsound)
	{
		gp_initSound(YM_SMPRATE,16,SEGMENTSIZE*2); 
		initsound=0; 
	}
	gp_disableIRQ();
	rTCFG0 |= (0xFF<<8);   // Presacler for timer 2,3,4 = 256
	rTCFG1 |= (0x03<<16);  // timer4  1/16
	rTCNTB4 = (long)pclk;
	rTCON  = (0x1<<22) | (0x1<<20); // start timer4, auto reload
	gp_installSWIIRQ(14,MyTIMER4IRQ);
	gp_enableIRQ();
}

//static 
void gp_stopSound()
{
	gp_disableIRQ();
	rTCON = BITCLEAR(rTCON,20); // timer4 off
	gp_removeSWIIRQ(14,MyTIMER4IRQ);
	gp_enableIRQ();
	gp_clearRingbuffer();
}

#define YM_DIR "dev0:\\GPMM\\YM\\"

void EnableCache();
ymMusicInfo_t info;

void printInfos(char* name)
{
	gp_clearFramebuffer16(framebuffer,0xffff);
	py = 10;
	println(name);
	println(info.pSongName);
	println(info.pSongAuthor);
	println(info.pSongType);
	println("PLAYING");
	py+=10;
	println("A     : NEXT");
	println("B     : PREVIOUS");
	println("L+R   : LOCK/UNLOCK");
	println("START : EXIT");
	py+=10;
}

F_HANDLE ym_file = 0;

unsigned int openFile(const char *path)
{
    char name[64];
    unsigned int size;

    if (smInit())
        return 1;

    sprintf(name,"%s%s",YM_DIR,path);
    println(name);
    smOpenFile(name,OPEN_R,&ym_file);
    if (ym_file == 0)
        return 0;

    smSeekFile(ym_file, FROM_END, 0, (int*) &size);
    smSeekFile(ym_file, FROM_BEGIN, 0, (int*) &size);

    return size;
}

void closeFile()
{
    smCloseFile(ym_file);
    ym_file = NULL;
}

int main()
{
	DIR dir_list;
	int nbf,cur,lock=0;
    unsigned int size;
    unsigned int tmp;
    unsigned char* data;

	framebuffer = (u16*) FRAMEBUFFER;
	gp_setCpuspeed(22);
	EnableCache();
	//gp_setMMU(framebuffer, ((unsigned char*)framebuffer)+153600-1, 0xFF2);
	gp_initFramebuffer(framebuffer,16,85);

	nbf = smc_dir(YM_DIR, &dir_list );
	cur = 2;
  
	while (1)
	{
		char buf[64];
		gp_clearFramebuffer16(framebuffer,0xffff);
		py = 10;
		sprintf(buf,"LOADING %s",dir_list.name[cur]);
		println(buf);
        size = openFile(dir_list.name[cur]);
        sprintf(buf,"%d",size);
        println(buf);
        data = (unsigned char*)malloc(size);
        smReadFile(ym_file, data, size, &tmp);
        sprintf(buf,"%d",tmp);
        println(buf);
		//fprintf(buf,"%s%s",YM_DIR,dir_list.name[cur]);
		//pMusic = ymMusicLoad(buf);
        pMusic = ymMusicLoadMemory(data,size);
		if (pMusic)
		{
			ymMusicGetInfo(pMusic,&info);
			printInfos(dir_list.name[cur]);
			ymMusicSetLoopMode(pMusic,YM_TRUE);
			ymMusicPlay(pMusic);
			gp_startSound();
			while(1)
			{
				int key = gp_getButton();
				if ((lock==0) && (key&BUTTON_START))
					gp_Reset();
				else if ((lock==0) && (key&BUTTON_A))
					cur = cur>2?cur-1:nbf-1;
				else if ((lock==0) && (key&BUTTON_B))
					cur = cur+1<nbf?cur+1:2;
				else if ((key&(BUTTON_L|BUTTON_R))==(BUTTON_L|BUTTON_R))
				{
					while(gp_getButton());
					printInfos(dir_list.name[cur]);
					lock ^= 1;
					if (lock)
						println("LOCKED");
					continue;
				}
				else
					continue;
				while(gp_getButton());
				break;
			}
			ymMusicStop(pMusic);
			gp_stopSound();
			ymMusicUnload(pMusic);
		}
		else
		{
			gp_drawString(20,30,5,"ERROR",0xF800,framebuffer);
			char* err = ymMusicGetLastError(pMusic);
			if (err!=NULL)
				gp_drawString(20,40,strlen(err),err,0xF800,framebuffer);
			if (pMusic)
				ymMusicUnload(pMusic);
			while(1)
			{
				int key = gp_getButton();
				if (key&BUTTON_START)
					gp_Reset();
				else if (key&BUTTON_A)
					cur = cur>2?cur-1:nbf-1;
				else if (key&BUTTON_B)
					cur = cur+1<nbf?cur+1:2;
				else
					continue;
				while(gp_getButton());
				break;
			}
		}
	}
}
