#ifdef __GP32__
#include "gp32.h"
#endif
#include "dbg.h"
#include "o3d.h"
#include "tables.h"
#include <stdlib.h>
#include <string.h>

#ifdef __GP32__
extern void error(char*);
extern void gp_setMMU(unsigned long,unsigned long,unsigned long);
extern void gp_waitDMA (unsigned char);
#endif

#define SPEED

#ifdef SPEED // 16bits, no zbuffer, no transparency, no cell shading, no stars
#define OVERLOADS
#define BITS_16
//#define NO_TEXTURE
//#define FORCE_FLAT

#else // Custom

#define BITS_16
#define ZBUFFER
//#define USE_TRANS
//#define CELLSHADING
//#define GHOST
//#define NO_TEXTURE
//#define FORCE_FLAT
#define OVERLOADS
//#define STARS
//#define POINT_LIGHT

#endif

#define QSORT2

int __drawn__;

#define LSHIFT (DIV_SHIFT+6) // 2^6 == 64 == (MULT*2)/32 : MULT*2 = max light vector length ; 32 : light range

extern unsigned int isqrt(unsigned int);

#define NEGATIVE_INFINITY 0x80808080
#define NEGATIVE_INFINITY_BYTE 0x80

#define MINX(a,b,c) (MIN(MIN(GET(drawvert,a).x,GET(drawvert,b).x),GET(drawvert,c).x))
#define MAXX(a,b,c) (MAX(MAX(GET(drawvert,a).x,GET(drawvert,b).x),GET(drawvert,c).x))
#define MINY(a,b,c) (MIN(MIN(GET(drawvert,a).y,GET(drawvert,b).y),GET(drawvert,c).y))
#define MAXY(a,b,c) (MAX(MAX(GET(drawvert,a).y,GET(drawvert,b).y),GET(drawvert,c).y))

#define QMINX(a,b,c,d) (MIN(MIN(MIN(GET(drawvert,a).x,GET(drawvert,b).x),GET(drawvert,c).x),GET(drawvert,d).x))
#define QMAXX(a,b,c,d) (MAX(MAX(MAX(GET(drawvert,a).x,GET(drawvert,b).x),GET(drawvert,c).x),GET(drawvert,d).y))
#define QMINY(a,b,c,d) (MIN(MIN(MIN(GET(drawvert,a).y,GET(drawvert,b).y),GET(drawvert,c).y),GET(drawvert,d).x))
#define QMAXY(a,b,c,d) (MAX(MAX(MAX(GET(drawvert,a).y,GET(drawvert,b).y),GET(drawvert,c).y),GET(drawvert,d).y))

#define MULTIPLY(m,v,r) \
            (r).x = ((v).x * (m)[0] + (v).y * (m)[4] + (v).z * (m)[8]  + (m)[12])>>SHIFT; \
            (r).y = ((v).x * (m)[1] + (v).y * (m)[5] + (v).z * (m)[9]  + (m)[13])>>SHIFT; \
            (r).z = ((v).x * (m)[2] + (v).y * (m)[6] + (v).z * (m)[10] + (m)[14])>>SHIFT

#define MULTIPLY_P(m,v,r,s) \
            (r).x = (((v).x * (m)[0] + (v).y * (m)[4] + (v).z * (m)[8]  + (m)[12])>>SHIFT) - (s).x; \
            (r).y = (((v).x * (m)[1] + (v).y * (m)[5] + (v).z * (m)[9]  + (m)[13])>>SHIFT) - (s).y; \
            (r).z = (((v).x * (m)[2] + (v).y * (m)[6] + (v).z * (m)[10] + (m)[14])>>SHIFT) - (s).z

#define MULTIPLY_X(m,v,r) \
            (r).x = ((v).x * (m)[0] + (v).y * (m)[4] + (v).z * (m)[8]  + (m)[12])>>SHIFT

#define MULTIPLY_Y(m,v,r) \
            (r).y = ((v).x * (m)[1] + (v).y * (m)[5] + (v).z * (m)[9]  + (m)[13])>>SHIFT

#define MULTIPLY_Z(m,v,r) \
            (r).z = ((v).x * (m)[2] + (v).y * (m)[6] + (v).z * (m)[10] + (m)[14])>>SHIFT

#define MULTIPLY_PX(m,v,r,s) \
            (r).x = (((v)->x * (m)[0] + (v)->y * (m)[4] + (v)->z * (m)[8]  + (m)[12])>>SHIFT) - (s)

#define MULTIPLY_PY(m,v,r,s) \
            (r).y = (((v)->x * (m)[1] + (v)->y * (m)[5] + (v)->z * (m)[9]  + (m)[13])>>SHIFT) - (s)

#define MULTIPLY_PZ(m,v,r,s) \
            (r).z = (((v)->x * (m)[2] + (v)->y * (m)[6] + (v)->z * (m)[10] + (m)[14])>>SHIFT) - (s)

#define DOT_PRODUCT(v0,v1) \
            ((v0).x * (v1).x + (v0).y * (v1).y + (v0).z * (v1).z)

const SVECTOR3D    zero = {0,0,0};

unsigned int    max_vert;
unsigned int    max_tri;
VERTEX3D        *vertices;
VECTOR2D        *drawvert;
REFTRI3D        *triangles;
VECTOR3D        *tnorms;
VERTEX3D        *tmid;
unsigned char   *vis;
unsigned int    *vlights;
VECTOR3D        *vnorms;

#define MAX_COLORS 512
LCOLOR3D*       lcolors[MAX_COLORS];
int             nbcolors;

#ifdef OVERLOADS
TRIANGLE3D_DATA**overloads;
#endif

#ifdef STARS
SVERTEX3D        stars[MAX_STARS];
int              nbStars;
int              zStars;
#endif

#ifdef BITS_16
unsigned short   *view3d;
#else
unsigned char    *view3d;
#endif

NODE3D            main3d;
NODE3D            *nlight3d;
VECTOR3D          light3d,tlight3d;
unsigned int      lightDef;
int               proj3d;
COLOR3D           col3d;
unsigned short    cscol3d_ext,cscol3d_int;
int               gdMin;
int               zMin;
int               zMax;
TRIANGLE3D_TYPE   shading3d;

unsigned short    _curv,_curt,_curs,_curId = 0;

#ifdef ZBUFFER

short* zbuffers[2];
short* zbuf;
short  zflip;

#else

NODE3D*           *nspr;
VECTOR3D          *vspr;
unsigned short    *sort_a;
#if !defined(HEAPSORT)
unsigned short    *sort_b;
#endif
#if defined(QSORT) || defined(QSORT2)
unsigned short    *sort_c;
#endif
#ifdef QSORT2
unsigned short    *sort_st;
#endif

#ifdef QSORT
void o3d_qsort(unsigned short*a,unsigned short*b,unsigned short start,unsigned short end);
#else
#ifdef QSORT2
void o3d_qsort2(int max);
#else
#ifdef HEAPSORT
void o3d_heapSort(unsigned short count);
#else
#define RADIXSORT
void o3d_radixSort(unsigned short*source,unsigned short*temp,unsigned short N);
#endif
#endif
#endif

#endif

#ifdef CELLSHADING

#define NO_FROM 0xffff
#define NO_FROM_BYTE 0xff

unsigned short*    cs_from;
unsigned short*    cs_obj;

#define GPLIGHT(a) (((a)>>LSHIFT)&0x1c) // 4 luminosity values

#else

#define GPLIGHT(a) ((a)>>LSHIFT)

#endif

unsigned char tabInit = 0;
unsigned char _tabCol_[258];
unsigned char *tabCol;

/**
 * MACROS
 */

#define MIN(a,b) (a<b?a:b)
#define MAX(a,b) (a>b?a:b)

#ifdef __GP32__
#define GPCOLX(r,g,b) ((tabCol[r]<<8)|(tabCol[g]<<3)|(tabCol[b]>>2))
#else
#define GPCOLX(r,g,b) ((tabCol[r]<<8)|(tabCol[g]<<3)|(tabCol[b]>>3))
#endif
#define GPCOL3DX(c)   GPCOLX((c)->r,(c)->g,(c)->b)

#define LIGHT(a,b,l)  (l>MULT? a+((b*(l-MULT))>>SHIFT) : ((a*l)>>SHIFT))

/*
 * Internal API
 */

void initMatrix(MATRIX3D,MATRIX_TYPE,short,short,short);
void initRotTransMatrix(MATRIX3D,ROTATION3D *,VECTOR3D*);
void multiplyMatrix(MATRIX3D,MATRIX3D,MATRIX3D);
void multiplyVector(MATRIX3D,VECTOR3D*,VECTOR3D*);
void multiplySVector(MATRIX3D,const SVECTOR3D*,VECTOR3D*);
void normalize(VECTOR3D*);
void clearBuffers(int);

void transformNode(NODE3D*,MATRIX3D);
void transformVertices(NODE3D*);

void showSprite(NODE3D*,VECTOR3D*);
void computeLights(unsigned short,unsigned short,short*,short*,short*,COLOR3D*);
#ifdef POINT_LIGHT
unsigned int computeLight(VERTEX3D*,VECTOR3D*n);
#else
unsigned int computeLight(VECTOR3D*n);
#endif
void computeColor(unsigned int,COLOR3D*,COLOR3D*);
void drawFlatTriangle(unsigned short,unsigned short,unsigned short,unsigned short,unsigned short);
void drawGouraudTriangle(unsigned short,unsigned short,unsigned short,unsigned short*,unsigned short);
void drawFlatTexturedTriangle(unsigned short,unsigned short,unsigned short,unsigned char,unsigned char,unsigned char,int l,TEXTURE3D_DATA*,TEXTURE3D*,unsigned short);
void drawGouraudTexturedTriangle(unsigned short,unsigned short,unsigned short,unsigned char,unsigned char,unsigned char,TEXTURE3D_DATA*,TEXTURE3D*,unsigned short);
#ifdef FORCE_FLAT
void drawTriangle(unsigned short,unsigned short,unsigned short,unsigned short,unsigned char,unsigned char);
#else
void drawTriangle(unsigned short,unsigned short,unsigned short,unsigned short,unsigned char);
#endif
void showStars(void);

#ifdef __GP32__

/**
 * DMA copy with dword alignment
 */
void gp_setDMA4( u32 s, u32 d, u32 size,u8 channel )
{
    // DMA Control register
    int DMD_HS = 0;  // 0=demand 1=handshake mode
    int SYNC   = 1;  // 0=sync to PCLK/APB 1=HCLK/AHB
    int INT    = 0;  // 0=irq disable 1=irq enable (when transfer is done)
    int TSZ    = 0;  // transfer size 0=one 1=four units
    int SERVMODE=1;  // 0=single service 1=whole service
    int HWSRCSEL=0;  // 0=nXDREQ0 DMA Source
    int SWHW_SEL=0;  // 0=SW request mode 1=source selected by bit HWSRCSEL
    int RELOAD = 1;  // 0=auto reload 1=turned off
    int DSZ    = 2;  // 0=8bit 1=16bit 2=32bit  data size
    int TC     = (size>>DSZ); // Transfercount DSZ*TSZ*TC

    u32 rdcon  = (DMD_HS<<30)|(SYNC<<29)|(INT<<28)|(TSZ<<27)|(SERVMODE<<26)
                |(HWSRCSEL<<24)|(SWHW_SEL<<23)|(RELOAD<<22)|(DSZ<<20)|(TC<<0);

    // DMA channel on in Software mode
    gp_waitDMA(channel);

    switch (channel)
    {
        case 0 : {  rDISRC0 = s; rDIDST0 = d; rDCON0 = rdcon; rDMASKTRIG0 = 3;  break; }
        case 1 : {  rDISRC1 = s; rDIDST1 = d; rDCON1 = rdcon; rDMASKTRIG1 = 3;  break; }
        case 2 : {  rDISRC2 = s; rDIDST2 = d; rDCON2 = rdcon; rDMASKTRIG2 = 3;  break; }
        case 3 : {  rDISRC3 = s; rDIDST3 = d; rDCON3 = rdcon; rDMASKTRIG3 = 3;  break; }
    }
}
#endif

void clearBuffers(int wait)
{
#ifdef ZBUFFER
#ifdef __GP32__
    unsigned long *ptr = (unsigned long*)zbuf;
    *ptr = NEGATIVE_INFINITY;
    gp_setDMA4((u32)ptr,(u32)(ptr+1),(SIZE<<1)-4,0);
#else
    memset(zbuf,NEGATIVE_INFINITY_BYTE,SIZE<<1);
#endif
#endif
#ifdef CELLSHADING
    memset(cs_from,NO_FROM_BYTE,SIZE<<1);
#endif
#ifdef __GP32__
#ifdef ZBUFFER
    if (wait) gp_waitDMA(0);
#endif
#endif
}

/**
 * Color tables
 */

void computeColor(unsigned int l,COLOR3D *col,COLOR3D *res)
{
    int c = LIGHT(col->r,col3d.r,l);
    res->r = MIN(c,GP_CC_MAX);
    c = LIGHT(col->g,col3d.g,l);
    res->g = MIN(c,GP_CC_MAX);
    c = LIGHT(col->b,col3d.b,l);
    res->b = MIN(c,GP_CC_MAX);
}

void initColorTab()
{
    int i;
    if (tabInit)
        return;
    tabCol = _tabCol_ + 1;
    for (i=0;i<256;++i)
        tabCol[i] = GPCOLU(i);
    // Precision error (+-1):
    _tabCol_[0] = tabCol[0];
    _tabCol_[257] = tabCol[255];
    tabInit = 1;
}

/**
 * Init engine with main scene
 */

NODE3D* o3d_init(SCENE3D *s,int maxv,int maxt,int maxs,int proj)
{
    memset(&main3d,0,sizeof(NODE3D));
    main3d.data.s = s;
    main3d.type = SCENE;
    proj3d = proj;
    col3d.r = 127;
    col3d.g = 127;
    col3d.b = 127;
    cscol3d_ext = 0;
    cscol3d_int = 0;
    nlight3d = &main3d;
    light3d.x = 0;
    light3d.y = 0;
    light3d.z = 0;
    lightDef = 16;
    gdMin = -1000;
    zMin = -1500;
    zMax = 500;
#ifdef STARS
    nbStars = 0;
#endif
    shading3d = FLAT;

    /* memory alloc */
    max_vert = maxv;
    max_tri = maxt;
    vertices    = NULL;
    drawvert    = (VECTOR2D*)MALLOC(sizeof(VECTOR2D),maxv);
    triangles   = (REFTRI3D*)MALLOC(sizeof(REFTRI3D),maxt);
    tnorms      = (VECTOR3D*)MALLOC(sizeof(VECTOR3D),maxt);
    vis         = (unsigned char*)MALLOC(sizeof(unsigned char),maxv);
    vlights     = NULL;
    vnorms      = NULL;
    tmid        = (VERTEX3D*)MALLOC(sizeof(VERTEX3D),maxt);

#ifdef OVERLOADS
    overloads   = (TRIANGLE3D_DATA**)MALLOC(sizeof(TRIANGLE3D_DATA*),maxt);
#endif

#ifdef ZBUFFER
    zbuffers[0] = (short*)MALLOC(sizeof(short),SIZE);
    zbuffers[1] = (short*)MALLOC(sizeof(short),SIZE);
#else
    nspr      = maxs==0? NULL : (NODE3D**)MALLOC(sizeof(NODE3D*),maxs);
    vspr      = maxs==0? NULL : (VECTOR3D*)MALLOC(sizeof(VECTOR3D),maxs);
    sort_a    = (unsigned short*)MALLOC(sizeof(unsigned short),maxt+maxs);
#if !defined(HEAPSORT)
    sort_b    = (unsigned short*)MALLOC(sizeof(unsigned short),maxt+maxs);
#endif
#if defined(QSORT) || defined(QSORT2)
    sort_c    = (unsigned short*)MALLOC(sizeof(unsigned short),maxt+maxs);
#endif
#ifdef QSORT2
    sort_st   = (unsigned short*)MALLOC(sizeof(unsigned short),(maxt+maxs)*3);
#endif
#endif

#ifdef CELLSHADING
    cs_from = (unsigned short*)MALLOC(sizeof(unsigned short),SIZE);
    cs_obj  = (unsigned short*)MALLOC(sizeof(unsigned short),maxt);
#endif
    initColorTab();

#ifdef ZBUFFER
    zbuf = zbuffers[1];
    clearBuffers(0);
    zflip = 0;
    zbuf = zbuffers[0];
#endif
    clearBuffers(1);

    _curv = 0;

    return &main3d;
}

void o3d_free()
{
    FREE(drawvert);
    FREE(triangles);
    FREE(tnorms);
    FREE(vis);
    if (vertices!=NULL)
    {
        FREE(vertices);
        FREE(vlights);
        FREE(vnorms);
    }
    FREE(tmid);

#ifdef OVERLOADS
    FREE(overloads);
#endif

#ifdef ZBUFFER
#ifdef __GP32__
    gp_waitDMA(0);
#endif
    FREE(zbuffers[0]);
    FREE(zbuffers[1]);
#else
    if (nspr!=NULL)
    {
        FREE(nspr);
        FREE(vspr);
    }
    FREE(sort_a);
#if !defined(HEAPSORT)
    FREE(sort_b);
#endif
#if defined(QSORT) || defined(QSORT2)
    FREE(sort_c);
#endif
#ifdef QSORT2
    FREE(sort_st);
#endif
#endif

#ifdef CELLSHADING
    FREE(cs_from);
    FREE(cs_obj);
#endif
}

VECTOR3D *o3d_setLightPos(NODE3D*n,int x,int y,int z)
{
    nlight3d = n == NULL ? &main3d : n;
    light3d.x = x;
    light3d.y = y;
    light3d.z = z;
#ifndef POINT_LIGHT
    normalize(&light3d);
#endif
    return &light3d;
}

COLOR3D *o3d_setLightColor(unsigned char r,unsigned char g,unsigned char b)
{
    unsigned int i,j;
    COLOR3D cl;
    col3d.r = r;
    col3d.g = g;
    col3d.b = b;
    for (i=0;i<32;++i)
    {
        for (j=0;j<nbcolors;++j)
        {
            computeColor((i*2*MULT)>>5,&lcolors[j]->col,&cl);
            lcolors[j]->pix[i] = GPCOLX(cl.r,cl.g,cl.b);
        }
    }
    for (j=0;j<nbcolors;++j)
        lcolors[j]->pix[32] = lcolors[j]->pix[31];
    return &col3d;
}

void o3d_setLightDefaut(unsigned int def)
{
    lightDef = MIN(31,def);
}

void o3d_setBorderColor(unsigned short external,unsigned short internal)
{
    cscol3d_ext = external;
    cscol3d_int = internal;
}

void o3d_setGDMin(int min)
{
    gdMin = min;
}

void o3d_setDepths(int zmin,int zmax)
{
    zMin = zmin;
    zMax = zmax;
}

void o3d_setShading(TRIANGLE3D_TYPE s)
{
    shading3d = s;
    if (shading3d==GOURAUD)
    {
        if (vertices!=NULL)
            return;
        vertices = (VERTEX3D*)MALLOC(sizeof(VERTEX3D),max_vert);
        vlights  = (unsigned int*)MALLOC(sizeof(unsigned int),max_vert);
        vnorms   = (VECTOR3D*)MALLOC(sizeof(VECTOR3D),max_vert);
    }
    else if (vertices!=NULL)
    {
        FREE(vertices);
        FREE(vlights);
        FREE(vnorms);
        vertices = NULL;
        vlights  = NULL;
        vnorms   = NULL;
    }
}

void o3d_initScene(SCENE3D *s)
{
    s->count = 0;
    s->reserved = 0;
    s->nodes = NULL;
}

void o3d_clearScene(SCENE3D *s)
{
    int i;
    for (i=0;i<s->count;i++)
    {
        if (s->nodes[i]->type == SCENE)
            o3d_clearScene(s->nodes[i]->data.s);
        FREE(s->nodes[i]);
    }
    s->count = 0;
}

NODE3D* o3d_addObject(SCENE3D *s,OBJECT3D* o,int x,int y,int z,unsigned short rx,unsigned short ry,unsigned short rz)
{
    return o3d_addNode(s,o,OBJECT,x,y,z,rx,ry,rz);
}

NODE3D* o3d_addScene(SCENE3D *s,SCENE3D* sc,int x,int y,int z,unsigned short rx,unsigned short ry,unsigned short rz)
{
    return o3d_addNode(s,sc,SCENE,x,y,z,rx,ry,rz);
}

NODE3D* o3d_addSprite(SCENE3D *s,SPRITE3D* sp,int x,int y,int z)
{
    return o3d_addNode(s,sp,SPRITE,x,y,z,0,0,0);
}

NODE3D* o3d_addAnimation(SCENE3D *s,ANIMATION3D* a,int x,int y,int z,unsigned short rx,unsigned short ry,unsigned short rz)
{
    return o3d_addNode(s,a,ANIMATION,x,y,z,rx,ry,rz);
}

NODE3D* o3d_addNode(SCENE3D *s,void* n,NODE3D_TYPE t,int x,int y,int z,unsigned short rx,unsigned short ry,unsigned short rz)
{
    NODE3D* node;
    if (s->reserved==0 || s->reserved==s->count)
    {
        NODE3D **save = s->nodes;
        s->reserved = s->reserved+NODE3D_ALLOC;
        s->nodes = (NODE3D**)MALLOC(sizeof(NODE3D*),s->reserved);
        if (s->count>0)
        {
            memcpy(s->nodes,save,sizeof(NODE3D*)*s->count);
            FREE(save);
        }
    }
    node = (NODE3D*)MALLOC(sizeof(NODE3D),1);
    node->type      = t;
    node->data.n    = n;
    node->pos.x     = x;
    node->pos.y     = y;
    node->pos.z     = z;
    node->rot.rx    = rx;
    node->rot.ry    = ry;
    node->rot.rz    = rz;
    node->overload  = NULL;
    node->hide      = 0;
    node->id        = _curId++;
    node->cur_frame = 0;
    s->nodes[s->count++] = node;
    return node;
}

void o3d_removeNodeAt(SCENE3D *s,int i)
{
    s->count-=1;
    FREE(s->nodes[i]);
    if (i<s->count)
        memcpy(&s->nodes[i],&s->nodes[i+1],sizeof(NODE3D*)*(s->count-i));
}

void o3d_removeNode(SCENE3D *s,NODE3D* o)
{
    int i;
    for (i=0;i<s->count;++i)
    {
        if (s->nodes[i]==o)
        {
            o3d_removeNodeAt(s,i);
            break;
        }
    }
}

SCENE3D *o3d_cloneScene(SCENE3D *src)
{
    int i;
    void *data;
    SCENE3D* dst;

    dst = (SCENE3D*)MALLOC(sizeof(SCENE3D),1);
    dst->count = 0;
    for (i=0;i<src->count;++i)
    {
        NODE3D *n = src->nodes[i];
        if (n->type==SCENE)
            data = o3d_cloneScene(n->data.s);
        else
            data = n->data.n;
        o3d_addNode(dst,data,n->type,n->pos.x,n->pos.y,n->pos.z,n->rot.rx,n->rot.ry,n->rot.rz);
    }
    return dst;
}

void o3d_destroyClone(SCENE3D *s)
{
    int i;
    for (i=0;i<s->count;++i)
    {
        if (s->nodes[i]->type==SCENE)
            o3d_destroyClone(s->nodes[i]->data.s);
        FREE(s->nodes[i]);
    }
    FREE(s->nodes);
    FREE(s);
}

void o3d_registerColor(LCOLOR3D* c)
{
    lcolors[nbcolors++] = c;
}

void o3d_unregisterColor(LCOLOR3D* c)
{
    int i;
    for (i=0;i<nbcolors;++i)
    {
        if (lcolors[i]==c)
        {
            nbcolors-=1;
            if (i<nbcolors)
                memcpy(&lcolors[i],&lcolors[i+1],sizeof(LCOLOR3D*)*(nbcolors-i));
        }
    }
}

void o3d_unregisterColors()
{
    nbcolors=0;
}

#ifdef CELLSHADING

int nolink(unsigned short t0,unsigned short t1)
{
    return !((GET(triangles,t0).a==GET(triangles,t1).a ||
              GET(triangles,t0).a==GET(triangles,t1).b ||
              GET(triangles,t0).a==GET(triangles,t1).c) ||
             (GET(triangles,t0).b==GET(triangles,t1).a ||
              GET(triangles,t0).b==GET(triangles,t1).b ||
              GET(triangles,t0).b==GET(triangles,t1).c) ||
             (GET(triangles,t0).c==GET(triangles,t1).a ||
              GET(triangles,t0).c==GET(triangles,t1).b ||
              GET(triangles,t0).c==GET(triangles,t1).c));
}

int nocol(unsigned short t0,unsigned short t1)
{
    TRIANGLE3D_TYPE tt0 = GET(triangles,t0).orig->type;
    TRIANGLE3D_TYPE tt1 = GET(triangles,t1).orig->type;
    COLOR3D *c0,*c1;
    if (tt0==TEXTURED_F || tt0==TEXTURED_G || tt0==TEXTURED)
    {
        if (tt1==TEXTURED_F || tt1==TEXTURED_G || tt1==TEXTURED)
            return GET(triangles,t0).orig->data.td->t!=GET(triangles,t1).orig->data.td->t;
        else
            return 1;
    }
    if (tt1==TEXTURED_F || tt1==TEXTURED_G || tt1==TEXTURED)
        return 1;
    c0 = &GET(triangles,t0).orig->data.c->col;
    c1 = &GET(triangles,t1).orig->data.c->col;
    return c0->r!=c1->r || c0->g!=c1->g || c0->b!=c1->b;
}

#endif

int __vis__;

void o3d_computeFrame(unsigned short *v, int clr)
{
    unsigned short j;
#ifndef ZBUFFER
    unsigned short l,m;
    int n;
#endif
#ifdef CELLSHADING
    unsigned short i,f,fw,fh;
    int kh,kl,k;
#endif

__drawn__ = 0;
__vis__ = 0;

    if (clr)
#ifdef __GP32__
    {
        unsigned long *fbuf = (unsigned long*)v;
        *fbuf = 0;
        gp_setDMA4((unsigned long)fbuf,(unsigned long)(fbuf+1),(SIZE<<1)-4,0);
    }
#else
        memset(v,0,SIZE<<1);
#endif

    view3d = v;
#ifdef ZBUFFER
    zbuf = zbuffers[zflip];
    zflip ^= 1;
#endif

    /* compute transformations */
    transformNode(&main3d,NULL);

    if (shading3d!=NONE)
    {
        MULTIPLY(nlight3d->m,light3d,tlight3d);
    }

    if (_curv)
        MEMSET(vis,0,_curv);
    _curv  = 0;
    _curt  = 0;
    _curs = 0;
    transformVertices(&main3d);

    /* drawing */
#ifdef ZBUFFER
    if (clr)
        gp_waitDMA(0);
    for (j=0;j<_curt;++j)
    {
        unsigned short a = GET(triangles,j).a;
        unsigned short b = GET(triangles,j).b;
        unsigned short c = GET(triangles,j).c;
        unsigned short d = GET(triangles,j).d;
        if ((drawvert[a].x >= 0     || drawvert[b].x >= 0     || drawvert[c].x >= 0)
         && (drawvert[a].x < WIDTH  || drawvert[b].x < WIDTH  || drawvert[c].x < WIDTH)
         && (drawvert[a].y >= 0     || drawvert[b].y >= 0     || drawvert[c].y >= 0)
         && (drawvert[a].y < HEIGHT || drawvert[b].y < HEIGHT || drawvert[c].y < HEIGHT))
#ifdef FORCE_FLAT
            drawTriangle(j,a,b,c,0,d!=NOTSET);
#else
            drawTriangle(j,a,b,c,0);
#endif
        if (d == NOTSET)
            continue;
        if ((drawvert[a].x >= 0     || drawvert[c].x >= 0     || drawvert[d].x >= 0)
         && (drawvert[a].x < WIDTH  || drawvert[c].x < WIDTH  || drawvert[d].x < WIDTH)
         && (drawvert[a].y >= 0     || drawvert[c].y >= 0     || drawvert[d].y >= 0)
         && (drawvert[a].y < HEIGHT || drawvert[c].y < HEIGHT || drawvert[d].y < HEIGHT))
#ifdef FORCE_FLAT
            drawTriangle(j,a,c,d,1,1);
#else
            drawTriangle(j,a,c,d,1);
#endif
    }
#else
    // Sort only visible triangles
    for (j=0,l=0;j<_curt;++j)
    {
        unsigned short a = GET(triangles,j).a;
        unsigned short b = GET(triangles,j).b;
        unsigned short c = GET(triangles,j).c;
        unsigned short d = GET(triangles,j).d;
        if (d==NOTSET
              && ((drawvert[a].x < 0       && drawvert[b].x < 0       && drawvert[c].x < 0)
               || (drawvert[a].x >= WIDTH  && drawvert[b].x >= WIDTH  && drawvert[c].x >= WIDTH)
               || (drawvert[a].y < 0       && drawvert[b].y < 0       && drawvert[c].y < 0)
               || (drawvert[a].y >= HEIGHT && drawvert[b].y >= HEIGHT && drawvert[c].y >= HEIGHT)))
            continue;
        else if (d!=NOTSET
              && ((drawvert[a].x < 0       && drawvert[b].x < 0       && drawvert[c].x < 0       && drawvert[d].x < 0)
               || (drawvert[a].x >= WIDTH  && drawvert[b].x >= WIDTH  && drawvert[c].x >= WIDTH  && drawvert[d].x >= WIDTH)
               || (drawvert[a].y < 0       && drawvert[b].y < 0       && drawvert[c].y < 0       && drawvert[d].y < 0)
               || (drawvert[a].y >= HEIGHT && drawvert[b].y >= HEIGHT && drawvert[c].y >= HEIGHT && drawvert[d].y >= HEIGHT)))
            continue;
        GET(sort_a,l++) = j;
    }
    // Sort only visible sprites
    m = _curt;
    for (j=0;j<_curs;++j,++m)
    {
        n = GET(vspr,j).z;
        if ((n<zMin) || (n>zMax))
            continue;
        GET(tmid,m).z = n;
        GET(sort_a,l++) = m;
    }
    // sort
#ifdef QSORT
    o3d_qsort(sort_a,sort_b,0,l);
#else
#ifdef QSORT2
    o3d_qsort2(l);
#else
#ifdef HEAPSORT
    o3d_heapSort(l);
#else // RADIXSORT
    o3d_radixSort(sort_a,sort_b,l);
#endif
#endif
#endif
#ifdef __GP32__
    if (clr)
        gp_waitDMA(0);
#endif
    // display triangles & sprite from back to front
    for (j=0;j<l;++j)
    {
#if defined(QSORT) || defined(QSORT2)
        int jj = GET(sort_c,j);
#else
        int jj = GET(sort_a,j);
#endif
        if (jj<_curt)
        {
#ifdef FORCE_FLAT
            drawTriangle(jj,GET(triangles,jj).a,GET(triangles,jj).b,GET(triangles,jj).c,0,GET(triangles,jj).d!=NOTSET);
            if (GET(triangles,jj).d != NOTSET)
                drawTriangle(jj,GET(triangles,jj).a,GET(triangles,jj).c,GET(triangles,jj).d,1,1);
#else
            drawTriangle(jj,GET(triangles,jj).a,GET(triangles,jj).b,GET(triangles,jj).c,0);
            if (GET(triangles,jj).d != NOTSET)
                drawTriangle(jj,GET(triangles,jj).a,GET(triangles,jj).c,GET(triangles,jj).d,1);
#endif
        }
        else
            showSprite(GET(nspr,jj-_curt),&GET(vspr,jj-_curt));
    }
#endif

#ifdef CELLSHADING
    k = 0;
    kh = HEIGHT;
    for (i=0;i<WIDTH-1;++i,++kh)
    {
        f  = GET(cs_from,k++);
        for (j=0;j<HEIGHT-1;++j)
        {
            fw = GET(cs_from,kh++);
            fh = GET(cs_from,k++);
            kl = 0;
            if ((f!=fw && (f==NO_FROM || fw==NO_FROM || (GET(cs_obj,f)!=GET(cs_obj,fw)) || (kl=nolink(f,fw)) || (kl=nocol(f,fw)))) ||
                (f!=fh && (f==NO_FROM || fh==NO_FROM || (GET(cs_obj,f)!=GET(cs_obj,fh)) || (kl=nolink(f,fh)) || (kl=nocol(f,fh)))))
            {
                if (kl==0)
                {
                    view3d[k-2]  = cscol3d_ext;
                    view3d[k-1]  = cscol3d_ext;
                    view3d[kh-1] = cscol3d_ext;
                    view3d[kh]   = cscol3d_ext;
                }
                else
                    view3d[k-2]  = cscol3d_int;
            }
            f = fh;
        }
    }
#endif

#ifdef STARS
    if (nbStars>0)
        showStars();
#endif

    clearBuffers(0);
}

#ifdef STARS

/**
 * Stars initialization
 */

void o3d_initStars(int nb,int r)
{
    int i,r2=r<<1;
    nbStars = nb;
    zStars = r;
    for (i=0;i<nb;++i)
    {
        stars[i].x = (rand()%r2)-r;
        stars[i].y = (rand()%r2)-r;
        stars[i].z = (rand()%r2)-r;
    }
}

#endif

/**
 * Matrice initialisation
 */

void initMatrix(MATRIX3D m,MATRIX_TYPE t,short v0,short v1,short v2)
{
    int i;
    memset(m,0,64);
    switch(t)
    {
        case VAL:
            for (i=0;i<16;m[i++]=v0);
            break;
        case ROTX:
            m[0]  =  MULT;
            m[5]  =  tcos[v0];
            m[6]  =  tsin[v0];
            m[9]  = -tsin[v0];
            m[10] =  tcos[v0];
            m[15] =  MULT;
            break;
        case ROTY:
            m[0]  =  tcos[v0];
            m[2]  = -tsin[v0];
            m[5]  =  MULT;
            m[8]  =  tsin[v0];
            m[10] =  tcos[v0];
            m[15] =  MULT;
            break;
        case ROTZ:
            m[0]  =  tcos[v0];
            m[1]  =  tsin[v0];
            m[4]  = -tsin[v0];
            m[5]  =  tcos[v0];
            m[10] =  MULT;
            m[15] =  MULT;
            break;
        case ROTXYZ:
            {
                int cx = tcos[v0],cy = tcos[v1],cz = tcos[v2],
                    sx = tsin[v0],sy = tsin[v1],sz = tsin[v2];
                m[0]  = (cy*cz)>>SHIFT;
                m[1]  = (cy*sz)>>SHIFT;
                m[2]  = -sy;
                m[4]  = (((sx*sy*cz)>>SHIFT)-(cx*sz))>>SHIFT;
                m[5]  = (((sx*sy*sz)>>SHIFT)+(cx*cz))>>SHIFT;
                m[6]  = (sx*cy)>>SHIFT;
                m[8]  = (((cx*sy*cz)>>SHIFT)+(sx*sz))>>SHIFT;
                m[9]  = (((cx*sy*sz)>>SHIFT)-(sx*cz))>>SHIFT;
                m[10] = (cx*cy)>>SHIFT;
                m[15] = MULT;
            }
            break;
        case SCALE:
            m[0]  = v0;
            m[5]  = v1;
            m[10] = v2;
            m[15] = MULT;
            break;
        case TRANS:
            m[12] = v0<<SHIFT;
            m[13] = v1<<SHIFT;
            m[14] = v2<<SHIFT;
            m[15] = MULT;
            break;
        case IDENT:
            m[0]  = MULT;
            m[5]  = MULT;
            m[10] = MULT;
            m[15] = MULT;
            break;
        case PROJ:
            m[0]  = MULT;
            m[5]  = MULT;
            m[10] = MULT;
            m[12] = v0<<SHIFT;
            m[15] = MULT;
            break;
    }
}

/**
 * Rotation and Trans Matrix initialisation
 */

void initRotTransMatrix(register MATRIX3D m,ROTATION3D *rot,VECTOR3D *pos)
{
    int cx = tcos[rot->rx],cy = tcos[rot->ry],cz = tcos[rot->rz],
        sx = tsin[rot->rx],sy = tsin[rot->ry],sz = tsin[rot->rz];
    m[0]  = (cy*cz)>>SHIFT;
    m[1]  = (cy*sz)>>SHIFT;
    m[2]  = -sy;
    m[3]  = 0;
    m[4]  = (((sx*sy*cz)>>SHIFT)-(cx*sz))>>SHIFT;
    m[5]  = (((sx*sy*sz)>>SHIFT)+(cx*cz))>>SHIFT;
    m[6]  = (sx*cy)>>SHIFT;
    m[7]  = 0;
    m[8]  = (((cx*sy*cz)>>SHIFT)+(sx*sz))>>SHIFT;
    m[9]  = (((cx*sy*sz)>>SHIFT)-(sx*cz))>>SHIFT;
    m[10] = (cx*cy)>>SHIFT;
    m[11] = 0;
    m[12] = pos->x<<SHIFT;
    m[13] = pos->y<<SHIFT;
    m[14] = pos->z<<SHIFT;
    m[15] = MULT;
}

/**
 * Matrice multiplication r = a*b (with a!=r && b!=r)
 */

void multiplyMatrix(register MATRIX3D a,register MATRIX3D b,register MATRIX3D r)
{
    r[0] = (a[0]*b[0] + a[4]*b[1] + a[8]*b[2]  + a[12]*b[3])>>SHIFT;
    r[1] = (a[1]*b[0] + a[5]*b[1] + a[9]*b[2]  + a[13]*b[3])>>SHIFT;
    r[2] = (a[2]*b[0] + a[6]*b[1] + a[10]*b[2] + a[14]*b[3])>>SHIFT;
    r[3] = (a[3]*b[0] + a[7]*b[1] + a[11]*b[2] + a[15]*b[3])>>SHIFT;

    r[4] = (a[0]*b[4] + a[4]*b[5] + a[8]*b[6]  + a[12]*b[7])>>SHIFT;
    r[5] = (a[1]*b[4] + a[5]*b[5] + a[9]*b[6]  + a[13]*b[7])>>SHIFT;
    r[6] = (a[2]*b[4] + a[6]*b[5] + a[10]*b[6] + a[14]*b[7])>>SHIFT;
    r[7] = (a[3]*b[4] + a[7]*b[5] + a[11]*b[6] + a[15]*b[7])>>SHIFT;

    r[8]  = (a[0]*b[8] + a[4]*b[9] + a[8]*b[10]  + a[12]*b[11])>>SHIFT;
    r[9]  = (a[1]*b[8] + a[5]*b[9] + a[9]*b[10]  + a[13]*b[11])>>SHIFT;
    r[10] = (a[2]*b[8] + a[6]*b[9] + a[10]*b[10] + a[14]*b[11])>>SHIFT;
    r[11] = (a[3]*b[8] + a[7]*b[9] + a[11]*b[10] + a[15]*b[11])>>SHIFT;

    r[12] = (a[0]*b[12] + a[4]*b[13] + a[8]*b[14]  + a[12]*b[15])>>SHIFT;
    r[13] = (a[1]*b[12] + a[5]*b[13] + a[9]*b[14]  + a[13]*b[15])>>SHIFT;
    r[14] = (a[2]*b[12] + a[6]*b[13] + a[10]*b[14] + a[14]*b[15])>>SHIFT;
    r[15] = (a[3]*b[12] + a[7]*b[13] + a[11]*b[14] + a[15]*b[15])>>SHIFT;
}

/**
 * Short Vertex multiplication r = m*v
 */

void multiplySVector(MATRIX3D m,const SVECTOR3D *v,VECTOR3D *r)
{
    r->x = (v->x * m[0] + v->y * m[4] + v->z *m[8]  + m[12])>>SHIFT;
    r->y = (v->x * m[1] + v->y * m[5] + v->z *m[9]  + m[13])>>SHIFT;
    r->z = (v->x * m[2] + v->y * m[6] + v->z *m[10] + m[14])>>SHIFT;
}

/**
 * Vertex multiplication r = m*v (with v!=r)
 */

void multiplyVector(MATRIX3D m,VECTOR3D *v,VECTOR3D *r)
{
    r->x = (v->x * m[0] + v->y * m[4] + v->z *m[8]  + m[12])>>SHIFT;
    r->y = (v->x * m[1] + v->y * m[5] + v->z *m[9]  + m[13])>>SHIFT;
    r->z = (v->x * m[2] + v->y * m[6] + v->z *m[10] + m[14])>>SHIFT;
}

/**
 * Length of vector
 */

int length(VECTOR3D *v)
{
    return isqrt((v->x*v->x)+(v->y*v->y)+(v->z*v->z));
}

/**
 * Length of vector to 1 (MULT)
 */

void normalize(VECTOR3D *v)
{
    int l = length(v);
    if (l>0)
    {
        if (l<=MAX_DIV)
        {
            v->x = DIVIDE_NSP(v->x,l)>>(DIV_SHIFT-SHIFT);
            v->y = DIVIDE_NSP(v->y,l)>>(DIV_SHIFT-SHIFT);
            v->z = DIVIDE_NSP(v->z,l)>>(DIV_SHIFT-SHIFT);
        }
        else
        {
            v->x = (v->x<<SHIFT)/l;
            v->y = (v->y<<SHIFT)/l;
            v->z = (v->z<<SHIFT)/l;
        }
    }
    else
    {
        v->x = v->y = 0;
        v->z = 1<<SHIFT;
    }
}

/**
 * Compute all 3D scene transformations matrices + triangle normals
 */

_MATRIX3D(tempMat);

void transformNode(NODE3D *node,MATRIX3D m0)
{
    unsigned short i;
    MATRIX3D m = m0==NULL ? node->m : tempMat;

    if (node->hide)
    {
        node->vis = 0;
        return;
    }

    initRotTransMatrix(m,&node->rot,&node->pos);
    if (m0!=NULL)
        multiplyMatrix(m0,m,node->m);

    node->vis = 1;

    if (node->type==SCENE)
    {
        SCENE3D *s=node->data.s;
        for (i=0;i<s->count;++i)
            transformNode(s->nodes[i],node->m);
    }
    else if (node->type==OBJECT || node->type==ANIMATION)
    {
        OBJECT3D* o = node->type==OBJECT ? node->data.o : node->data.a->anims[node->cur_frame];
        VECTOR3D  v;

        int x,y,mask=0;
        for (i=0;(i<8) && (mask!=15);++i)
        {
            MULTIPLY_Z(node->m,o->bbox[i],v);
            if (v.z==proj3d)
                continue;
            v.z = proj3d-v.z;
            MULTIPLY_X(node->m,o->bbox[i],v);
            v.x *= proj3d;
            x = WIDTH2-DIVIDE(v.x,v.z);
            MULTIPLY_Y(node->m,o->bbox[i],v);
            v.y *= proj3d;
            y = HEIGHT2+DIVIDE(v.y,v.z);
            if (x>=0 && x<WIDTH && y>=0 && y<HEIGHT)
                break;
            mask |= (x>=0 ? 1 : 0) | (x<WIDTH ? 2 : 0) | (y>=0 ? 4 : 0) | (y<HEIGHT ? 8 : 0);
        }
        if (i==8 && (mask!=15))
        {
            node->vis = 0;
            return;
        }
        __vis__+=1;
    }
}

/**
 * Compute light
 */

#ifdef POINT_LIGHT
inline unsigned int computeLight(VERTEX3D *v,VECTOR3D *n)
{
    VECTOR3D vlight;
    vlight.x = tlight3d.x-v->x;
    vlight.y = tlight3d.y-v->y;
    vlight.z = tlight3d.z-v->z;
    normalize(&vlight);
    vlight.x += n->x;
    vlight.y += n->y;
    vlight.z += n->z;
    return length(&vlight);
}
#else
__inline unsigned int computeLight(VECTOR3D *n)
{
    return MULT+((tlight3d.x*n->x+tlight3d.y*n->y+tlight3d.z*n->z)>>SHIFT);
}
#endif

/**
 * Transform all visible vertices
 */

void transformVertices(NODE3D *node)
{
    int i;

    if (node->hide || !node->vis)
        return;

    if (node->type==SCENE)
    {
        SCENE3D *s=node->data.s;
        for (i=0;i<s->count;++i)
            transformVertices(s->nodes[i]);
    }
    else if (node->type==SPRITE)
    {
#ifdef ZBUFFER
        VECTOR3D v;
        MULTIPLY(node->m,zero,v);
        showSprite(node,&v);
#else
        MULTIPLY(node->m,zero,GET(vspr,_curs));
        GET(nspr,_curs++) = node;
#endif
    }
    else // Object or animation
    {
        OBJECT3D   *o  = node->type==OBJECT ? node->data.o : node->data.a->anims[node->cur_frame];
        SVERTEX3D  *v  = o->vert;
        TRIANGLE3D *t  = o->tri;
        SVECTOR3D  *tn = o->tnorm;
        SVECTOR3D  *m  = o->tmid;
        VECTOR3D    vz;
        VECTOR3D    vcam;
        MATRIX3D    nm = node->m;
        MULTIPLY(nm,zero,vz);

        for (i=0;i<o->nbt;++i)
        {
            MULTIPLY_P(nm,tn[i],tnorms[_curt],vz);
            MULTIPLY(nm,m[i],tmid[_curt]);
            vcam.x = -tmid[_curt].x;
            vcam.y = -tmid[_curt].y;
            vcam.z = proj3d - tmid[_curt].z;
            if (DOT_PRODUCT(vcam,tnorms[_curt])>0 && tmid[_curt].z>=zMin && tmid[_curt].z<=zMax)
            {
                GET(vis,GET(triangles,_curt).a  = t[i].a+_curv) = 1;
                GET(vis,GET(triangles,_curt).b  = t[i].b+_curv) = 1;
                GET(vis,GET(triangles,_curt).c  = t[i].c+_curv) = 1;
                if (t[i].d!=NOTSET)
                    GET(vis,GET(triangles,_curt).d  = t[i].d+_curv) = 1;
                else
                    GET(triangles,_curt).d = NOTSET;
#ifdef OVERLOADS
                GET(overloads,_curt) = node->overload;
#endif
#ifdef CELLSHADING
                GET(cs_obj,_curt) = node->id;
#endif
                GET(triangles,_curt++).orig = &t[i];
            }
        }
        if (shading3d==GOURAUD)
        {
            SVECTOR3D  *vn = o->vnorm;
            for (i=0;i<o->nbv;++i,++_curv)
            {
                if (GET(vis,_curv))
                {
                    MULTIPLY(nm,v[i],GET(vertices,_curv));
                    GET(drawvert,_curv).x = WIDTH2  - DIVIDE(GET(vertices,_curv).x*proj3d,proj3d-GET(vertices,_curv).z);
                    GET(drawvert,_curv).y = HEIGHT2 + DIVIDE(GET(vertices,_curv).y*proj3d,proj3d-GET(vertices,_curv).z);
                    // compute vertice light
                    MULTIPLY_P(nm,vn[i],GET(vnorms,_curv),vz);
#ifdef POINT_LIGHT
                    GET(vlights,_curv) = computeLight(vertices+_curv,vnorms+_curv);
#else
                    GET(vlights,_curv) = computeLight(vnorms+_curv);
#endif
                }
            }
        }
        else
        {
            int vx,vy,vz;
            for (i=0;i<o->nbv;++i,++_curv)
            {
                if (GET(vis,_curv))
                {
                    vx = proj3d*((v[i].x * nm[0] + v[i].y * nm[4] + v[i].z * nm[8] + nm[12])>>SHIFT);
                    vy = proj3d*((v[i].x * nm[1] + v[i].y * nm[5] + v[i].z * nm[9]  + nm[13])>>SHIFT);
                    vz = proj3d-((v[i].x * nm[2] + v[i].y * nm[6] + v[i].z * nm[10] + nm[14])>>SHIFT);
                    GET(drawvert,_curv).x = WIDTH2  - DIVIDE(vx,vz);
                    GET(drawvert,_curv).y = HEIGHT2 + DIVIDE(vy,vz);
                }
            }
        }
    }
}

/**
 * Show sprite
 */

void showSprite(NODE3D *node,VECTOR3D *v)
{
    SPRITE3D *spr = node->data.p;
    TEXTURE3D *t = spr->t;
    int i,j,k,l,x,y,w,h,fx,fy,lx,ly,dx,tw,dk,dl,fl,curk;
    int w2 = spr->w>>1;
    int h2 = spr->h>>1;
    register unsigned int trans = t->trans;
    register unsigned int pix;
#ifdef BITS_16
    unsigned short *pal = t->gpal;
#endif
    unsigned char *tex = t->pix;

    j = WIDTH2-DIVIDE((v->x-w2)*proj3d,proj3d-v->z);
    k = WIDTH2-DIVIDE((v->x+w2)*proj3d,proj3d-v->z);
    if (j==k)
        return;
    x = k;
    w = j-k;

    j = HEIGHT2+DIVIDE((v->y-h2)*proj3d,proj3d-v->z);
    k = HEIGHT2+DIVIDE((v->y+h2)*proj3d,proj3d-v->z);
    if (j>=k)
        return;
    y = j;
    h = k-j-1;

    if ((x>=WIDTH)||(x+w<0)||(y>=HEIGHT)||(y+h<0))
        return;

    fx = MAX(x,0);
    lx = x+w+1;
    lx = MIN(lx,WIDTH);
    fy = MAX(y,0);
    ly = y+h+1;
    ly = MIN(ly,HEIGHT);
    dx = fx*HEIGHT;
    tw = t->w+1;
    dk = DIVIDE_NSP(spr->w,w);
    k = (fx-x)*dk;
    dl = DIVIDE_NSP(spr->h,h);
    fl = (h-(fy-y))*dl+(spr->dy<<DIV_SHIFT);

    for (i=fx;i<lx;++i)
    {
        l = fl;
        curk = (k>>DIV_SHIFT)+spr->dx;
#ifdef SPEED
        for (j=fy;j<ly;++j,l-=dl)
        {
            pix = tex[(l>>DIV_SHIFT)*tw+curk];
            if (trans!=pix)
                view3d[dx+j] = pal[pix];
        }
#else
        for (j=fy;j<ly;++j)
        {
#ifdef ZBUFFER
            if (GET(zbuf,dx+j)<v->z)
#endif
            {
                pix = tex[(l>>DIV_SHIFT)*tw+curk];
                if (trans!=pix)
                {
#ifdef BITS_16
                    view3d[dx+j] = pal[pix];
#else
                    view3d[dx+j] = (unsigned char)pix;
#endif
#ifdef ZBUFFER
                    GET(zbuf,dx+j) = v->z;
#endif
                }
            }
            l-=dl;
        }
#endif // SPEED
        k+=dk;
        dx+=HEIGHT;
    }
}

/**
 * Draw a "flat light" triangle
 */

void drawFlatTriangle(unsigned short a,unsigned short b,unsigned short c,unsigned short nc,unsigned short nt)
{
    int dax = GET(drawvert,a).x;
    int dbx = GET(drawvert,b).x;
    int dcx = GET(drawvert,c).x;
    int ax = MAX(dax,0);
    int bx = MIN(dbx,WIDTH);
    int cx = MIN(dcx,WIDTH-1);
    int dx;
    int ymin,ymax;
    struct
    {
        int xdy,xcy;
#ifdef ZBUFFER
        int xdz,xcz;
#endif
    } s0,s1,s2;
#ifdef ZBUFFER
    int cz,dz,curz;
#endif
    int cy0,cy1;

    if (bx>0 && ax<bx)
    {
        dx = ax*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,dcx-dax);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-dax)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,b).y-GET(drawvert,a).y,dbx-dax);
        s1.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-dax)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,dcx-dax);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-dax)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,b).z-GET(vertices,a).z,dbx-dax);
        s1.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-dax)*s1.xdz;
#endif

        if (s0.xdy > s1.xdy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
#ifdef SPEED
            for (;ymin<ymax;++ymin)
                view3d[dx+ymin] = nc;
#else
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
                        view3d[dx+ymin] = nc;
#ifdef CELLSHADING
                        GET(cs_from,dx+ymin) = nt;
#endif
#ifdef ZBUFFER
                        GET(zbuf,dx+ymin) = cz;
                    }
                    curz+=dz;
#endif
                }
            }
#endif // SPEED
            if (++ax == bx)
                break;
            dx+=HEIGHT;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
        }
    }

    if (bx<cx)
    {
        bx = MAX(bx,0);
        dx = bx*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,dcx-dax);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(bx-dax)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,b).y,dcx-dbx);
        s1.xcy = (GET(drawvert,b).y<<DIV_SHIFT)+(bx-dbx)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,dcx-dax);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(bx-dax)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,b).z,dcx-dbx);
        s1.xcz = (GET(vertices,b).z<<DIV_SHIFT)+(bx-dbx)*s1.xdz;
#endif

        if (s0.xcy > s1.xcy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
#ifdef SPEED
            for (;ymin<ymax;++ymin)
                view3d[dx+ymin] = nc;
#else
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
                        view3d[dx+ymin] = nc;
#ifdef CELLSHADING
                        GET(cs_from,dx+ymin) = nt;
#endif
#ifdef ZBUFFER
                        GET(zbuf,dx+ymin) = cz;
                    }
                    curz+=dz;
#endif
                }
            }
#endif // SPEED
            if (++bx > cx)
                break;
            dx+=HEIGHT;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
        }
    }
}

/**
 * Draw a "gouraud lighted" triangle
 */
void drawGouraudTriangle(unsigned short a,unsigned short b,unsigned short c,unsigned short *col,unsigned short nt)
{
    int dax = GET(drawvert,a).x;
    int dbx = GET(drawvert,b).x;
    int dcx = GET(drawvert,c).x;
    int ax = MAX(dax,0);
    int bx = MIN(dbx,WIDTH);
    int cx = MIN(dcx,WIDTH-1);
    int dx;
    int ymin,ymax;
    struct
    {
        int xdy,xcy;
#ifdef ZBUFFER
        int xdz,xcz;
#endif
        int xl,xdl;
    } s0,s1,s2;
#ifdef ZBUFFER
    int cz,dz,curz;
#endif
    int curl,dl;
    int dab,dac,dbc;
    int cy0,cy1;

    dac = DIVIDE_NS(GET(vlights,c)-GET(vlights,a),dcx-dax);

    if (bx>0 && ax<bx)
    {
        dx = ax*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,dcx-dax);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-dax)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,b).y-GET(drawvert,a).y,dbx-dax);
        s1.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-dax)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,dcx-dax);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-dax)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,b).z-GET(vertices,a).z,dbx-dax);
        s1.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-dax)*s1.xdz;
#endif

        dab = DIVIDE_NS(GET(vlights,b)-GET(vlights,a),dbx-dax);
        s0.xl = (GET(vlights,a)<<DIV_SHIFT)+(ax-dax)*dac;
        s1.xl = (GET(vlights,a)<<DIV_SHIFT)+(ax-dax)*dab;

        s0.xdl = dac;
        s1.xdl = dab;

        if (s0.xdy > s1.xdy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                dl = DIVIDE_NSP((s1.xl-s0.xl)>>DIV_SHIFT,cy1-cy0);
                curl = s0.xl+dl*(ymin-cy0);
#ifdef SPEED
                for (;ymin<ymax;++ymin,curl+=dl)
                    view3d[dx+ymin] = col[GPLIGHT(curl)];
#else
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
                        view3d[dx+ymin] = col[GPLIGHT(curl)];
#ifdef CELLSHADING
                        GET(cs_from,dx+ymin) = nt;
#endif
#ifdef ZBUFFER
                        GET(zbuf,dx+ymin) = cz;
                    }
                    curz+=dz;
#endif
                    curl+=dl;
                }
#endif// SPEED
            }
            if (++ax == bx)
                break;
            dx+=HEIGHT;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
            s0.xl+=s0.xdl;
            s1.xl+=s1.xdl;
        }
    }

    if (bx<cx)
    {
        bx = MAX(bx,0);
        dx = bx*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,dcx-dax);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(bx-dax)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,b).y,dcx-dbx);
        s1.xcy = (GET(drawvert,b).y<<DIV_SHIFT)+(bx-dbx)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,dcx-dax);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(bx-dax)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,b).z,dcx-dbx);
        s1.xcz = (GET(vertices,b).z<<DIV_SHIFT)+(bx-dbx)*s1.xdz;
#endif
        dbc = DIVIDE_NS(GET(vlights,c)-GET(vlights,b),dcx-dbx);
        s0.xl = (GET(vlights,a)<<DIV_SHIFT)+(bx-dax)*dac;
        s1.xl = (GET(vlights,b)<<DIV_SHIFT)+(bx-dbx)*dbc;

        s0.xdl = dac;
        s1.xdl = dbc;

        if (s0.xcy > s1.xcy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                dl = DIVIDE_NSP((s1.xl-s0.xl)>>DIV_SHIFT,cy1-cy0);
                curl = s0.xl+dl*(ymin-cy0);
#ifdef SPEED
                for (;ymin<ymax;++ymin,curl+=dl)
                    view3d[dx+ymin] = col[GPLIGHT(curl)];
#else
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
                        view3d[dx+ymin] = col[GPLIGHT(curl)];
#ifdef CELLSHADING
                        GET(cs_from,dx+ymin) = nt;
#endif
#ifdef ZBUFFER
                        GET(zbuf,dx+ymin) = cz;
                    }
                    curz+=dz;
#endif
                    curl+=dl;
                }
#endif // SPEED
            }
            if (++bx > cx)
                break;
            dx+=HEIGHT;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
            s0.xl+=s0.xdl;
            s1.xl+=s1.xdl;
        }
    }
}

/**
 * Draw a "flat light" textured triangle
 */

void drawFlatTexturedTriangle(unsigned short a,unsigned short b,unsigned short c,
                              unsigned char oa,unsigned char ob,unsigned char oc,
                              int l,TEXTURE3D_DATA* td,TEXTURE3D *tx,unsigned short nt)
{
#ifndef GHOST
    TEXTURE3D *t = tx==NULL ? td->t : tx;
#endif
    int ax = MAX(GET(drawvert,a).x,0);
    int bx = MIN(GET(drawvert,b).x,WIDTH);
    int cx = MIN(GET(drawvert,c).x,WIDTH-1);
    int dx;
#ifndef GHOST
    int dw=t->w+1;
#ifndef SPEED
    int pix;
#endif
#endif

    int ctx,cty,cdtx,cdty;

    int ymin,ymax;
    struct
    {
        int xdy,xcy;
        int tx,ty;
        int dtx,dty;
#ifdef ZBUFFER
        int xdz,xcz;
#endif
    } s0,s1,s2;
#ifdef ZBUFFER
    int cz,dz,curz;
#endif
    int cy0,cy1;

#ifndef GHOST
#ifdef USE_TRANS
    int trans=t->trans;
#endif
#ifdef BITS_16
    unsigned short *pal = t->lpal==NULL ? (t->gpal) : (t->lpal+((t->nbc+1)*(l>31?31:l)));
#endif
    unsigned char *tex = t->pix;
#endif

    if (bx>0 && ax<bx)
    {
        dx = ax*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,b).y-GET(drawvert,a).y,GET(drawvert,b).x-GET(drawvert,a).x);
        s1.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,b).z-GET(vertices,a).z,GET(drawvert,b).x-GET(drawvert,a).x);
        s1.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s1.xdz;
#endif

        s0.tx = td->x[oa]<<DIV_SHIFT;
        s0.ty = td->y[oa]<<DIV_SHIFT;
        s1.tx = s0.tx;
        s1.ty = s0.ty;

        s0.dtx = DIVIDE_NSP(td->x[oc]-td->x[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s0.dty = DIVIDE_NSP(td->y[oc]-td->y[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s1.dtx = DIVIDE_NSP(td->x[ob]-td->x[oa],GET(drawvert,b).x-GET(drawvert,a).x);
        s1.dty = DIVIDE_NSP(td->y[ob]-td->y[oa],GET(drawvert,b).x-GET(drawvert,a).x);

        s0.tx+=s0.dtx*(ax-GET(drawvert,a).x);
        s0.ty+=s0.dty*(ax-GET(drawvert,a).x);
        s1.tx+=s1.dtx*(ax-GET(drawvert,a).x);
        s1.ty+=s1.dty*(ax-GET(drawvert,a).x);

        if (s0.xdy > s1.xdy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                cdtx = DIVIDE_NSP((s1.tx-s0.tx)>>DIV_SHIFT,cy1-cy0);
                cdty = DIVIDE_NSP((s1.ty-s0.ty)>>DIV_SHIFT,cy1-cy0);
                ctx = s0.tx+(ymin-cy0)*cdtx;
                cty = s0.ty+(ymin-cy0)*cdty;
#ifdef SPEED
                for (;ymin<ymax;++ymin,ctx+=cdtx,cty+=cdty)
                    view3d[dx+ymin] = pal[tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw]];
#else
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
#ifndef GHOST
                        pix = tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw];
#ifdef USE_TRANS
                        if (pix!=trans)
                        {
#endif
#ifdef BITS_16
                            view3d[dx+ymin] = pal[pix];
#else
                            view3d[dx+ymin] = (unsigned char)pix;
#endif
#endif
#ifdef ZBUFFER
                            GET(zbuf,dx+ymin) = cz;
#endif
#ifdef CELLSHADING
                            GET(cs_from,dx+ymin) = nt;
#endif
#ifndef GHOST
#ifdef USE_TRANS
                        }
#endif
#endif
#ifdef ZBUFFER
                    }
                    curz+=dz;
#endif
                    ctx+=cdtx;
                    cty+=cdty;
                }
#endif // SPEED
            }
            if (++ax == bx)
                break;
            dx+=HEIGHT;
            s0.tx+=s0.dtx;
            s0.ty+=s0.dty;
            s1.tx+=s1.dtx;
            s1.ty+=s1.dty;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
        }
    }

    if (bx<cx)
    {
        bx = MAX(bx,0);
        dx = bx*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(bx-GET(drawvert,a).x)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,b).y,GET(drawvert,c).x-GET(drawvert,b).x);
        s1.xcy = (GET(drawvert,b).y<<DIV_SHIFT)+(bx-GET(drawvert,b).x)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(bx-GET(drawvert,a).x)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,b).z,GET(drawvert,c).x-GET(drawvert,b).x);
        s1.xcz = (GET(vertices,b).z<<DIV_SHIFT)+(bx-GET(drawvert,b).x)*s1.xdz;
#endif

        s0.tx = td->x[oa]<<DIV_SHIFT;
        s0.ty = td->y[oa]<<DIV_SHIFT;
        s1.tx = td->x[ob]<<DIV_SHIFT;
        s1.ty = td->y[ob]<<DIV_SHIFT;

        s0.dtx = DIVIDE_NSP(td->x[oc]-td->x[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s0.dty = DIVIDE_NSP(td->y[oc]-td->y[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s1.dtx = DIVIDE_NSP(td->x[oc]-td->x[ob],GET(drawvert,c).x-GET(drawvert,b).x);
        s1.dty = DIVIDE_NSP(td->y[oc]-td->y[ob],GET(drawvert,c).x-GET(drawvert,b).x);

        s0.tx+=s0.dtx*(bx-GET(drawvert,a).x);
        s0.ty+=s0.dty*(bx-GET(drawvert,a).x);
        s1.tx+=s1.dtx*(bx-GET(drawvert,b).x);
        s1.ty+=s1.dty*(bx-GET(drawvert,b).x);

        if (s0.xcy > s1.xcy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                cdtx = DIVIDE_NSP((s1.tx-s0.tx)>>DIV_SHIFT,cy1-cy0);
                cdty = DIVIDE_NSP((s1.ty-s0.ty)>>DIV_SHIFT,cy1-cy0);
                ctx = s0.tx+(ymin-cy0)*cdtx;
                cty = s0.ty+(ymin-cy0)*cdty;
#ifdef SPEED
                for (;ymin<ymax;++ymin,ctx+=cdtx,cty+=cdty)
                    view3d[dx+ymin] = pal[tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw]];
#else
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
#ifndef GHOST
                        pix = tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw];
#ifdef USE_TRANS
                        if (pix!=trans)
                        {
#endif
#ifdef BITS_16
                            view3d[dx+ymin] = pal[pix];
#else
                            view3d[dx+ymin] = (unsigned char)pix;
#endif
#endif
#ifdef ZBUFFER
                            GET(zbuf,dx+ymin) = cz;
#endif
#ifdef CELLSHADING
                            GET(cs_from,dx+ymin) = nt;
#endif
#ifndef GHOST
#ifdef USE_TRANS
                        }
#endif
#endif
#ifdef ZBUFFER
                    }
                    curz+=dz;
#endif
                    ctx+=cdtx;
                    cty+=cdty;
                }
#endif
            }
            if (++bx > cx)
                break;
            dx+=HEIGHT;
            s0.tx+=s0.dtx;
            s0.ty+=s0.dty;
            s1.tx+=s1.dtx;
            s1.ty+=s1.dty;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
        }
    }
}

/**
 * Draw a "gouraud light" textured triangle
 */

void drawGouraudTexturedTriangle(unsigned short a,unsigned short b,unsigned short c,
                                 unsigned char oa,unsigned char ob,unsigned char oc,
                                 TEXTURE3D_DATA* td,TEXTURE3D *t,unsigned short nt)
{
    int ax = MAX(GET(drawvert,a).x,0);
    int bx = MIN(GET(drawvert,b).x,WIDTH);
    int cx = MIN(GET(drawvert,c).x,WIDTH-1);
    int dx, dw = t->w+1,nbc = t->nbc+1;
#ifdef USE_TRANS
    int trans=t->trans;
#endif
#ifndef SPEED
    int pix;
#endif
    unsigned char *tex = t->pix;

    struct
    {
        int tx,ty;
        int dtx,dty;
        int xdy,xcy;
#ifdef ZBUFFER
        int xdz,xcz;
#endif
        int xl, xdl;
    } s0,s1,s2;

    int ctx,cty,cdtx,cdty;
    int ymin,ymax;
    int curl,dl;
#ifdef ZBUFFER
    int cz,dz,curz;
#endif
    int dab,dac,dbc;
    int cy0,cy1;

    dac = DIVIDE_NS(GET(vlights,c)-GET(vlights,a),GET(drawvert,c).x-GET(drawvert,a).x);

    if (bx>0 && ax<bx)
    {
        dx = ax*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,b).y-GET(drawvert,a).y,GET(drawvert,b).x-GET(drawvert,a).x);
        s1.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,b).z-GET(vertices,a).z,GET(drawvert,b).x-GET(drawvert,a).x);
        s1.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*s1.xdz;
#endif

        s0.tx = td->x[oa]<<DIV_SHIFT;
        s0.ty = td->y[oa]<<DIV_SHIFT;
        s1.tx = s0.tx;
        s1.ty = s0.ty;

        s0.dtx = DIVIDE_NSP(td->x[oc]-td->x[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s0.dty = DIVIDE_NSP(td->y[oc]-td->y[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s1.dtx = DIVIDE_NSP(td->x[ob]-td->x[oa],GET(drawvert,b).x-GET(drawvert,a).x);
        s1.dty = DIVIDE_NSP(td->y[ob]-td->y[oa],GET(drawvert,b).x-GET(drawvert,a).x);

        s0.tx+=s0.dtx*(ax-GET(drawvert,a).x);
        s0.ty+=s0.dty*(ax-GET(drawvert,a).x);
        s1.tx+=s1.dtx*(ax-GET(drawvert,a).x);
        s1.ty+=s1.dty*(ax-GET(drawvert,a).x);

        dab = DIVIDE_NS(GET(vlights,b)-GET(vlights,a),GET(drawvert,b).x-GET(drawvert,a).x);
        s0.xl = (GET(vlights,a)<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*dac;
        s1.xl = (GET(vlights,a)<<DIV_SHIFT)+(ax-GET(drawvert,a).x)*dab;

        s0.xdl = dac;
        s1.xdl = dab;

        if (s0.xdy > s1.xdy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                cdtx = DIVIDE_NSP((s1.tx-s0.tx)>>DIV_SHIFT,cy1-cy0);
                cdty = DIVIDE_NSP((s1.ty-s0.ty)>>DIV_SHIFT,cy1-cy0);
                ctx = s0.tx+(ymin-cy0)*cdtx;
                cty = s0.ty+(ymin-cy0)*cdty;

                dl = DIVIDE_NSP((s1.xl-s0.xl)>>DIV_SHIFT,cy1-cy0);
                curl = s0.xl+dl*(ymin-cy0);

#ifdef SPEED
                for (;ymin<ymax;++ymin)
                {
                    view3d[dx+ymin] = t->lpal[GPLIGHT(curl)*nbc+tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw]];
                    ctx+=cdtx;
                    cty+=cdty;
                    curl+=dl;
                }
#else
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
                        pix = tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw];
#ifdef USE_TRANS
                        if (pix!=trans)
                        {
#endif
                            view3d[dx+ymin] = t->lpal[GPLIGHT(curl)*nbc+pix];
#ifdef ZBUFFER
                            GET(zbuf,dx+ymin) = cz;
#endif
#ifdef CELLSHADING
                            GET(cs_from,dx+ymin) = nt;
#endif
#ifdef USE_TRANS
                        }
#endif
#ifdef ZBUFFER
                    }
                    curz+=dz;
#endif
                    curl+=dl;
                    ctx+=cdtx;
                    cty+=cdty;
                }
#endif // SPEED
            }
            if (++ax == bx)
                break;
            dx+=HEIGHT;
            s0.tx+=s0.dtx;
            s0.ty+=s0.dty;
            s1.tx+=s1.dtx;
            s1.ty+=s1.dty;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
            s0.xl+=s0.xdl;
            s1.xl+=s1.xdl;
        }
    }

    if (bx<cx)
    {
        bx = MAX(bx,0);
        dx = bx*HEIGHT;

        s0.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,a).y,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcy = (GET(drawvert,a).y<<DIV_SHIFT)+(bx-GET(drawvert,a).x)*s0.xdy;
        s1.xdy = DIVIDE_NSP(GET(drawvert,c).y-GET(drawvert,b).y,GET(drawvert,c).x-GET(drawvert,b).x);
        s1.xcy = (GET(drawvert,b).y<<DIV_SHIFT)+(bx-GET(drawvert,b).x)*s1.xdy;

#ifdef ZBUFFER
        s0.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,a).z,GET(drawvert,c).x-GET(drawvert,a).x);
        s0.xcz = (GET(vertices,a).z<<DIV_SHIFT)+(bx-GET(drawvert,a).x)*s0.xdz;
        s1.xdz = DIVIDE_NSP(GET(vertices,c).z-GET(vertices,b).z,GET(drawvert,c).x-GET(drawvert,b).x);
        s1.xcz = (GET(vertices,b).z<<DIV_SHIFT)+(bx-GET(drawvert,b).x)*s1.xdz;
#endif

        s0.tx = td->x[oa]<<DIV_SHIFT;
        s0.ty = td->y[oa]<<DIV_SHIFT;
        s1.tx = td->x[ob]<<DIV_SHIFT;
        s1.ty = td->y[ob]<<DIV_SHIFT;

        s0.dtx = DIVIDE_NSP(td->x[oc]-td->x[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s0.dty = DIVIDE_NSP(td->y[oc]-td->y[oa],GET(drawvert,c).x-GET(drawvert,a).x);
        s1.dtx = DIVIDE_NSP(td->x[oc]-td->x[ob],GET(drawvert,c).x-GET(drawvert,b).x);
        s1.dty = DIVIDE_NSP(td->y[oc]-td->y[ob],GET(drawvert,c).x-GET(drawvert,b).x);

        s0.tx+=s0.dtx*(bx-GET(drawvert,a).x);
        s0.ty+=s0.dty*(bx-GET(drawvert,a).x);
        s1.tx+=s1.dtx*(bx-GET(drawvert,b).x);
        s1.ty+=s1.dty*(bx-GET(drawvert,b).x);

        dbc = DIVIDE_NS(GET(vlights,c)-GET(vlights,b),GET(drawvert,c).x-GET(drawvert,b).x);
        s0.xl = (GET(vlights,a)<<DIV_SHIFT)+(bx-GET(drawvert,a).x)*dac;
        s1.xl = (GET(vlights,b)<<DIV_SHIFT)+(bx-GET(drawvert,b).x)*dbc;

        s0.xdl = dac;
        s1.xdl = dbc;

        if (s0.xcy > s1.xcy)
        {
            s2 = s0;
            s0 = s1;
            s1 = s2; 
        }

        while(1)
        {
            cy0 = s0.xcy>>DIV_SHIFT;
            cy1 = s1.xcy>>DIV_SHIFT;
            ymin = MAX(cy0,0);
            ymax = MIN(cy1,HEIGHT);
            if (ymin<ymax)
            {
#ifdef ZBUFFER
                dz = DIVIDE_NSP((s1.xcz-s0.xcz)>>DIV_SHIFT,cy1-cy0);
                curz = s0.xcz+dz*(ymin-cy0);
#endif
                cdtx = DIVIDE_NSP((s1.tx-s0.tx)>>DIV_SHIFT,cy1-cy0);
                cdty = DIVIDE_NSP((s1.ty-s0.ty)>>DIV_SHIFT,cy1-cy0);
                ctx = s0.tx+(ymin-cy0)*cdtx;
                cty = s0.ty+(ymin-cy0)*cdty;

                dl = DIVIDE_NSP((s1.xl-s0.xl)>>DIV_SHIFT,cy1-cy0);
                curl = s0.xl+dl*(ymin-cy0);

#ifdef SPEED
                for (;ymin<ymax;++ymin)
                {
                    view3d[dx+ymin] = t->lpal[GPLIGHT(curl)*nbc+tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw]];
                    ctx+=cdtx;
                    cty+=cdty;
                    curl+=dl;
                }
#else
                for (;ymin<ymax;++ymin)
                {
#ifdef ZBUFFER
                    cz = curz>>DIV_SHIFT;
                    if (cz>GET(zbuf,dx+ymin))
                    {
#endif
                        pix = tex[(ctx>>DIV_SHIFT)+(cty>>DIV_SHIFT)*dw];
#ifdef USE_TRANS
                        if (pix!=trans)
                        {
#endif
                            view3d[dx+ymin] = t->lpal[GPLIGHT(curl)*nbc+pix];
#ifdef ZBUFFER
                            GET(zbuf,dx+ymin) = cz;
#endif
#ifdef CELLSHADING
                            GET(cs_from,dx+ymin) = nt;
#endif
#ifdef USE_TRANS
                        }
#endif
#ifdef ZBUFFER
                    }
                    curz+=dz;
#endif
                    curl+=dl;
                    ctx+=cdtx;
                    cty+=cdty;
                }
#endif // SPEED
            }
            if (++bx > cx)
                break;
            dx+=HEIGHT;
            s0.tx+=s0.dtx;
            s0.ty+=s0.dty;
            s1.tx+=s1.dtx;
            s1.ty+=s1.dty;
            s0.xcy+=s0.xdy;
            s1.xcy+=s1.xdy;
#ifdef ZBUFFER
            s0.xcz+=s0.xdz;
            s1.xcz+=s1.xdz;
#endif
            s0.xl+=s0.xdl;
            s1.xl+=s1.xdl;
        }
    }
}

/**
 *
 */

#ifdef FORCE_FLAT
void drawTriangle(unsigned short n,unsigned short a,unsigned short b,unsigned short c,unsigned char quad,unsigned char flat)
#else
void drawTriangle(unsigned short n,unsigned short a,unsigned short b,unsigned short c,unsigned char quad)
#endif
{
    unsigned short aa,bb,cc;
    unsigned char  oa,ob,oc;
    TRIANGLE3D *t = GET(triangles,n).orig;
    unsigned int col;
    TRIANGLE3D_TYPE type = 0;
    TRIANGLE3D_DATA *data = NULL;

__drawn__++;

    if (GET(drawvert,a).x<GET(drawvert,b).x)
    {
        if (GET(drawvert,a).x<GET(drawvert,c).x)
        {
            aa = a;
            oa = 0;
            if (GET(drawvert,b).x<GET(drawvert,c).x)
            {
                bb = b;
                cc = c;
                ob = quad?2:1;
                oc = quad?3:2;
            }
            else
            {
                bb = c;
                cc = b;
                ob = quad?3:2;
                oc = quad?2:1;
            }
        }
        else
        {
            aa = c;
            bb = a;
            cc = b;
            oa = quad?3:2;
            ob = 0;
            oc = quad?2:1;
        }
    }
    else
    {
        if (GET(drawvert,b).x<GET(drawvert,c).x)
        {
            aa = b;
            oa = quad?2:1;
            if (GET(drawvert,a).x<GET(drawvert,c).x)
            {
                bb = a;
                cc = c;
                ob = 0;
                oc = quad?3:2;
            }
            else
            {
                bb = c;
                cc = a;
                ob = quad?3:2;
                oc = 0;
            }
        }
        else
        {
            aa = c;
            bb = b;
            cc = a;
            oa = quad?3:2;
            ob = quad?2:1;
            oc = 0;
        }
    }

    if (t->type==TEXTURED_N)
        type = TEXTURED;
    else if (t->type==COLORED_N)
        type = COLORED;
    else
    {
        switch(shading3d)
        {
#ifdef NO_TEXTURE
#ifdef FORCE_FLAT
            case GOURAUD: type = (GET(tmid,n).z>=gdMin && !flat) ? GOURAUD : FLAT; break;
#else
            case GOURAUD: type = (GET(tmid,n).z>=gdMin) ? GOURAUD : FLAT; break;
#endif
            case FLAT:    type = FLAT; break;
            default:      type = COLORED;
#else
            case GOURAUD:
#ifdef FORCE_FLAT
                if (GET(tmid,n).z>=gdMin && !flat)
#else
                if (GET(tmid,n).z>=gdMin)
#endif
                {
                    if (t->type==COLORED)
                        type = GOURAUD;
                    else
                        type = TEXTURED_G;
                    break;
                } /* else FLAT */
            case FLAT:
                if (t->type==COLORED)
                    type = FLAT;
                else
                    type = TEXTURED_F;
                break;
            default:
                if ((t->type==TEXTURED) && (GET(tmid,n).z<=gdMin))
                    type = COLORED;
                else
                    type = t->type;
                break;
#endif
        }
    }

#ifdef OVERLOADS
    data = GET(overloads,n)?GET(overloads,n):&t->data;
#else
    data = &t->data;
#endif

    switch(type)
    {
        case COLORED:
            drawFlatTriangle(aa,bb,cc,(t->type==TEXTURED?data->td->t->lcol.pix[lightDef]:data->c->pix[lightDef]),n);
            break;
        case FLAT:
#ifdef POINT_LIGHT
            col = computeLight(&GET(tmid,n),&GET(tnorms,n))>>(SHIFT-4);
#else
            col = computeLight(&GET(tnorms,n))>>(SHIFT-4);
            col = MIN(col,31);
#endif
            drawFlatTriangle(aa,bb,cc,(t->type==TEXTURED?data->td->t->lcol.pix[col]:data->c->pix[col]),n);
            break;
        case GOURAUD:
            drawGouraudTriangle(aa,bb,cc,(t->type==TEXTURED)?data->td->t->lcol.pix:data->c->pix,n);
            break;
        case TEXTURED:
#ifdef OVERLOADS
            drawFlatTexturedTriangle(aa,bb,cc,oa,ob,oc,lightDef,data->td,GET(overloads,n)?data->t:NULL,n);
#else
            drawFlatTexturedTriangle(aa,bb,cc,oa,ob,oc,lightDef,data->td,NULL,n);
#endif
            break;
        case TEXTURED_F:
#ifdef POINT_LIGHT
            col = computeLight(&GET(tmid,n),&GET(tnorms,n))>>(SHIFT-4);
#else
            col = computeLight(&GET(tnorms,n))>>(SHIFT-4);
            col = MIN(col,31);
#endif
#ifdef OVERLOADS
            drawFlatTexturedTriangle(aa,bb,cc,oa,ob,oc,col,t->data.td,GET(overloads,n)?data->t:NULL,n);
#else
            drawFlatTexturedTriangle(aa,bb,cc,oa,ob,oc,col,t->data.td,NULL,n);
#endif
            break;
        case TEXTURED_G:
#ifdef OVERLOADS
            drawGouraudTexturedTriangle(aa,bb,cc,oa,ob,oc,t->data.td,GET(overloads,n)?data->t:t->data.td->t,n);
#else
            drawGouraudTexturedTriangle(aa,bb,cc,oa,ob,oc,t->data.td,t->data.td->t,n);
#endif
        default:
            break;
    }
}

#ifdef STARS

/**
 * Show stars in view
 */

void showStars()
{
    int i,x,y,z=(zMax+zStars)/31;
    VERTEX3D v;
    unsigned char c;
    for (i=0;i<nbStars;++i)
    {
        MULTIPLY(main3d.m,stars[i],v);
        if (v.z>zMax)
            continue;

        x = WIDTH2-((v.x*proj3d)/(proj3d-v.z));
        if (x<0||x>=WIDTH)
            continue;

        y = ((v.y*proj3d)/(proj3d-v.z))+HEIGHT2;
        if  (   y<0
             || y>=HEIGHT
#ifdef ZBUFFER
             || GET(zbuf,y*HEIGHT+x)>v.z
#endif
            )
            continue;

        c = 31-((zMax-v.z)/z);
        view3d[x*HEIGHT+y] = (c<<11)|(c<<6)|(c<<1);
    }
}

#endif

#ifndef ZBUFFER

/**
 * Sort
 */

#ifdef QSORT

void o3d_qsort(unsigned short*a,unsigned short*b,unsigned short start,unsigned short end)
{
    int i,j,k;
    int aa = GET(a,start);
    int za = GET(tmid,aa).z;

    for (i=start+1,j=start,k=end;i<end;++i)
        GET(b,(GET(tmid,GET(a,i)).z<za)?j++:--k) = GET(a,i);
    GET(sort_c,j) = aa;

    if (j>start)
    {
        if (start+1<j)
            o3d_qsort(b,a,start,j);
        else
            GET(sort_c,start) = GET(b,start);
    }

    if (++j<end)
    {
        if (j+1<end)
            o3d_qsort(b,a,j,end);
        else
            GET(sort_c,j) = GET(b,j);
    }
}

#else
#ifdef QSORT2

void o3d_qsort2(int max)
{
    int aa,i,j,k;
    int za;
    int start,end,swa,swb;
    unsigned short*stackptr=sort_st;
    unsigned short*a,*b;
    unsigned short*swap[2];
    
    swap[0] = sort_a;
    swap[1] = sort_b;

    start = 0;
    end = max;
    swa = 0;

    while(1)
    {
        a = swap[swa];
        b = swap[swb=1-swa];

        aa = GET(a,start);
        za = GET(tmid,aa).z;
        for (i=start+1,j=start,k=end;i<end;++i)
            GET(b,(GET(tmid,GET(a,i)).z<za)?j++:--k) = GET(a,i);
        GET(sort_c,j) = aa;

        if (j>start)
        {
            if (start+1<j)
            {
                *stackptr++ = swb;
                *stackptr++ = j;
                *stackptr++ = start;
            }
            else
                GET(sort_c,start) = GET(b,start);
        }

        if (++j<end)
        {
            if (j+1<end)
            {
                swa = swb;
                start = j;
                continue;
            }
            GET(sort_c,j) = GET(b,j);
        }

        if (stackptr--==sort_st)
            break;

        start = *stackptr--;
        end   = *stackptr--;
        swa   = *stackptr;
    }
}

#else
#ifdef RADIXSORT

unsigned short rs_count[256];
unsigned short rs_index[256];

void o3d_radixStep(int shift, int N, unsigned short* source, unsigned short* dest)
{
    int i;

    memset(rs_count,0,sizeof(unsigned short)<<8);
    for (i=0;i<N;++i)
        rs_count[(GET(tmid,GET(source,i)).z>>shift)&0xff]++;

    rs_index[0]=0;
    for (i=1;i<256;++i)
        rs_index[i]=rs_index[i-1]+rs_count[i-1];

    for (i=0;i<N;++i)
        GET(dest,rs_index[(GET(tmid,GET(source,i)).z>>shift)&0xff]++) = GET(source,i);
}

void o3d_radixSort(unsigned short* source, unsigned short* temp, unsigned short N)
{
    o3d_radixStep(0, N, source, temp);
    o3d_radixStep(4, N, temp, source);
    o3d_radixStep(8, N, source, temp);
    o3d_radixStep(12, N, temp, source);
}

#else // HEAPSORT

#define ZDATA(i) GET(tmid,GET(sort_a,i)).z
#define ZSWAP(i,j) tmp = GET(sort_a,i); GET(sort_a,i) = GET(sort_a,j); GET(sort_a,j) = tmp

void o3d_siftDown(unsigned short root, unsigned short end)
{
    unsigned short child = (root<<1) + 1;
    unsigned short tmp;
    while (child <= end)
    {
        if (child<end && ZDATA(child)<ZDATA(child+1))
            child+=1;
        if (ZDATA(root)<ZDATA(child))
        {
            ZSWAP(root, child);
            root = child;
        }
        else
            return;
        child = (root<<1) + 1;
    }
}

void o3d_heapSort(unsigned short count)
{
    unsigned short end = count - 1;
    unsigned short tmp;
    for (tmp=(count>>1)-1;tmp>0;--tmp)
        o3d_siftDown(tmp,end);
    o3d_siftDown(0,end);
    while (end > 0)
    {
        ZSWAP(end, 0);
        o3d_siftDown(0, --end);
    }
}

#endif // RADIX

#endif // QSORT2

#endif // QSORT

#endif // ZBUFFER

void o3d_initVNormals(OBJECT3D *o)
{
    int i;
    VECTOR3D v;

    for (i=0;i<o->nbv;++i)
        VNORM(o,i,0,0,0)
    for (i=0;i<o->nbt;++i)
    {
        o->vnorm[o->tri[i].a].x+=o->tnorm[i].x;
        o->vnorm[o->tri[i].a].y+=o->tnorm[i].y;
        o->vnorm[o->tri[i].a].z+=o->tnorm[i].z;
        o->vnorm[o->tri[i].b].x+=o->tnorm[i].x;
        o->vnorm[o->tri[i].b].y+=o->tnorm[i].y;
        o->vnorm[o->tri[i].b].z+=o->tnorm[i].z;
        o->vnorm[o->tri[i].c].x+=o->tnorm[i].x;
        o->vnorm[o->tri[i].c].y+=o->tnorm[i].y;
        o->vnorm[o->tri[i].c].z+=o->tnorm[i].z;
        if (o->tri[i].d!=NOTSET)
        {
            o->vnorm[o->tri[i].d].x+=o->tnorm[i].x;
            o->vnorm[o->tri[i].d].y+=o->tnorm[i].y;
            o->vnorm[o->tri[i].d].z+=o->tnorm[i].z;
        }
    }
    for (i=0;i<o->nbv;++i)
    {
        v.x = o->vnorm[i].x;
        v.y = o->vnorm[i].y;
        v.z = o->vnorm[i].z;
        normalize(&v);
        VNORM(o,i,v.x,v.y,v.z)
    }
}

void o3d_initTNormals(OBJECT3D *o)
{
    VECTOR3D v;
    int i;
    for (i=0;i<o->nbt;++i)
    {
        v.x = ((o->vert[o->tri[i].c].y-o->vert[o->tri[i].b].y)*(o->vert[o->tri[i].b].z-o->vert[o->tri[i].a].z))-
              ((o->vert[o->tri[i].c].z-o->vert[o->tri[i].b].z)*(o->vert[o->tri[i].b].y-o->vert[o->tri[i].a].y));
        v.y = ((o->vert[o->tri[i].c].z-o->vert[o->tri[i].b].z)*(o->vert[o->tri[i].b].x-o->vert[o->tri[i].a].x))-
              ((o->vert[o->tri[i].c].x-o->vert[o->tri[i].b].x)*(o->vert[o->tri[i].b].z-o->vert[o->tri[i].a].z));
        v.z = ((o->vert[o->tri[i].c].x-o->vert[o->tri[i].b].x)*(o->vert[o->tri[i].b].y-o->vert[o->tri[i].a].y))-
              ((o->vert[o->tri[i].c].y-o->vert[o->tri[i].b].y)*(o->vert[o->tri[i].b].x-o->vert[o->tri[i].a].x));
        normalize(&v);
        TNORM(o,i,v.x,v.y,v.z)
        if (o->tri[i].d==NOTSET)
        {
            v.x = (o->vert[o->tri[i].a].x+o->vert[o->tri[i].b].x+o->vert[o->tri[i].c].x)/3;
            v.y = (o->vert[o->tri[i].a].y+o->vert[o->tri[i].b].y+o->vert[o->tri[i].c].y)/3;
            v.z = (o->vert[o->tri[i].a].z+o->vert[o->tri[i].b].z+o->vert[o->tri[i].c].z)/3;
        }
        else
        {
            v.x = (o->vert[o->tri[i].a].x+o->vert[o->tri[i].b].x+o->vert[o->tri[i].c].x+o->vert[o->tri[i].d].x)>>2;
            v.y = (o->vert[o->tri[i].a].y+o->vert[o->tri[i].b].y+o->vert[o->tri[i].c].y+o->vert[o->tri[i].d].y)>>2;
            v.z = (o->vert[o->tri[i].a].z+o->vert[o->tri[i].b].z+o->vert[o->tri[i].c].z+o->vert[o->tri[i].d].z)>>2;
        }
        VECTOR(o,i,tmid,v.x,v.y,v.z)
    }
}

void o3d_initBBox(OBJECT3D *o)
{
    int i;
    short maxx = o->vert[0].x;
    short maxy = o->vert[0].y;
    short maxz = o->vert[0].z;
    short minx = maxx;
    short miny = maxy;
    short minz = maxz;
    for (i=1;i<o->nbv;++i)
    {
        maxx = MAX(o->vert[i].x,maxx);
        maxy = MAX(o->vert[i].y,maxy);
        maxz = MAX(o->vert[i].z,maxz);
        minx = MIN(o->vert[i].x,minx);
        miny = MIN(o->vert[i].y,miny);
        minz = MIN(o->vert[i].z,minz);
    }
    o->bbox[0].x = minx;
    o->bbox[0].y = maxy;
    o->bbox[0].z = maxz;
    o->bbox[1].x = maxx;
    o->bbox[1].y = maxy;
    o->bbox[1].z = maxz;
    o->bbox[2].x = maxx;
    o->bbox[2].y = miny;
    o->bbox[2].z = maxz;
    o->bbox[3].x = minx;
    o->bbox[3].y = miny;
    o->bbox[3].z = maxz;
    o->bbox[4].x = minx;
    o->bbox[4].y = maxy;
    o->bbox[4].z = minz;
    o->bbox[5].x = maxx;
    o->bbox[5].y = maxy;
    o->bbox[5].z = minz;
    o->bbox[6].x = maxx;
    o->bbox[6].y = miny;
    o->bbox[6].z = minz;
    o->bbox[7].x = minx;
    o->bbox[7].y = miny;
    o->bbox[7].z = minz;
}

unsigned short read16(unsigned char **data)
{
    unsigned short n = *((*data)++);
    return (n<<8)|(*((*data)++));
}

unsigned char* o3d_readObject(OBJECT3D* o,unsigned char* data,unsigned char readT,TEXTURE3D *t)
{
    int i,j,n;
    unsigned int q;

    if (readT)
    {
        o->nbc = n = read16(&data); // nb colors
        o->colors = n==0 ? NULL:(LCOLOR3D*)MALLOC(sizeof(LCOLOR3D),n);
        for (i=0;i<n;++i)
        {
            o->colors[i].col.r = *data++;
            o->colors[i].col.g = *data++;
            o->colors[i].col.b = *data++;
            o3d_registerColor(&o->colors[i]);
        }
    }
    else
    {
        o->nbc = 0;
        o->colors = NULL;
    }

    // vertices
    o->nbv = n = read16(&data);
    o->vert = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),n);
    for (i=0;i<n;++i)
    {
        o->vert[i].x = read16(&data);
        o->vert[i].y = read16(&data);
        o->vert[i].z = read16(&data);
    }
    // bounding box
    o->bbox = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),8);
    for (i=0;i<8;++i)
    {
        o->bbox[i].x = read16(&data);
        o->bbox[i].y = read16(&data);
        o->bbox[i].z = read16(&data);
    }
    // vertex normals
    n = read16(&data); // nbvn
    o->vnorm = n==0 ? NULL : (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),n);
    for (i=0;i<n;++i)
    {
        o->vnorm[i].x = read16(&data);
        o->vnorm[i].y = read16(&data);
        o->vnorm[i].z = read16(&data);
    }
     // triangles
    if (readT)
    {
        n = read16(&data);
        q = (n&0x8000) != 0;
        n &= 0x7fff;
        o->nbt = n;
        o->tri = (TRIANGLE3D*)MALLOC(sizeof(TRIANGLE3D),n);
        o->tmid = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),n);
        for (i=0;i<n;++i)
        {
            o->tri[i].a = read16(&data);
            o->tri[i].b = read16(&data);
            o->tri[i].c = read16(&data);
            o->tri[i].d = q ? read16(&data) : NOTSET;
            j = read16(&data);
            if (j&0x8000) // textured
            {
                TEXTURE3D_DATA *td = (TEXTURE3D_DATA*)MALLOC(sizeof(TEXTURE3D_DATA),1);
                o->tri[i].type = TEXTURED;
                o->tri[i].data.td = td;
                td->t = &t[j&0x7fff];
                td->x[0] = *data++;
                td->x[1] = *data++;
                td->x[2] = *data++;
                if (q)
                    td->x[3] = *data++;
                td->y[0] = *data++;
                td->y[1] = *data++;
                td->y[2] = *data++;
                if (q)
                    td->y[3] = *data++;
            }
            else // colored
            {
                o->tri[i].type = COLORED;
                o->tri[i].data.c = &o->colors[j];
            }
        }
    }
    // triangle normales
    n = read16(&data); // nbtn
    o->tnorm = n==0 ? NULL : (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),n);
    for (i=0;i<n;++i)
    {
        o->tnorm[i].x = read16(&data);
        o->tnorm[i].y = read16(&data);
        o->tnorm[i].z = read16(&data);
        o->tmid[i].x = (o->vert[o->tri[i].a].x+o->vert[o->tri[i].b].x+o->vert[o->tri[i].c].x)/3;
        o->tmid[i].y = (o->vert[o->tri[i].a].y+o->vert[o->tri[i].b].y+o->vert[o->tri[i].c].y)/3;
        o->tmid[i].z = (o->vert[o->tri[i].a].z+o->vert[o->tri[i].b].z+o->vert[o->tri[i].c].z)/3;
    }
    return data;
}

NODE3D_TYPE o3d_createObject(unsigned char* data,TEXTURE3D* t,NODE3D_DATA* ret,unsigned char doVN)
{
    NODE3D_TYPE    type;
    OBJECT3D *o;
    int i,n;

    switch(*data++)
    {
        case 0: // scene
            ret->s = (SCENE3D*)MALLOC(sizeof(SCENE3D),1);
            ret->s->count = 0;
            n = *data++;
            for (i=0;i<n;++i)
            {
                o = (OBJECT3D*)MALLOC(sizeof(OBJECT3D),1);
                data = o3d_readObject(o,data,1,t);
                if (o->tnorm==NULL)
                {
                    o->tnorm = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),o->nbt);
                    o->tmid = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),o->nbt);
                    o3d_initTNormals(o);
                }
                o3d_addObject(ret->s,o,0,0,0,0,0,0);
            }
            type = SCENE;
            break;
        case 1: // animation
            ret->a = (ANIMATION3D*)MALLOC(sizeof(ANIMATION3D),1);
            type = ANIMATION;
            ret->a->anim_count = n = *data++;
            ret->a->anim_start = (unsigned short*)MALLOC(sizeof(unsigned short),n);
            ret->a->anim_end = (unsigned short*)MALLOC(sizeof(unsigned short),n);
            for (i=0;i<n;++i)
            {
                ret->a->anim_start[i] = read16(&data);
                ret->a->anim_end[i] = read16(&data);
            }
            ret->a->frame_count = read16(&data);
            n = ret->a->frame_count;
            ret->a->anims = (OBJECT3D**)MALLOC(sizeof(OBJECT3D*),n);
            for (i=0;i<n;++i)
            {
                o = ret->a->anims[i] = (OBJECT3D*)MALLOC(sizeof(OBJECT3D),1);
                data = o3d_readObject(o,data,i==0,t);
                if (i>0)
                {
                    o->nbt = ret->a->anims[0]->nbt;
                    o->tri = ret->a->anims[0]->tri;
                }
                if (o->tnorm==NULL)
                {
                    o->tnorm = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),o->nbt);
                    o->tmid = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),o->nbt);
                    o3d_initTNormals(o);
                    if (doVN)
                    {
                        o->vnorm = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),o->nbv);
                        o3d_initVNormals(o);
                    }
                }
            }
            break;
        default:
        case 2: // object
            ret->o = (OBJECT3D*)MALLOC(sizeof(OBJECT3D),1);
            type = OBJECT;
            o3d_readObject(ret->o,data,1,t);
            if (ret->o->tnorm==NULL)
            {
                ret->o->tnorm = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),ret->o->nbt);
                ret->o->tmid = (SVECTOR3D*)MALLOC(sizeof(SVECTOR3D),ret->o->nbt);
                o3d_initTNormals(ret->o);
            }
            break;
    }
    return type;
}

void o3d_freeObject(OBJECT3D* o,unsigned char freeT)
{
    int i;
    if (o->colors!=NULL)
    {
        for (i=0;i<o->nbc;++i)
            o3d_registerColor(&o->colors[i]);
        FREE(o->colors);
    }

    FREE(o->vert);
    if (o->vnorm!=NULL)
        FREE(o->vnorm);
    for (i=0;freeT && i<o->nbt;++i)
    {
        if (o->tri[i].type==TEXTURED)
            FREE(o->tri[i].data.td);
    }
    if (freeT)
        FREE(o->tri);
    FREE(o->tnorm);
    FREE(o->bbox);
    FREE(o);
}

void o3d_destroyObject(NODE3D_DATA* data,NODE3D_TYPE t)
{
    int i;

    switch(t)
    {
        case OBJECT:
            o3d_freeObject(data->o,1);
            break;
        case SCENE:
            for (i=0;i<data->s->count;++i)
            {
                o3d_destroyObject(&data->s->nodes[i]->data,data->s->nodes[i]->type);
                FREE(data->s->nodes[i]);
            }
            FREE(data->s->nodes);
            FREE(data->s);
            break;
        case ANIMATION:
            for (i=0;i<data->a->frame_count;++i)
                o3d_freeObject(data->a->anims[i],i==0);
            FREE(data->a->anim_start);
            FREE(data->a->anim_end);
            FREE(data->a->anims);
            FREE(data->a);
        default:
            break;
    }
}

void o3d_lightTextures(TEXTURE3D* t,int count)
{
#ifdef BITS_16
    int i,j,k,l;
    COLOR3D col;

    initColorTab();
    for (i=0;i<count;++i)
    {
        if (t[i].lpal==NULL)
            continue;
        l = t[i].nbc+1;
        for (j=0;j<l;++j)
        {
            for (k=0;k<32;++k)
            {
                computeColor((k+1)<<(SHIFT-4),&t[i].pal[j],&col);
                t[i].lpal[k*l+j] = GPCOLX(col.r,col.g,col.b);
            }
        }
    }
#endif
}

void o3d_createTextures(unsigned char *data,int light,TEXTURE3DSET *tset)
{
    int i,j,n,l;
    TEXTURE3D* t;
    unsigned long smem;
    unsigned short* gpal0;

    n = tset->count = (*data++)+1;
    t = tset->set = (TEXTURE3D*)MALLOC(sizeof(TEXTURE3D),n);

#ifdef __GP32__
    smem = sizeof(unsigned short)*256*n+4096;
    tset->gpal = (unsigned long)MALLOC(smem,1);
    gpal0 = (unsigned short*)((tset->gpal+4095)&0xfffff000);
    gp_setMMU((unsigned long)gpal0, tset->gpal+smem-1,0xFFA);
#else
    smem = sizeof(unsigned short)*256*n;
    tset->gpal = (unsigned long)MALLOC(smem,1);
    gpal0 = (unsigned short*)tset->gpal;
#endif

    for (i=0;i<n;++i)
    {
        t[i].trans = *data++;
// TODO ??
t[i].trans = 0xff;
        t[i].lcol.col.r = *data++;
        t[i].lcol.col.g = *data++;
        t[i].lcol.col.b = *data++;
        o3d_registerColor(&t[i].lcol);
        t[i].nbc = *data++;
        l = t[i].nbc+1;
        t[i].pal = (COLOR3D*)MALLOC(sizeof(COLOR3D),l);
        for (j=0;j<l;++j)
        {
            t[i].pal[j].r = *data++;
            t[i].pal[j].g = *data++;
            t[i].pal[j].b = *data++;
        }
        t[i].gpal = gpal0+i*256;
#ifdef __GP32__
        for (j=0;j<l;++j)
            t[i].gpal[j] = read16(&data);
#else
        for (j=0;j<l;++j)
        {
            unsigned short gcol = read16(&data);
            gcol = (gcol&0xffc0) | ((gcol&0x3f)>>1);
            t[i].gpal[j] = gcol;
        }
#endif
#ifdef BITS_16
        t[i].lpal = light?(unsigned short*)MALLOC(sizeof(unsigned short),l*32):NULL;
#endif
        t[i].w = *data++;
        t[i].h = *data++;
        l = (t[i].w+1)*(t[i].h+1);
        t[i].pix = (unsigned char*)MALLOC(sizeof(unsigned char),l);
        memcpy(t[i].pix,data,l);
        data+=l;
    }
#ifdef BITS_16
    if (light)
        o3d_lightTextures(t,n);
#endif
}

void o3d_destroyTextures(TEXTURE3DSET* tset)
{
    int i;
    for (i=0;i<tset->count;++i)
    {
        FREE(tset->set[i].pal);
        FREE(tset->set[i].pix);
#ifdef BITS_16
        if (tset->set[i].lpal!=NULL)
            FREE(tset->set[i].lpal);
#endif
        o3d_unregisterColor(&tset->set[i].lcol);
    }
    if (tset->set)
        FREE(tset->set);
    if (tset->gpal)
        FREE((void*)tset->gpal);
    tset->set = NULL;
    tset->gpal = 0;
    tset->count = 0;
}
