#ifndef _O3D_H_
#define _O3D_H_

#ifndef NULL
#define NULL 0
#endif

#define GP_CC_MAX 0xf8
#define GPCOLU(a)		((a)>=GP_CC_MAX?GP_CC_MAX:((a)&GP_CC_MAX)+(((a)&4)<<1))
#ifdef __GP32__
#define GPCOL(r,g,b)	((GPCOLU(r)<<8)|(GPCOLU(g)<<3)|(GPCOLU(b)>>2))
#else
#define GPCOL(r,g,b)	((GPCOLU(r)<<8)|(GPCOLU(g)<<3)|(GPCOLU(b)>>3))
#endif
#define GPCOL3D(c)		GPCOL((c)->r,(c)->g,(c)->b)

#define VECTOR(o,i,t,xi,yi,zi) \
	{ \
	(o)->t[i].x = (xi); \
	(o)->t[i].y = (yi); \
	(o)->t[i].z = (zi); \
	}

#define VNORM(o,i,xi,yi,zi)  VECTOR(o,i,vnorm,xi,yi,zi)
#define TNORM(o,i,xi,yi,zi)  VECTOR(o,i,tnorm,xi,yi,zi)

#define WIDTH	320
#define HEIGHT	240
#define SIZE	(WIDTH*HEIGHT)

#define WIDTH2	160
#define HEIGHT2	120

#define MAX_STARS 1000

#define NODE3D_ALLOC 16

#define NOTSET 0xffff /* d value for triangles, set for quad */

typedef enum
	{
		VAL = 0,
		ROTX,
		ROTY,
		ROTZ,
		ROTXYZ,
		SCALE,
		TRANS,
		IDENT,
		PROJ
	} MATRIX_TYPE;

#define _MATRIX3D(m) int m[16]

typedef int* MATRIX3D;

typedef struct
	{
		int x,y;
	} VECTOR2D;

typedef struct
	{
		int x,y,z;
	} VECTOR3D;

typedef struct
	{
		short x,y,z;
	} SVECTOR3D;

#define VERTEX3D VECTOR3D
#define SVERTEX3D SVECTOR3D

typedef enum
	{
		NONE = 0,
		FLAT,
		GOURAUD,
		TEXTURED_F,
		TEXTURED_G,
		COLORED,
		TEXTURED,
		COLORED_N,
		TEXTURED_N
	} TRIANGLE3D_TYPE;

typedef struct
	{
		unsigned char r,g,b;
	} COLOR3D;

typedef struct
	{
		COLOR3D col;
		unsigned short pix[33];
	} LCOLOR3D;

typedef struct Texture3D
	{
		unsigned char	w,h; /* -1 */
		unsigned char	nbc; /* nb colors in pal -1 */
		COLOR3D			*pal;
		unsigned short	*gpal;
		unsigned char	*pix;
		unsigned char	trans;
		LCOLOR3D		lcol;
		unsigned short	*lpal;
	} TEXTURE3D;

typedef struct
	{
		unsigned char	x[4];
		unsigned char	y[4];
		TEXTURE3D		*t;
	} TEXTURE3D_DATA;

typedef union
	{
		void			*data;
		LCOLOR3D		*c;
		TEXTURE3D_DATA	*td;
		TEXTURE3D		*t;
	} TRIANGLE3D_DATA;

typedef struct
	{
		unsigned short	a,b,c,d;
		TRIANGLE3D_TYPE type;
		TRIANGLE3D_DATA	data;
	} TRIANGLE3D;

typedef struct
	{
		unsigned short	a,b,c,d;
		TRIANGLE3D		*orig;
	} REFTRI3D;

typedef struct
	{
		unsigned short rx,ry,rz;
	} ROTATION3D;

typedef struct Object3D
	{
		unsigned short	nbv;
		SVERTEX3D		*vert;
		SVECTOR3D		*vnorm;
		unsigned short	nbt;
		TRIANGLE3D		*tri;
		SVECTOR3D		*tnorm;
        SVECTOR3D       *tmid;
		SVERTEX3D		*bbox;
		unsigned short	nbc;
		LCOLOR3D		*colors;
	} OBJECT3D;

typedef enum
	{
		OBJECT = 0,
		SPRITE,
		SCENE,
		ANIMATION,
		UNKNOWN
	} NODE3D_TYPE;

struct Scene3D;
struct Animation3D;

typedef union
	{
		struct Object3D		*o;
		struct Scene3D		*s;
		struct Sprite3D		*p;
		struct Animation3D	*a;
		void				*n;
	} NODE3D_DATA;

typedef struct
	{
		NODE3D_TYPE		type;
		NODE3D_DATA		data;
		VECTOR3D		pos;
		ROTATION3D		rot;
		_MATRIX3D		(m);
		unsigned char	vis;
		unsigned char	hide;
		TRIANGLE3D_DATA	*overload;
		unsigned short	id;
		unsigned short	cur_frame;
	} NODE3D;

typedef struct Scene3D
	{
		unsigned short	count;
		unsigned short	reserved;
		NODE3D			**nodes;
	} SCENE3D;

typedef struct Animation3D
	{
		unsigned char	anim_count;
		unsigned short	*anim_start;
		unsigned short	*anim_end;
		unsigned short	frame_count;
		unsigned short	cur_frame;
		OBJECT3D		**anims;
	} ANIMATION3D;

typedef struct Sprite3D
	{
		unsigned short	dx,dy;
		unsigned short	w,h;
		TEXTURE3D		*t;
	} SPRITE3D;

typedef struct Texture3DSet
	{
		int count;
		unsigned long gpal;
		TEXTURE3D *set;
	} TEXTURE3DSET;

/*
 * Public usage
 */

NODE3D*		o3d_init			(SCENE3D*,int,int,int,int);
void		o3d_free			(void);
VECTOR3D*	o3d_setLightPos		(NODE3D*,int,int,int);
COLOR3D*	o3d_setLightColor	(unsigned char,unsigned char,unsigned char);
void		o3d_setLightDefaut	(unsigned int);
void		o3d_setBorderColor	(unsigned short,unsigned short);
void	 	o3d_initScene		(SCENE3D*);
void		o3d_clearScene		(SCENE3D*);
NODE3D* 	o3d_addObject		(SCENE3D*,OBJECT3D*,int,int,int,unsigned short,unsigned short,unsigned short);
NODE3D* 	o3d_addScene		(SCENE3D*,SCENE3D*,int,int,int,unsigned short,unsigned short,unsigned short);
NODE3D*		o3d_addSprite		(SCENE3D*,SPRITE3D*,int,int,int);
NODE3D*		o3d_addAnimation	(SCENE3D*,ANIMATION3D*,int,int,int,unsigned short,unsigned short,unsigned short);
NODE3D* 	o3d_addNode			(SCENE3D*,void*,NODE3D_TYPE,int,int,int,unsigned short,unsigned short,unsigned short);
void	 	o3d_removeNode		(SCENE3D*,NODE3D*);
void	 	o3d_removeNodeAt	(SCENE3D*,int);
SCENE3D*	o3d_cloneScene		(SCENE3D*);
void		o3d_destroyClone	(SCENE3D*);
void        o3d_registerColor   (LCOLOR3D*);
void        o3d_unregisterColors(void);
void        o3d_unregisterColor (LCOLOR3D*);
void		o3d_setGDMin		(int);
void		o3d_setDepths		(int,int);
void 		o3d_computeFrame	(unsigned short *, int);
void		o3d_initStars		(int,int);
void		o3d_setShading		(TRIANGLE3D_TYPE);
void		o3d_initVNormals	(OBJECT3D *o);
void		o3d_initTNormals	(OBJECT3D *o);
NODE3D_TYPE	o3d_createObject	(unsigned char*,TEXTURE3D*,NODE3D_DATA*,unsigned char);
void		o3d_destroyObject	(NODE3D_DATA*,NODE3D_TYPE);
void		o3d_freeObject		(OBJECT3D*,unsigned char);
void		o3d_createTextures	(unsigned char*,int,TEXTURE3DSET*);
void		o3d_destroyTextures	(TEXTURE3DSET*);
void		o3d_lightTextures	(TEXTURE3D*,int);
void		o3d_initBBox		(OBJECT3D*);

#endif
