#include "o3d.h"
#include "o3dbasic.h"
#include "tables.h"
#include <stdlib.h>

void o3d_alloc(OBJECT3D* o)
{
	o->vert		= (SVERTEX3D*)malloc(sizeof(SVERTEX3D)*o->nbv);
	o->vnorm	= (SVECTOR3D*)malloc(sizeof(SVECTOR3D)*o->nbv);
	o->tri		= (TRIANGLE3D*)malloc(sizeof(TRIANGLE3D)*o->nbt);
	o->tnorm	= (SVECTOR3D*)malloc(sizeof(SVECTOR3D)*o->nbt);
    o->tmid     = (SVECTOR3D*)malloc(sizeof(SVECTOR3D)*o->nbt);
	o->bbox		= (SVERTEX3D*)malloc(sizeof(SVERTEX3D)*8);
	o->colors	= o->nbc == 0 ? NULL : (LCOLOR3D*)malloc(sizeof(LCOLOR3D)*o->nbc);
}

OBJECT3D *o3d_createNewObject(int nbv,int nbt,int nbc)
{
	OBJECT3D *o = (OBJECT3D*)malloc(sizeof(OBJECT3D));

	o->nbv	= nbv;
	o->nbt	= nbt;
	o->nbc	= nbc;
	o3d_alloc(o);

	return o;
}

OBJECT3D *o3d_createSprite(int x,int y,int w,int h,TEXTURE3D* t,TRIANGLE3D_TYPE type)
{
	TEXTURE3D_DATA* data;
	OBJECT3D *o = o3d_createNewObject(4,2,0);
	int w2 = w>>1;
	int h2 = h>>1;

	VERTEX(o,0,w2,h2,0)
	VERTEX(o,1,w2,-h2,0)
	VERTEX(o,2,-w2,-h2,0)
	VERTEX(o,3,-w2,h2,0)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,x,y+h-1,x+w-1,y+h-1,x,y,t)
	TRIANGLE(o,0,1,2,0,type,data)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,x+w-1,y,x,y,x+w-1,y+h-1,t)
	TRIANGLE(o,1,3,0,2,type,data)

	o3d_initTNormals(o);
	o3d_initVNormals(o);
	o3d_initBBox(o);

	return o;
}

OBJECT3D *o3d_createColoredBox(int w,int h,int d,LCOLOR3D *col)
{
	OBJECT3D *o = o3d_createNewObject(8,12,0);

	VERTEX(o,0,w,h,d)
	VERTEX(o,1,w,-h,d)
	VERTEX(o,2,-w,-h,d)
	VERTEX(o,3,-w,h,d)
	VERTEX(o,4,-w,h,-d)
	VERTEX(o,5,-w,-h,-d)
	VERTEX(o,6,w,-h,-d)
	VERTEX(o,7,w,h,-d)

	TRIANGLE(o,0,1,2,0,COLORED,col)
	TRIANGLE(o,1,3,0,2,COLORED,col)
	TRIANGLE(o,2,5,6,4,COLORED,col)
	TRIANGLE(o,3,7,4,6,COLORED,col)
	TRIANGLE(o,4,6,1,7,COLORED,col)
	TRIANGLE(o,5,0,7,1,COLORED,col)
	TRIANGLE(o,6,6,5,1,COLORED,col)
	TRIANGLE(o,7,2,1,5,COLORED,col)
	TRIANGLE(o,8,2,5,3,COLORED,col)
	TRIANGLE(o,9,4,3,5,COLORED,col)
	TRIANGLE(o,10,4,7,3,COLORED,col)
	TRIANGLE(o,11,0,3,7,COLORED,col)

	o3d_initTNormals(o);
	o3d_initVNormals(o);
	o3d_initBBox(o);

	return o;
}

OBJECT3D *o3d_createTexturedBox(int w,int h,int d,int xt,int yt,int wt,int ht,TEXTURE3D* t,TRIANGLE3D_TYPE type)
{
	TEXTURE3D_DATA* data;
	OBJECT3D *o = o3d_createNewObject(8,12,0);

	VERTEX(o,0,w,h,d)
	VERTEX(o,1,w,-h,d)
	VERTEX(o,2,-w,-h,d)
	VERTEX(o,3,-w,h,d)
	VERTEX(o,4,-w,h,-d)
	VERTEX(o,5,-w,-h,-d)
	VERTEX(o,6,w,-h,-d)
	VERTEX(o,7,w,h,-d)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,0,1,2,0,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,1,3,0,2,type,data)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,2,5,6,4,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,3,7,4,6,type,data)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,4,6,1,7,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,5,0,7,1,type,data)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,6,6,5,1,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,7,2,1,5,type,data)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,8,2,5,3,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,9,4,3,5,type,data)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,10,4,7,3,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,11,0,3,7,type,data)

	o3d_initTNormals(o);
	o3d_initVNormals(o);
	o3d_initBBox(o);

	return o;
}

OBJECT3D *o3d_createMixedBox(int w,int h,int d,LCOLOR3D *col,int xt,int yt,int wt,int ht,TEXTURE3D* t,TRIANGLE3D_TYPE type)
{
	TEXTURE3D_DATA* data;
	OBJECT3D *o = o3d_createNewObject(8,12,0);

	VERTEX(o,0,w,h,d)
	VERTEX(o,1,w,-h,d)
	VERTEX(o,2,-w,-h,d)
	VERTEX(o,3,-w,h,d)
	VERTEX(o,4,-w,h,-d)
	VERTEX(o,5,-w,-h,-d)
	VERTEX(o,6,w,-h,-d)
	VERTEX(o,7,w,h,-d)

	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt,t)
	TRIANGLE(o,0,1,2,0,type,data)
	data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
	TEXDATA(data,xt+wt-1,yt+ht-1,xt+wt-1,yt,xt,yt+ht-1,t)
	TRIANGLE(o,1,3,0,2,type,data)

	TRIANGLE(o,2,5,6,4,COLORED,col)
	TRIANGLE(o,3,7,4,6,COLORED,col)
	TRIANGLE(o,4,6,1,7,COLORED,col)
	TRIANGLE(o,5,0,7,1,COLORED,col)
	TRIANGLE(o,6,6,5,1,COLORED,col)
	TRIANGLE(o,7,2,1,5,COLORED,col)
	TRIANGLE(o,8,2,5,3,COLORED,col)
	TRIANGLE(o,9,4,3,5,COLORED,col)
	TRIANGLE(o,10,4,7,3,COLORED,col)
	TRIANGLE(o,11,0,3,7,COLORED,col)

	o3d_initTNormals(o);
	o3d_initVNormals(o);
	o3d_initBBox(o);

	return o;
}

OBJECT3D *o3d_createTriangle(int s,LCOLOR3D *col)
{
	OBJECT3D *o = o3d_createNewObject(3,1,0);
	int a = ((PRECISION2)*tdiv[6])>>DIV_SHIFT;

	VERTEX(o,0,0,0,0)
	VERTEX(o,1,((-s*tcos[a])>>SHIFT),-((s*tsin[a])>>SHIFT),0)
	VERTEX(o,2,((-s*tcos[a])>>SHIFT), ((s*tsin[a])>>SHIFT),0)

	TRIANGLE(o,0,1,2,0,COLORED,col)

	o3d_initTNormals(o);
	o3d_initVNormals(o);
	o3d_initBBox(o);

	return o;
}

OBJECT3D *o3d_createColoredBoxQ(int w,int h,int d,LCOLOR3D *col)
{
    OBJECT3D *o = o3d_createNewObject(8,6,0);

    VERTEX(o,0,w,h,d)
    VERTEX(o,1,w,-h,d)
    VERTEX(o,2,-w,-h,d)
    VERTEX(o,3,-w,h,d)
    VERTEX(o,4,-w,h,-d)
    VERTEX(o,5,-w,-h,-d)
    VERTEX(o,6,w,-h,-d)
    VERTEX(o,7,w,h,-d)

    QUAD(o,0,0,1,2,3,COLORED,col);
    QUAD(o,1,4,5,6,7,COLORED,col);
    QUAD(o,2,7,6,1,0,COLORED,col);
    QUAD(o,3,3,2,5,4,COLORED,col);
    QUAD(o,4,7,0,3,4,COLORED,col);
    QUAD(o,5,1,6,5,2,COLORED,col);

    o3d_initTNormals(o);
    o3d_initVNormals(o);
    o3d_initBBox(o);

    return o;
}


OBJECT3D *o3d_createTexturedBoxQ(int w,int h,int d,int xt,int yt,int wt,int ht,TEXTURE3D* t,TRIANGLE3D_TYPE type)
{
    TEXTURE3D_DATA* data;
    OBJECT3D *o = o3d_createNewObject(8,6,0);

    VERTEX(o,0,w,h,d)
    VERTEX(o,1,w,-h,d)
    VERTEX(o,2,-w,-h,d)
    VERTEX(o,3,-w,h,d)
    VERTEX(o,4,-w,h,-d)
    VERTEX(o,5,-w,-h,-d)
    VERTEX(o,6,w,-h,-d)
    VERTEX(o,7,w,h,-d)

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,0,0,1,2,3,type,data);

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,1,4,5,6,7,type,data);

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,2,7,6,1,0,type,data);

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,3,3,2,5,4,type,data);

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,4,7,0,3,4,type,data);

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,5,1,6,5,2,type,data);

    o3d_initTNormals(o);
    o3d_initVNormals(o);
    o3d_initBBox(o);

    return o;
}

OBJECT3D *o3d_createMixedBoxQ(int w,int h,int d,LCOLOR3D *col,int xt,int yt,int wt,int ht,TEXTURE3D* t,TRIANGLE3D_TYPE type)
{
    TEXTURE3D_DATA* data;
    OBJECT3D *o = o3d_createNewObject(8,6,0);

    VERTEX(o,0,w,h,d)
    VERTEX(o,1,w,-h,d)
    VERTEX(o,2,-w,-h,d)
    VERTEX(o,3,-w,h,d)
    VERTEX(o,4,-w,h,-d)
    VERTEX(o,5,-w,-h,-d)
    VERTEX(o,6,w,-h,-d)
    VERTEX(o,7,w,h,-d)

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,xt,yt,xt,yt+ht-1,xt+wt-1,yt+ht-1,xt+wt-1,yt,t);
    QUAD(o,0,0,1,2,3,type,data);

    QUAD(o,1,4,5,6,7,COLORED,col);
    QUAD(o,2,7,6,1,0,COLORED,col);
    QUAD(o,3,3,2,5,4,COLORED,col);
    QUAD(o,4,7,0,3,4,COLORED,col);
    QUAD(o,5,1,6,5,2,COLORED,col);

    o3d_initTNormals(o);
    o3d_initVNormals(o);
    o3d_initBBox(o);

    return o;
}

OBJECT3D *o3d_createSpriteQ(int x,int y,int w,int h,TEXTURE3D* t,TRIANGLE3D_TYPE type)
{
    TEXTURE3D_DATA* data;
    OBJECT3D *o = o3d_createNewObject(4,1,0);
    int w2 = w>>1;
    int h2 = h>>1;

    VERTEX(o,0,w2,h2,0)
    VERTEX(o,1,w2,-h2,0)
    VERTEX(o,2,-w2,-h2,0)
    VERTEX(o,3,-w2,h2,0)

    data = (TEXTURE3D_DATA*)malloc(sizeof(TEXTURE3D_DATA));
    QTEXDATA(data,x,y,x,y+h-1,x+w-1,y+h-1,x+w-1,y,t);
    QUAD(o,0,0,1,2,3,type,data);

    o3d_initTNormals(o);
    o3d_initVNormals(o);
    o3d_initBBox(o);

    return o;
}

OBJECT3D *o3d_createQuad(int s,LCOLOR3D *col)
{
    OBJECT3D *o = o3d_createNewObject(4,1,0);

    VERTEX(o,0,s,s,0)
    VERTEX(o,1,s,-s,0)
    VERTEX(o,2,-s,-s,0)
    VERTEX(o,3,-s,s,0)

    QUAD(o,0,0,1,2,3,COLORED,col);

    o3d_initTNormals(o);
    o3d_initVNormals(o);
    o3d_initBBox(o);

    return o;
}
