#ifdef __GP32__
#include "gp32.h"
#include "fileio.h"
#else
#include "win32def.h"
#include "wingp.h"
#endif
#include "filezda.h"
#include <stdlib.h>
#include <string.h>

#include "dbg.h"
#include "gfx.h"
#include "tables.h"
#include "o3d.h"
#include "o3dbasic.h"

#include "levels.h"

#include "top.h"

u32 getTickCount();
void gp_Reset();
u32 gp_getRTC();

#ifdef __GP32__
int gp_getButton();
void gp_drawString16(int,int,char*,u16,u16*);
void gp_drawSString16(int,int,char*,u16,u16,u16*);
#define GEEPEE
#endif

//#define SHOW_FPS

#ifdef SHOW_FPS
unsigned long fps;
unsigned long fps_tick;
char fps_str[100];
extern int __drawn__;
extern int __vis__;
#endif

void gp_playMusic(void*,u32);
void gp_playSoundFX(s16*,u32,u8);
void gp_stopSoundFX();
void gp_stopMusic();
void gp_startSound();
void gp_stopSound();
char* gp_getMusicTitle();
char* gp_getMusicAuthor();

#ifdef __GP32__
#ifdef GEEPEE
#define FPS_CONST 256
#else
#define FPS_CONST 30 // 45
#endif
#else
#define FPS_CONST 80
#endif

#define CSIZE 256
#define WSIZE WIDTH

#define IDLE  0
#define LEFT  1
#define RIGHT 2
#define UP    4
#define DOWN  8

#define HORIZONTAL (LEFT|RIGHT)
#define VERTICAL (UP|DOWN)

enum
{
    BLOCK_NONE = 0,
    BLOCK_WALL,
    BLOCK_MIDWALL,
    BLOCK_GROUND,
    BLOCK_UNCOLORED,
    BLOCK_COLOR1,
    BLOCK_COLOR2,
    BLOCK_COLOR3,
    BLOCK_COLOR4
};

enum
{
    ANIM_IDLE = 0,
    ANIM_RUN,
    ANIM_SALUTE
};

SCENE3D     main_scene;
VECTOR3D*   p_light;
NODE3D*     p_main_scene;

extern u32 nflip;
extern u32 framebuffers[3];

#define CHAR_COUNT 3
const char* anims_name[CHAR_COUNT] = {"CUBEX_Alien","CUBEX_Yoshi","CUBEX_Dragon"};
NODE3D_DATA anims_data[CHAR_COUNT];
NODE3D_TYPE anims_type[CHAR_COUNT];
TEXTURE3DSET commonSet;

enum
{
    ROAR = 0,
    ENERGY,
    DOODLOOP,
    STOMP,
    GROWL,
    SND_COUNT
};
const char* snd_name[SND_COUNT] = {"roar","energy","doodloop","stomp","growl"};
s16* snd_data[SND_COUNT];
int snd_len[SND_COUNT];

#define BOX 14
#define SPC_BOX (BOX+2)
#define BOX_MASK 31

unsigned char level[13][13];
NODE3D *nlevel[13][13];

SCENE3D level_scene;
SCENE3D game_scene;
SCENE3D msg_scene;
SCENE3D pwd_scene;
SCENE3D text_scenes[5];

int nbb,nbb2;
int curlvl;
int curchar;
int player_x,player_y,player_dir,player_a,player_anim;
int line_dr;
int tuto_idx;
const char* tuto_text;

const int scramble[4] = {3,1,0,2};
const int bscramble[4] = {0,1,1,0};
char passwd[7];

#define PAL_SIZE 8
const unsigned char palette[PAL_SIZE][3] =
{
    {160, 160, 160}, // wall - grey
    {192, 192, 192}, // midwall - light grey
    {255, 192,  64}, // ground - orange
    {160,  96,   0}, // uncolored - brown
    {144, 224, 160}, // green
    { 64, 224, 224}, // blue
    {224, 112, 224}, // violet
    {224,  64,  64}
};

LCOLOR3D colors[PAL_SIZE];
TRIANGLE3D_DATA odata[PAL_SIZE];

OBJECT3D* p_cubes[5];
OBJECT3D* p_boxes[16+PAL_SIZE];
OBJECT3D* p_petobj;
TEXTURE3D p_textures[4][4];
NODE3D* p_player;
NODE3D* p_petnode;
NODE3D* p_level_scene;
NODE3D* p_game_scene;
NODE3D* p_msg_scene;
NODE3D* p_pwd_scene;
NODE3D* p_text_scenes[5];

NODE3D* p_push;
int push_x,push_y;
int petmode;

#define MUS_COUNT 21
const char* mus_names[MUS_COUNT] =
{
    "onceagain",
    "sharpness",
    "4urloader1",
    "4urloader3",
    "akscreen",
    "beatnick",
    "briefhistoryoftime",
    "chamberstitle",
    "chicago",
    "colorshock",
    "cuddlymenu",
    "cuddlyreset",
    "cuddlyscreen4",
    "ddreset",
    "forabyss",
    "noname",
    "odyssey",
    "phallanx",
    "stars",
    "thatsthewayitis",
    "thinktwice"
};
int mus_size[MUS_COUNT];
u8* music[MUS_COUNT];
int curmus;
int mus_cnt = 0;

SPRITE3D font[48];

unsigned short *backgrd;

void loadTextures(char* name,TEXTURE3DSET* tset)
{
    u8* data = loadZDAFile(name,"o3d",NULL);
    if (data==NULL)
        error(name);
    o3d_createTextures(data,1,tset);
    commonSet.set[2].trans = 0;
    FREE(data);
}

NODE3D_TYPE loadObject(char* name,NODE3D_DATA* nd,TEXTURE3D *tex)
{
    NODE3D_TYPE t;
    u8* data = loadZDAFile(name,"o3d",NULL);
    if (data==NULL)
        return UNKNOWN;
    t = o3d_createObject(data,tex,nd,1);
    FREE(data);
    return t;
}

u8* loadFile(char* name, char* ext, int* size)
{
    u8* data = loadZDAFile(name,ext,size);
    if (data==NULL)
        error(name);
    return data;
}

void loadCommon()
{
    int i;
    loadTextures("CUBEX_tex",&commonSet);
    for (i=0;i<CHAR_COUNT;++i)
        anims_type[i] = loadObject((char*)anims_name[i],&anims_data[i],commonSet.set);
}

void unloadCommon()
{
    int i;
    for (i=0;i<CHAR_COUNT;++i)
        o3d_destroyObject(&anims_data[i],anims_type[i]);
    o3d_destroyTextures(&commonSet);
    o3d_free();
    for (i=0;i<SND_COUNT;++i)
        FREE(snd_data[i]);
}

void createColor(LCOLOR3D *lcol, unsigned char r, unsigned char g, unsigned char b)
{
    lcol->col.r = r;
    lcol->col.g = g;
    lcol->col.b = b;
    o3d_registerColor(lcol);
}

unsigned char lumColor(int c, int i)
{
    return (unsigned char)((c*i)/255);
}

void createTexture(TEXTURE3D *t, LCOLOR3D *lcol, int id)
{
    int i;
    t->w = 23;
    t->h = 23;
    t->lcol = *lcol;
    t->pal = (COLOR3D*)malloc(9*sizeof(COLOR3D));
    t->gpal = (unsigned short*)malloc(9*sizeof(unsigned short));
    t->nbc = 8;
    t->pix = (unsigned char*)top_lum[id];
    for (i=0;i<9;++i)
    {
        t->pal[i].r = lumColor(lcol->col.r,i*32);
        t->pal[i].g = lumColor(lcol->col.g,i*32);
        t->pal[i].b = lumColor(lcol->col.b,i*32);
        t->gpal[i] = GPCOL(t->pal[i].r, t->pal[i].g, t->pal[i].b);
    }
    t->lpal = (unsigned short*)malloc(sizeof(unsigned short)*9*32);
    t->trans = -1;
}

void initO3D()
{
    int i,j;

    o3d_unregisterColors();

    o3d_initScene(&main_scene);
    o3d_initScene(&game_scene);
    o3d_initScene(&level_scene);
    o3d_initScene(&msg_scene);
    o3d_initScene(&pwd_scene);

    p_main_scene = o3d_init(&main_scene,2000,4000,128,640);
    o3d_setGDMin(-2000);
    o3d_setDepths(-2560,512);
    o3d_setShading(GOURAUD);
    p_light = o3d_setLightPos(NULL,MULT,MULT,MULT);
    o3d_setBorderColor(GPCOL(0,0,0),GPCOL(0,0,0));
    p_main_scene->pos.x = 0;
    p_main_scene->pos.y = 0;
    p_main_scene->pos.z = 0;
    p_main_scene->rot.rx = 0;
    p_main_scene->rot.ry = 0;
    p_main_scene->rot.rz = 0;

    p_petobj = o3d_createColoredBoxQ(5,5,5,&colors[0]);

    for (i=0;i<PAL_SIZE;++i)
    {
        createColor(&colors[i],palette[i][0],palette[i][1],palette[i][2]);
        p_boxes[i] = o3d_createColoredBoxQ(BOX,BOX,BOX,&colors[i]);
        odata[i].c = &colors[i]; 
    }

    for (i=0;i<4;++i)
    {
        for (j=0;j<4;++j)
        {
            createTexture(&p_textures[i][j],&colors[4+j],i);
            p_boxes[PAL_SIZE+i*4+j] = o3d_createMixedBoxQ(BOX,BOX,BOX,&colors[4+j],0,0,24,24,&p_textures[i][j],TEXTURED);
        }
    }

    o3d_setLightColor(80,80,80);
    o3d_setLightDefaut(24);
    o3d_lightTextures(commonSet.set,commonSet.count);
    o3d_lightTextures((TEXTURE3D*)p_textures,16);
}

void createFont()
{
    int i;
    //commonSet.set[3].trans = 0;
    for (i=0;i<48;i++)
    {
        font[i].dx = (i&0x7)<<5;
        font[i].dy = (i>>3)*25;
        font[i].w = 32;
        font[i].h = 25;
        font[i].t = &commonSet.set[2];
    }
}

void setBackground(int type,int color)
{
    int i,j,k,l = 0;
    unsigned short *gpal = commonSet.set[3].gpal;
    unsigned char *pix = commonSet.set[3].pix + ((type==10?8:type)<<4) + (color<<12);
    for (i=0;i<320;++i)
    {
        k = ((i&0xf)<<8);
        for (j=0;j<CSIZE;++j)
            backgrd[l++] = gpal[pix[(j&0xf)+k]];
    }
}

void backLine(int set, int dx, int dy)
{
    int iset = dx+set;
    while (iset>=WSIZE)
        iset-=WSIZE;
    memcpy((void*)(framebuffers[nflip]+(set*HEIGHT*2)),backgrd+(dy&0xf)+(iset*CSIZE),HEIGHT*2);
}

void exitApp()
{
    int i;
    for (i=0;i<PAL_SIZE+12;i++)
        o3d_freeObject(p_boxes[i], 1);

    gp_stopSound();
    for (i=0;i<MUS_COUNT;++i)
        free(music[i]);
    unloadCommon();
    free(backgrd);
    gp_Reset();
}

void playMusic()
{
    gp_stopMusic();
    gp_playMusic(music[curmus],mus_size[curmus]);
}

void removeText()
{
    if (p_pwd_scene != NULL)
    {
        o3d_clearScene(&pwd_scene);
        o3d_removeNode(&main_scene, p_pwd_scene);
        p_pwd_scene = NULL;
    }
    if (p_msg_scene != NULL)
    {
        o3d_clearScene(&msg_scene);
        o3d_removeNode(&main_scene, p_msg_scene);
        p_msg_scene = NULL;
    }
}

void addChar(SCENE3D* st,char c,int dx,int dy, int dz)
{
    int cc;
    if (c>='A' && c<='Z')
        cc = c-'A'+16;
    else if (c>='0' && c<='9')
        cc = c-'0';
    else
    {
        switch(c)
        {
            case ' ': return;
            case ':': cc=10; break;
            case ';': cc=11; break;
            case '+': cc=12; break;
            case ',': cc=13; break;
            case '-': cc=14; break;
            case '.': cc=15; break;
            case '!': cc=42; break;
            case '\"': cc=43; break;
            case '(': cc=44; break;
            case ')': cc=45; break;
            case '?': cc=46; break;
            case '/': cc=47; break;
            default : return;
        }
    }
    o3d_addSprite(st,&font[cc],dx,dy,dz);
}

NODE3D* addText(SCENE3D* s,SCENE3D* st,const char* text,int x,int y,int z)
{
    int i,dx=0,dy=0;
    NODE3D* p_scene = NULL;
    if (s!=NULL)
    {
        o3d_initScene(st);
        p_scene = o3d_addScene(s,st,x,y,z,0,0,0);
    }
    else
        o3d_clearScene(st);
    for (i=0;text[i]!=0;++i)
    {
        if (text[i] == 10)
        {
            dy-=27;
            dx=0;
            continue;
        }
        addChar(st,text[i],dx,dy,0);
        dx-=32;
    }
    return p_scene;
}

void initLevel()
{
    int i,j,k,y;
    NODE3D *n;

    o3d_clearScene(&main_scene);
    o3d_clearScene(&game_scene);
    o3d_clearScene(&level_scene);
    o3d_clearScene(&msg_scene);
    o3d_clearScene(&pwd_scene);

    if (curlvl >= 0)
    {
        player_x = startx[curlvl];
        player_y = starty[curlvl];
    }
    else
    {
        player_x = TUTO_STARTX;
        player_y = TUTO_STARTY;
    }
    player_a = 1;

    p_game_scene = o3d_addScene(&main_scene,&game_scene,0,0,-512,128,0,0/*PRECISION1-24*/);
    p_player = o3d_addAnimation(&game_scene,anims_data[curchar].a,0,BOX,0,0,0,0);
    p_petnode = o3d_addObject(&game_scene,p_petobj,0,BOX+36,0,0,64,64);
    p_petnode->hide = 1;
    p_player->cur_frame = p_player->data.a->anim_start[ANIM_IDLE];
    p_level_scene = o3d_addScene(&game_scene,&level_scene,0,0,0,0,0,0);
    if (curlvl>=0)
        p_msg_scene = addText(&main_scene, &msg_scene, "READY?", 80, -56, 128);
    else
        p_msg_scene = addText(&main_scene, &msg_scene, "TUTORIAL", 112, 80, 128);
    if (curlvl>0)
    {
        char buf[32];
        sprintf(buf,"PASSWORD:\n  %s",passwd);
        p_pwd_scene = addText(&main_scene, &pwd_scene, buf, 120, 96, 0);
    }

    line_dr = 0;

    p_level_scene->pos.x = (player_x-13)*SPC_BOX*2;
    p_level_scene->pos.y = -SPC_BOX;
    p_level_scene->pos.z = -player_y*SPC_BOX*2;

    if (curlvl>=0)
        memcpy(level,levels[curlvl],sizeof(level));
    else
        memcpy(level,tuto_level,sizeof(level));

    nbb = 0;
    for (i=0;i<13;++i)
    {
        for (j=0;j<13;++j)
        {
            nlevel[i][j] = NULL;
            k = level[i][j];
            if (k)
            {
                switch(k)
                {
                    case BLOCK_WALL:
                        y = (SPC_BOX*3)>>1;
                        break;
                    case BLOCK_MIDWALL:
                        y = SPC_BOX;
                        break;
                    case BLOCK_GROUND:
                        y = 0;
                        break;
                    default:
                        y = SPC_BOX<<1;
                }
                if (k>BLOCK_GROUND) // add ground
                {
                    o3d_addObject(&level_scene,p_boxes[BLOCK_GROUND-1],(13-j)*SPC_BOX*2,0,i*SPC_BOX*2,0,0,0);
                    if (k>=BLOCK_ANCHOR)
                    {
                        int kol = (k&0xf)-5;
                        switch(k&0xf0)
                        {
                            case BLOCK_ANCHOR: k = 0; break;
                            case BLOCK_PUZZLE: k = 1; break;
                            case BLOCK_PET:    k = 2; break;
                            case BLOCK_JOKER:  k = 3; break;
                        }
                        k = PAL_SIZE+k*4+kol+1;
                    }
                    n = o3d_addObject(&level_scene,p_boxes[k-1],(13-j)*SPC_BOX*2,y,i*SPC_BOX*2,384,0,0);
                    if (k>BLOCK_UNCOLORED)
                        nbb++;
                }
                else
                    n = o3d_addObject(&level_scene,p_boxes[k-1],(13-j)*SPC_BOX*2,y,i*SPC_BOX*2,0,0,0);
                if (k != BLOCK_GROUND)
                    nlevel[i][j] = n;
            }
        }
    }
    nbb2 = nbb;
    player_anim = ANIM_IDLE;
    p_push = NULL;
    petmode = 0;
    i = gp_getRTC();
    setBackground(i&15,(i>>4)&7);
}

void removeBlock(int x,int y)
{
    if (level[y][x]&BLOCK_PET)
    {
        p_petnode->hide = 0;
        petmode = level[y][x]&0xf;
        p_petnode->overload = &odata[petmode-1];
    }
    level[y][x] = BLOCK_GROUND;
    o3d_removeNode(&level_scene,nlevel[y][x]);
    nlevel[y][x] = NULL;
    nbb2 -= 1;
}

unsigned int checkBlock(int x,int y)
{
    if (level[push_y][push_x]==BLOCK_UNCOLORED)
        return 0;
    if ((level[y][x]&0xf)==(level[push_y][push_x]&0xf))
    {
        if (((level[push_y][push_x]&BLOCK_PUZZLE)!=0 && (level[y][x]&BLOCK_PUZZLE)!=0) ||
            ((level[push_y][push_x]&BLOCK_PUZZLE)==0 && (level[y][x]&BLOCK_PUZZLE)==0))
        {
            removeBlock(x,y);
            return 1;
        }
        return 0;
    }
    if ((level[push_y][push_x]&BLOCK_JOKER)!=0 && (level[y][x]&BLOCK_JOKER)!=0)
    {
        removeBlock(x,y);
        return 1;
    }
    return 0;
}

void playSoundFX(u16 id,u8 loop)
{
    gp_playSoundFX(snd_data[id]+22,snd_len[id]-44,loop);
}

void checkPos()
{
    static int dx[4] = { 0,-1,1,0};
    static int dy[4] = {-1, 0,0,1};
    int i;
    unsigned int b = 0;
    player_dir = IDLE;
    if (p_push==NULL)
        return;
    if ((p_petnode->hide==0) && (petmode==(level[push_y][push_x]&0xf)))
        b = 1;
    for (i=0;i<4;++i)
        b |= checkBlock(push_x+dx[i],push_y+dy[i]);
    if (b)
    {
        removeBlock(push_x,push_y);
        playSoundFX(ENERGY,0);
    }
    p_push = NULL;
}

void updatePos()
{
    if ((player_dir&HORIZONTAL) && ((p_level_scene->pos.x&BOX_MASK)==0))
    {
        player_x += (player_dir==LEFT ? -1 : 1);
        checkPos();
    }
    else if ((player_dir&VERTICAL) && ((p_level_scene->pos.z&BOX_MASK)==0))
    {
        player_y += (player_dir==UP ? -1 : 1);
        checkPos();
    }
}

void changeMusic(int key)
{
    if (mus_cnt>0)
        return;
    else if (key&BUTTON_L)
    {
        if (--curmus<0)
            curmus=MUS_COUNT-1;
        mus_cnt = 60;
        playMusic();
    }
    else if (key&BUTTON_R)
    {
        if (++curmus==MUS_COUNT)
            curmus = 0;
        mus_cnt = 60;
        playMusic();
    }
}

int waitKeyDown()
{
    int key;
    while ((key = gp_getButton())==0)
#ifdef __ARM_ADS
        __asm { nop }
#else
        asm volatile("nop\n");
#endif
    return key;
}

void waitKeyUp()
{
    int key;
    while ((key = gp_getButton())!=0)
#ifdef __ARM_ADS
        __asm { nop }
#else
        asm volatile("nop\n");
#endif
}

int handlePlayer(int ingame)
{
    int dx,dz,b=0,sx,sy;
    int key = gp_getButton();

    p_petnode->rot.rx = (p_petnode->rot.rx+4)&PMASK;

    if (key&BUTTON_A)
    {
        p_game_scene->rot.rx = (p_game_scene->rot.rx+2)&PMASK;
        line_dr -= 4;
    }
    else if (key&BUTTON_B)
    {
        p_game_scene->rot.rx = (p_game_scene->rot.rx-2)&PMASK;
        line_dr += 4;
    }
    else if ((key&BUTTON_START) && (key&BUTTON_SELECT))
    {
        waitKeyUp();
        return 1;
    }

    changeMusic(key);

    if ((p_msg_scene==NULL) && // welldone/ready/go...
        (player_dir==IDLE) &&
        ((p_level_scene->pos.x&BOX_MASK)==0) &&
        ((p_level_scene->pos.z&BOX_MASK)==0))
    {
        if (!ingame)
        {
            if (key&BUTTON_START)
            {
                nbb = 0;
                waitKeyUp();
                return 0;
            }
            key = tuto_moves[tuto_idx++];
            tuto_text = tuto_texts[tuto_moves[tuto_idx++]];
            b = 1;
        }
        if (key&BUTTON_LEFT)
        {
            player_dir = LEFT;
            p_player->rot.ry = 128;
        }
        else if (key&BUTTON_RIGHT)
        {
            player_dir = RIGHT;
            p_player->rot.ry = 384;
        }
        else if (key&BUTTON_DOWN)
        {
            player_dir = DOWN;
            p_player->rot.ry = 0;
        }
        else if (key&BUTTON_UP)
        {
            player_dir = UP;
            p_player->rot.ry = 256;
        }
        if (ingame)
        {
            if (key&BUTTON_START)
                nbb = -1;
            else
                b = 1;
        }
    }
    else if (nbb2==0 && (key&BUTTON_START))
    {
        nbb = 0;
        waitKeyUp();
        return 0;
    }

    switch(player_dir)
    {
        case LEFT:  dx = -1; dz =  0; break;
        case RIGHT: dx =  1; dz =  0; break;
        case DOWN:  dx =  0; dz = -1; break;
        case UP:    dx =  0; dz =  1; break;
        default:
        case IDLE:  dx =  0; dz =  0; break;
    }

    sy = player_y - dz;
    sx = player_x + dx;
    if ((b==0) || (level[sy][sx] >= BLOCK_GROUND))
    {
        if (b==1)
        {
            p_push = nlevel[sy][sx];
            if (p_push!=NULL)
            {
                push_x = sx + dx;
                push_y = sy - dz;
                if ((level[push_y][push_x]==BLOCK_GROUND) &&
                    ((level[sy][sx]&BLOCK_ANCHOR) == 0))
                {
                    level[push_y][push_x]  = level[sy][sx];
                    nlevel[push_y][push_x] = nlevel[sy][sx];
                    level[sy][sx]  = BLOCK_GROUND;
                    nlevel[sy][sx] = NULL;
                }
                else
                {
                    p_push = NULL;
                    player_dir = IDLE;
                    dx = dz = 0;
                }
            }
        }
        p_level_scene->pos.x += (dx<<2);
        p_level_scene->pos.z += (dz<<2);
        if (p_push!=NULL)
        {
            p_push->pos.x -= (dx<<2);
            p_push->pos.z -= (dz<<2);
        }
    }
    else
        player_dir = IDLE;

    player_anim = player_dir==IDLE ? ANIM_IDLE : ANIM_RUN;

    if (p_player->cur_frame<p_player->data.a->anim_start[player_anim] ||
        p_player->cur_frame>=p_player->data.a->anim_end[player_anim])
    {
        p_player->cur_frame = p_player->data.a->anim_start[player_anim];
    }
    else if (--player_a==0)
    {
        player_a = 2;
        if (++p_player->cur_frame>=p_player->data.a->anim_end[player_anim])
        {
            p_player->cur_frame = p_player->data.a->anim_start[player_anim];
        }
    }
    return 0;
}

//#define SPLASHSCREEN

void loadData()
{
    int i;
#ifdef SPLASHSCREEN
    u8* data;
#endif
    
    clearScreen();

    if (openZDA("DATACUBX.ZDA"))
        error("BAD COMMON DATA FILE !!!");

#ifdef SPLASHSCREEN
    data = loadFile("revival","bin",NULL);
    memcpy((void*)framebuffers[nflip],(void*)data,SIZE<<1);
    gp_drawSString16(130,229,"LOADING...",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
#else
    gp_drawSString16(116,119,"LOADING...",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
#endif
    flip(1);

#ifndef GEEPEE
     gp_startSound();
#endif

    curmus = 0;
    for (i=0;i<MUS_COUNT;++i)
    {
        music[i] = loadFile((char*)mus_names[i],"ym",&mus_size[i]);
        if (i==0)
            playMusic();
    }
    for (i=0;i<SND_COUNT;++i)
        snd_data[i] = (s16*)loadFile((char*)snd_name[i],"wav",&snd_len[i]);

    loadCommon();
    closeZDA();

    createFont();

#ifdef SPLASHSCREEN
    memcpy((void*)framebuffers[nflip],(void*)data,SIZE<<1);
    FREE(data);
    gp_drawSString16(68,229,"PRESS A BUTTON TO START",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
    flip(1);
    waitKeyDown();
#endif
    waitKeyUp();
}

void musicInfos()
{
    char buf[64];
    char *title = gp_getMusicTitle();
    sprintf(buf,"Music: %s",((title==NULL)||(title[0]==0))?mus_names[curmus]:title);
    gp_drawSString16(4,219,buf,COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
    sprintf(buf,"By %s",gp_getMusicAuthor());
    gp_drawSString16(4,229,buf,COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
}

/**
 * Enter passwd screen
 */
void enterpass()
{
    u32 tick,last = getTickCount();
    volatile int key;
    int a=0,b=0,i=12,j=0,k,dx=0;
    int idx = 0;

    o3d_clearScene(&main_scene);
    p_text_scenes[0] = addText(&main_scene, &text_scenes[0], "  ENTER\nPASSWORD:", 120, 96, 0);
    p_text_scenes[1] = addText(&main_scene, &text_scenes[1], "ABCDEFGHIJKLMNOPQRSTUVWXYZ", 384, 0, 0);
    p_text_scenes[2] = addText(&main_scene, &text_scenes[2], passwd, 64, -56, 0);
    p_text_scenes[1]->data.s->nodes[i]->pos.z = 128;
    p_text_scenes[1]->data.s->nodes[i]->pos.y = -8;
#ifndef __GP32__
    setBackground(last&15,0);
#else
    k = gp_getRTC();
    setBackground(k&15,(k>>4)&7);
#endif
    key = 0;
    mus_cnt = 0;
    while ((key&BUTTON_START)==0 && idx<5)
    {
        tick = getTickCount();
#ifdef __GP32__
        if ((tick-last)<FPS_CONST)
            continue;
#else
        syncFPS(last,tick,FPS_CONST);
#endif
        last = tick;
        if (mus_cnt)
            mus_cnt-=1;
        if (j==0)
        {
            key = gp_getButton();
            if (key!=0)
            {
                j = 8;
                dx = 0;
                k = idx;
                if (key & (BUTTON_UP|BUTTON_LEFT))
                {
                    dx = i>0?-4:0;
                    i = i>0?i-1:0;
                }
                else if (key & (BUTTON_DOWN|BUTTON_RIGHT))
                {
                    dx = i<25?4:0;
                    i = i+(i<25?1:0);
                }
                else if (key & BUTTON_A)
                    passwd[idx++] = i+'A';
                else if ((key & BUTTON_B) && idx>0)
                    passwd[--idx] = '.';
                if (k!=idx)
                    addText(NULL, &text_scenes[2], passwd, 0, 0, 0);
            }
            else
                dx = 0;
            changeMusic(key);
        }
        else
        {
            j -= 1;
            if (dx!=0)
            {
                p_text_scenes[1]->pos.x += dx;
                for (k=(i>0?i-1:0);(k<26)&&(k<i+2);++k)
                {
                    if (k!=i && p_text_scenes[1]->data.s->nodes[k]->pos.z > 0)
                    {
                        p_text_scenes[1]->data.s->nodes[k]->pos.z -= 16;
                        p_text_scenes[1]->data.s->nodes[k]->pos.y += 1;
                    }
                    else if (k==i && p_text_scenes[1]->data.s->nodes[k]->pos.z < 128)
                    {
                        p_text_scenes[1]->data.s->nodes[k]->pos.z += 16;
                        p_text_scenes[1]->data.s->nodes[k]->pos.y -= 1;
                    }
                }
            }
        }

        for (k=0;k<WSIZE;++k)
            backLine(k,(MULT+tsin[a]+MULT+tcos[b])>>6,(MULT+tcos[a]+MULT+tsin[(a*3)&PMASK])>>6);
        a=(a+4)&PMASK;
        b=(b+1)&PMASK;

        o3d_computeFrame((u16*)framebuffers[nflip],0);
        musicInfos();
#ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
    waitKeyUp();
}

int checkpass()
{
    // check passwd validity
    int i=0,j=0,k,a;
    for (k=0;k<4;++k)
    {
        a=(passwd[scramble[3-k]]-'A')^bscramble[3-k];
        i+=a;
        j=j*26+a;
    }

    // Verify validation key
    if ((25-(i%26)) != (passwd[4]-'A'))
        return 0;

    if (j&0x40000)
        j = ((j&1) ? (j>>2) : (j>>1)) & 0xffff;

    // Get data from bits
    curlvl= (j>>5)&0x1f;
    curchar = (j>>4)&0x1;

    // check data validity
    if (curlvl>=LEVEL_COUNT)
        return 0;

    return 1;
}

/**
 * Generate password
 */
void genpass()
{
    unsigned int a;
    int s=0,i,j;

    // Put data on 16 bits
    a  = curlvl<<5;
    a |= curchar<<4;

    i = gp_getRTC()&3;
    switch(i)
    {
        case 2:
            if (((a<<2)|0x40001)<456976)
            {
                a = (a<<2)|0x40001;
                break;
            }
        case 1:
            a = (a<<1)|0x40000;
        default:
            break;
    }

    // Transform data to letters and compute validation key
    for (i=0;i<4;i++)
    {
        j = a%26;
        s += j;
        passwd[scramble[i]] = (j^bscramble[i]) + 'A';
        a /= 26;
    }
    passwd[4] = (25-(s%26))+'A';
    passwd[5] = 0;
}

/**
 * Tutorial
 */
void tutorial()
{
    u32 last = getTickCount(), tick;
    u32 t0 = 0;
    int t;
    int save_rx=0;
    char banner[41];
    int set = 0, dx = 0, dy = 0;
    int started = 0, cnt = 1024;
#ifdef SHOW_FPS
    int fps_sum = 0;
    fps_str[0] = 0;
    fps = 0;
#endif

    curlvl = -1;
    initLevel();
    banner[0] = 0;
    t = 0;
    player_dir = IDLE;
    tuto_idx = 0;
    tuto_text = NULL;

    mus_cnt = 0;
    curmus = gp_getRTC()%MUS_COUNT;
    playMusic();

    while(1)
    {
        if (set==0)
        {
            dx = ((640+p_level_scene->pos.x)>>2);
            dy = ((352+p_level_scene->pos.z+line_dr)>>2);
        }
        tick = getTickCount();
#ifdef __GP32__
        if ((tick-last)<FPS_CONST)
        {
            if (set<WSIZE)
                backLine(set++, dx, dy);
            continue;
        }
#else
        syncFPS(last,tick,FPS_CONST);
#endif
        last = tick;
        if (mus_cnt)
            mus_cnt-=1;
        for (;set<WSIZE;++set)
            backLine(set, dx, dy);
        set = 0;

        if (started)
        {
            updatePos();

            if (nbb2==0)
            {
                if (p_msg_scene==NULL)
                    p_msg_scene = addText(&main_scene, &msg_scene, "PLAY!", 64, 64, 128);
                player_anim = ANIM_IDLE;
            }
            if (handlePlayer(0))
                return;
            if (nbb<=0)
                return;
            else if (nbb2>0)
            {
                tick = getTickCount();
#ifdef GEEPEE
                t = ((tick-t0)/3000);
#else
                t = ((tick-t0)/1000);
#endif
//                sprintf(banner,"Tutorial  Time:%03d",t);
                if (tuto_text!=NULL)
                    sprintf(banner,"%s",tuto_text);
            }
        }
        else // Level intro
        {
            cnt-=16;
            if (!cnt)
            {
                removeText();
                started = 1;
                t0 = getTickCount();
            }
            else if (cnt<512)
            {
                p_game_scene->pos.z = -cnt;
                p_game_scene->rot.rx = 96+(save_rx-128)+(cnt>>4);
            }
            else if (cnt==512)
            {
                banner[0] = 0;
                removeText();
                save_rx = p_game_scene->rot.rx; 
                p_msg_scene = addText(&main_scene, &msg_scene, "LEARN!", 80, -56, 128);
            }
            handlePlayer(0);
        }

        o3d_computeFrame((u16*)framebuffers[nflip],0);

        if (banner[0])
        {
            gp_drawSString16(160-(strlen(banner)*4),4,banner,COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
            musicInfos();
        }
        else
        {
            gp_drawSString16(4,217,"L or R to change the music",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
            gp_drawSString16(4,227,"A or B to rotate the board",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
        }

#ifdef SHOW_FPS
        if (tick-fps>1000)
        {
            sprintf(fps_str,"fps=%ld %ld %d %d",fps_tick,fps_sum/fps_tick,p_level_scene->pos.z,__vis__);
            fps_tick = 0;
            fps_sum = 0;
            fps = tick;
        }
        else
        {
            fps_tick++;
            fps_sum+=__drawn__;
        }
        gp_drawString16(5,15,fps_str,COLOR_WHITE,(u16*)framebuffers[nflip]);
#endif

#ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
}

/**
 * Game main loop
 */
int game()
{
    u32 last = getTickCount(), tick;
    u32 t0 = 0;
    int t;
    int save_rx=0;
    char banner[41];
    int set = 0, dx = 0, dy = 0;
    int started = 0, cnt = 1024;
#ifdef SHOW_FPS
    int fps_sum = 0;
    fps_str[0] = 0;
    fps = 0;
#endif

    initLevel();
    banner[0] = 0;
    t = 0;
    player_dir = IDLE;

    mus_cnt = 0;
    curmus = gp_getRTC()%MUS_COUNT;
    playMusic();

    while(1)
    {
        if (set==0)
        {
            dx = ((640+p_level_scene->pos.x)>>2);
            dy = ((352+p_level_scene->pos.z+line_dr)>>2);
        }
        tick = getTickCount();
#ifdef __GP32__
        if ((tick-last)<FPS_CONST)
        {
            if (set<WSIZE)
                backLine(set++, dx, dy);
            continue;
        }
#else
        syncFPS(last,tick,FPS_CONST);
#endif
        last = tick;
        if (mus_cnt)
            mus_cnt-=1;
        for (;set<WSIZE;++set)
            backLine(set, dx, dy);
        set = 0;

        if (started)
        {
            updatePos();

            if (nbb2==0)
            {
                if (p_msg_scene==NULL)
                    p_msg_scene = addText(&main_scene, &msg_scene, "WELL\nDONE!", 48, 64, 128);
                player_anim = ANIM_IDLE;
            }
            if (handlePlayer(1))
                return 0;
            if (nbb<=0)
            {
                if (nbb==0 && ++curlvl==LEVEL_COUNT)
                    return 1;
                cnt = 1024;
                banner[0] = 0;
                started = 0;
                genpass();
                initLevel();
            }
            else if (nbb2>0)
            {
                tick = getTickCount();
#ifdef GEEPEE
                t = ((tick-t0)/3000);
#else
                t = ((tick-t0)/1000);
#endif
                sprintf(banner,"Stage:%02d Time:%03d",curlvl+1,t);
            }
        }
        else // Level intro
        {
            cnt-=16;
            if (!cnt)
            {
                removeText();
                started = 1;
                t0 = getTickCount();
            }
            else if (cnt<512)
            {
                p_game_scene->pos.z = -cnt;
                p_game_scene->rot.rx = 96+(save_rx-128)+(cnt>>4);
            }
            else if (cnt==512)
            {
                banner[0] = 0;
                removeText();
                save_rx = p_game_scene->rot.rx; 
                p_msg_scene = addText(&main_scene, &msg_scene, "GO!", 32, -56, 128);
            }
            handlePlayer(1);
        }

        o3d_computeFrame((u16*)framebuffers[nflip],0);

        if (banner[0])
        {
            gp_drawSString16(92,4,banner,COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
            musicInfos();
        }
        else
        {
            gp_drawSString16(4,217,"L or R to change the music",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
            gp_drawSString16(4,227,"A or B to rotate the board",COLOR_WHITE,COLOR_BLACK,(u16*)framebuffers[nflip]);
        }

#ifdef SHOW_FPS
        if (tick-fps>1000)
        {
            sprintf(fps_str,"fps=%ld %ld %d %d",fps_tick,fps_sum/fps_tick,p_level_scene->pos.z,__vis__);
            fps_tick = 0;
            fps_sum = 0;
            fps = tick;
        }
        else
        {
            fps_tick++;
            fps_sum+=__drawn__;
        }
        gp_drawString16(5,15,fps_str,COLOR_WHITE,(u16*)framebuffers[nflip]);
#endif

#ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
    return 0;
}

/**
 * Character selection
 */
void gp32Way()
{
    int i, b = 0, loop = 0;
    volatile int key;
    u32 last = getTickCount(), tick;
    NODE3D *dragon;
    int cur_a = ANIM_RUN;
#ifdef SHOW_FPS
    int fps_sum = 0;
    fps_str[0] = 0;
    fps = 0;
#endif

    gp_stopSoundFX();
    gp_stopMusic();

    o3d_clearScene(&main_scene);
    dragon = o3d_addAnimation(&main_scene,anims_data[2].a,0,8,-1024,0,0,0);
    dragon->cur_frame = dragon->data.a->anim_start[cur_a];

    i = key = 0;
    player_a = 1;
    mus_cnt = 0; 
    playSoundFX(STOMP,1);

    while ((loop<5) && (key&(BUTTON_A|BUTTON_START))==0)
    {
        tick = getTickCount();
#ifdef __GP32__
        if ((tick-last)<FPS_CONST)
            continue;
#else
        syncFPS(last,tick,FPS_CONST);
#endif
        last = tick;
        if (mus_cnt)
            mus_cnt-=1;
        if (i==0)
        {
            key = gp_getButton();
            if (key!=0)
            {
                i = 12;
            }
        }
        else
            i -= 1;

        o3d_computeFrame((u16*)framebuffers[nflip],1);

        if (--player_a==0)
        {
            player_a = 2;
            dragon->cur_frame += 1;
            if (dragon->cur_frame>=dragon->data.a->anim_end[cur_a])
            {
                switch(cur_a)
                {
                    case ANIM_RUN:
                        if (dragon->pos.z>=256)
                        {
                            cur_a = ANIM_IDLE;
                            playSoundFX(GROWL,0);
                        }
                        break;
                    case ANIM_IDLE:
                        cur_a = ANIM_SALUTE;
                        b = 1;
                        playSoundFX(ROAR,0);
                        loop += 1;
                        break;
                    case ANIM_SALUTE:
                        cur_a = ANIM_IDLE;
                        loop += 1;
                        break;
                }
                dragon->cur_frame = dragon->data.a->anim_start[cur_a];
            }
        }
        if (cur_a==ANIM_RUN)
        {
            dragon->pos.z += 24;
            if (dragon->pos.z>=256)
                player_a = 1;
        }
        if (b)
            gp_drawSString16(40,224,"GP32: THE WAY YOU SHOULD PLAY!",COLOR_WHITE,COLOR_RED,(u16*)framebuffers[nflip]);

#ifdef SHOW_FPS
        if (tick-fps>1000)
        {
            sprintf(fps_str,"fps=%ld %ld %d",fps_tick,fps_sum/fps_tick,__vis__);
            fps_tick = 0;
            fps_sum = 0;
            fps = tick;
        }
        else
        {
            fps_tick++;
            fps_sum+=__drawn__;
        }
        gp_drawString16(5,15,fps_str,COLOR_WHITE,(u16*)framebuffers[nflip]);
#endif

        #ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
    waitKeyUp();
    gp_stopSoundFX();
}

/**
 * Character selection
 */
int selectCharacter()
{
    int i, j, k, set = 0, dx = 0, dy = 0;
    int a = 0, b = 0;
    volatile int key;
    u32 tick, last = getTickCount();
    NODE3D *p_char[CHAR_COUNT-1];

    o3d_clearScene(&main_scene);
    for (i=0;i<CHAR_COUNT-1;++i)
    {
        p_char[i] = o3d_addAnimation(&main_scene,anims_data[i].a,32-i*64,-16,128,0,0,0);
        p_char[i]->cur_frame = p_char[i]->data.a->anim_start[ANIM_IDLE];
    }
    addText(&main_scene, &msg_scene, "  SELECT\nCHARACTER:", 136, 80, 0);
#ifndef __GP32__
    setBackground(last&15,0);
#else
    k = gp_getRTC();
    setBackground(k&15,(k>>4)&7);
#endif

    i = j = key = 0;
    player_a = 1;
    mus_cnt = 0; 
    while ((key&(BUTTON_A|BUTTON_START))==0)
    {
        tick = getTickCount();
        if (set==0)
        {
            dx = (MULT+tsin[a]+MULT+tcos[b])>>6;
            dy = (MULT+tcos[a]+MULT+tsin[(a*3)&PMASK])>>6;
        }
#ifdef __GP32__
        if ((tick-last)<FPS_CONST)
        {
            if (set<WSIZE)
                backLine(set++, dx, dy);
            continue;
        }
#else
        syncFPS(last,tick,FPS_CONST);
#endif
        last = tick;

        for (;set<WSIZE;++set)
            backLine(set,dx,dy);
        set = 0;
        a = (a+4)&PMASK;
        b = (b+1)&PMASK;

        if (mus_cnt)
            mus_cnt-=1;
        if (j==0)
        {
            key = gp_getButton();
            if (key!=0)
            {
                if (key & (BUTTON_UP|BUTTON_LEFT))
                    i=(i==0?CHAR_COUNT-2:i-1);
                else if (key & (BUTTON_DOWN|BUTTON_RIGHT))
                    i=(i==CHAR_COUNT-2?0:i+1);
                j = 12;
            }
            changeMusic(key);
        }
        else
            j -= 1;

        o3d_computeFrame((u16*)framebuffers[nflip],0);
        musicInfos();

        for (k=0;k<CHAR_COUNT-1;++k)
        {
            if (i==k)
            {
                if (p_char[k]->pos.z<256)
                    p_char[k]->pos.z+=16;
                else
                    p_char[k]->rot.ry=(p_char[k]->rot.ry+8)&PMASK;
            }
            else if (i!=k)
            {
                if (p_char[k]->pos.z>128)
                    p_char[k]->pos.z-=16;
                if (p_char[k]->rot.ry>0)
                    p_char[k]->rot.ry=(p_char[k]->rot.ry+8)&PMASK;
            }
        }

        if (--player_a==0)
        {
            player_a = 2;
            for (k=0;k<CHAR_COUNT-1;++k)
            {
                p_char[k]->cur_frame += 1;
                if (p_char[k]->cur_frame>=p_char[k]->data.a->anim_end[ANIM_IDLE])
                    p_char[k]->cur_frame = p_char[k]->data.a->anim_start[ANIM_IDLE];
            }
        }

#ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
    gp_stopMusic();
    playSoundFX(DOODLOOP,0);
    waitKeyUp();
    o3d_clearScene(&main_scene);
    return i;
}

const char* scroller =
"    "\
"CUBEX BY (SOD)THOR"\
"  -  RELEASED FOR GP32 REVIVAL CODING COMPO "\
"  -  REMAKE BASED ON CUBERXTREME FROM DIGIS SEED ENTERTAINMENT (HTTP://WWW.DIGISONLINE.COM/CUBERXTREME/)"\
"  -  YM MUSIC LIBRARY BY LEONARD/OXYGEN (HTTP://LEONARD.OXG.FREE.FR)"\
"  -  MUSIC BY JESS, MADMAX, TAO AND SOME OTHER ATARI ST MUSIC AUTHORS"\
"  -  3D MODELS BY REED HAWKER (ALIEN) AND CHRIS SHELTON (YOSHI)";

/**
 * Game Intro
 */
void intro()
{
    int a = 0, b = 0, k, idx = 0, cpt, dax = 352;
    int set = 0, dx = 0, dy = 0;
    volatile int key;
    NODE3D *p_char;
    u32 tick, last = getTickCount();
#ifdef SHOW_FPS
    int fps_sum = 0;
    fps_str[0] = 0;
    fps = 0;
#endif

    curmus = 0;
    playMusic();

    o3d_clearScene(&main_scene);
    for (k=0;k<5;++k)
    {
        p_cubes[k] = o3d_createTexturedBoxQ(24,24,24,k*48,0,48,48,&commonSet.set[4],TEXTURED);
        o3d_addObject(&main_scene,p_cubes[k],128-(k*64)+dax,0,-128,0,0,128);
    }
    p_char = o3d_addAnimation(&main_scene,anims_data[1].a,-64,-32,32,0,0,0);
    p_char->cur_frame = p_char->data.a->anim_start[ANIM_SALUTE];
    o3d_clearScene(&msg_scene);
    p_msg_scene = o3d_addScene(&main_scene,&msg_scene,0,0,0,0,0,0);
#ifndef __GP32__
    setBackground(last&15,0);
#else
    k = gp_getRTC();
    setBackground(k&15,(k>>4)&7);
#endif
    key = 0;
    player_a = 1;
    cpt = 8;

    while ((key&(BUTTON_A|BUTTON_START))==0)
    {
        tick = getTickCount();
        if (set == 0)
        {
            dx = (MULT+tsin[a]+MULT+tcos[b])>>6;
            dy = (MULT+tcos[a]+MULT+tsin[(a*3)&PMASK])>>6;
        }
#ifdef __GP32__
        if ((tick-last)<FPS_CONST)
        {
            if (set<WSIZE)
                backLine(set++, dx, dy);
            continue;
        }
#else
        syncFPS(last,tick,FPS_CONST);
#endif
        last = tick;
        // animate background
        key = gp_getButton();
        while (set<WSIZE)
            backLine(set++, dx, dy);
        set = 0;

        // animate cubes
        for (k=0;k<5;++k)
        {
            main_scene.nodes[k]->rot.rx = a;
            main_scene.nodes[k]->rot.ry = (b+(k<<3)) & PMASK;
            main_scene.nodes[k]->pos.x = 128-(k*64)+dax;
            main_scene.nodes[k]->pos.y = (tsin[(a+(k<<4))&PMASK]>>4)+64;
            main_scene.nodes[k]->pos.z = (tcos[(b+(k<<3))&PMASK]>>4)-128;
        }
        if (dax>0)
            dax-=4;

        for (k=msg_scene.count-1;k>=0;--k)
        {
            if (msg_scene.nodes[k]->pos.x>208)
            {
                o3d_removeNodeAt(&msg_scene,k);
                continue;
            }
            msg_scene.nodes[k]->pos.x += 4;
            msg_scene.nodes[k]->pos.y = ((tsin[(msg_scene.nodes[k]->pos.x+PRECISION2)&PMASK]<<5)>>SHIFT);
            msg_scene.nodes[k]->pos.z = ((tcos[msg_scene.nodes[k]->pos.x&PMASK]<<7)>>SHIFT)-32;
        }
        if (--cpt==0)
        {
            addChar(&msg_scene,scroller[idx++],-208,0,0);
            cpt = 8;
            if (scroller[idx]==0)
                idx = 0;
        }

        a = (a+7)&PMASK;
        b = (b+3)&PMASK;

        if (--player_a==0)
        {
            player_a = 1;
            p_char->cur_frame += 1;
            if (p_char->cur_frame>=p_char->data.a->anim_end[ANIM_SALUTE])
                p_char->cur_frame = p_char->data.a->anim_start[ANIM_SALUTE];
        }
        p_char->rot.ry=(p_char->rot.ry+4)&PMASK;

        // show all
        o3d_computeFrame((u16*)framebuffers[nflip],0);
        musicInfos();

#ifdef SHOW_FPS
        if (tick-fps>1000)
        {
            sprintf(fps_str,"fps=%ld %ld %d",fps_tick,fps_sum/fps_tick,__vis__);
            fps_tick = 0;
            fps_sum = 0;
            fps = tick;
        }
        else
        {
            fps_tick++;
            fps_sum+=__drawn__;
        }
        gp_drawString16(5,15,fps_str,COLOR_WHITE,(u16*)framebuffers[nflip]);
#endif

#ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
    gp_stopMusic();
    playSoundFX(DOODLOOP,0);
    waitKeyUp();
}

/**
 * Game end
 */
void congrats()
{
    int a=0,b=0,k,cpt;
    int ang[5];
    volatile int key;
#ifndef __GP32__
    u32 tick,last;
#endif
    curmus = 0;
    playMusic();
    o3d_clearScene(&main_scene);
    for (k=0;k<5;++k)
    {
        p_cubes[k] = o3d_createTexturedBox(24,24,24,k*48,0,48,48,&commonSet.set[4],TEXTURED);
        ang[k] = (PRECISION1*k)/5;
        o3d_addObject(&main_scene,p_cubes[k],tcos[ang[k]]>>(SHIFT-7),tsin[ang[k]]>>(SHIFT-7),-128,0,0,128);
    }
    o3d_clearScene(&msg_scene);
    p_msg_scene = addText(&main_scene,&msg_scene,"COMPLETED\n\nCONGRATS!",128,32,0);
#ifndef __GP32__
    last = getTickCount();
    setBackground(last&15,0);
#else
    k = gp_getRTC();
    setBackground(k&15,(k>>4)&7);
#endif
    key = 0;
    cpt = 8;
    while ((key&(BUTTON_A|BUTTON_START))==0)
    {
#ifndef __GP32__
        tick = getTickCount();
        syncFPS(last,tick,FPS_CONST);
        last = tick;
#endif
        // animate background
        key = gp_getButton();
        for (k=0;k<WSIZE;++k)
            backLine(k,(MULT+tsin[a]+MULT+tcos[b])>>6,(MULT+tcos[a]+MULT+tsin[(a*3)&PMASK])>>6);

        // animate cubes
        for (k=0;k<5;++k)
        {
            main_scene.nodes[k]->rot.rx = a;
            main_scene.nodes[k]->rot.ry = (b+(k<<3)) & PMASK;
            ang[k] = (ang[k]+4)&PMASK;
            main_scene.nodes[k]->pos.x = tcos[ang[k]]>>(SHIFT-7);
            main_scene.nodes[k]->pos.y = tsin[ang[k]]>>(SHIFT-7);
        }

        a = (a+4)&PMASK;
        b = (b+1)&PMASK;

        // show all
        o3d_computeFrame((u16*)framebuffers[nflip],0);
        musicInfos();
#ifdef GEEPEE
        flip(1);
#else
        flip(0);
#endif
    }
    gp_stopMusic();
    playSoundFX(DOODLOOP,0);
    waitKeyUp();
}

/**
 * Main menu
 */
void menu()
{
    const char* entries[5] =
    {
        "NEW GAME",
        "CONTINUE",
        "PASSWORD",
        "CREDITS",
        "EXIT"
    };
    NODE3D* p_curs;
    int i,j,k;
    int a=0,b=0;
    volatile int key;
#ifndef __GP32__
    u32 tick,last;
#endif

    strcpy(passwd,".....");
    p_msg_scene = NULL;
    p_pwd_scene = NULL;

    while (1)
    {
        curmus = 1;
        playMusic();
        o3d_clearScene(&main_scene);
        o3d_clearScene(&msg_scene);
        p_msg_scene = o3d_addScene(&main_scene,&msg_scene,0,96,-64,0,0,0);
        for (i=0;i<5;++i)
            o3d_addObject(&msg_scene,p_cubes[i],128-(i*64),0,0,0,0,128);
        for (i=0;i<5;++i)
            p_text_scenes[i] = addText(&main_scene, &text_scenes[i], entries[i], 96, 40-i*30, 0);
        p_curs = o3d_addSprite(&main_scene,&font[14],132,40,32);
#ifndef __GP32__
        last = getTickCount();
        setBackground(last&15,0);
#else
        k = gp_getRTC();
        setBackground(k&15,(k>>4)&7);
#endif
        key = i = j = 0;
        mus_cnt = 0;
        p_text_scenes[i]->pos.z = 12;
        while ((key&(BUTTON_A|BUTTON_START))==0)
        {
#ifndef __GP32__
            tick = getTickCount();
            syncFPS(last,tick,FPS_CONST);
            last = tick;
#endif
            if (mus_cnt)
               mus_cnt-=1;
            if (j==0)
            {
                key = gp_getButton();
                if (key!=0)
                {
                    if (key & (BUTTON_UP|BUTTON_LEFT))
                        i=(i==0?4:i-1);
                    else if (key & (BUTTON_DOWN|BUTTON_RIGHT))
                        i=(i==4?0:i+1);
                    j = 12;
                }
                changeMusic(key);
            }
            else
                j -= 1;

            k = 40-i*30;
            if (p_curs->pos.y<k)
                p_curs->pos.y += 6;
            else if (p_curs->pos.y>k)
                p_curs->pos.y -= 6;

            for (k=0;k<5;++k)
            {
                if (k!=i && p_text_scenes[k]->pos.z>0)
                    p_text_scenes[k]->pos.z -= 4;
                else if (k==i && p_text_scenes[k]->pos.z<32)
                    p_text_scenes[k]->pos.z += 4;
            }

            p_msg_scene->pos.x = (tsin[a]>>3);
            for (k=0;k<5;++k)
                msg_scene.nodes[k]->rot.rz = (p_msg_scene->pos.x+128)&PMASK;
            for (k=0;k<WSIZE;++k)
                backLine(k,(MULT+tsin[a]+MULT+tcos[b])>>6,(MULT+tcos[a]+MULT+tsin[(a*3)&PMASK])>>6);
            a = (a+4)&PMASK;
            b = (b+1)&PMASK;

            o3d_computeFrame((u16*)framebuffers[nflip],0);
            musicInfos();
#ifdef GEEPEE
            flip(1);
#else
            flip(0);
#endif
        }
        playSoundFX(DOODLOOP,0);
        waitKeyUp();
        switch(i)
        {
            case 2:
                strcpy(passwd,".....");
                enterpass();
            case 1:
                if (checkpass())
                {
                    if (game())
                        congrats();
                }
                break;
            case 0:
                curchar = selectCharacter();
                tutorial();
                curlvl = 0;
                if (game())
                    congrats();
                break;
            case 4:
                exitApp();
            case 3:
                intro();
        }
        waitKeyUp();
    }
}

/**
 * Main loop between intro and main menu
 */
void mainLoop()
{
    loadData();
    initO3D();
    backgrd = (unsigned short*)malloc(WIDTH*CSIZE*2);
    gp32Way();
    intro();
    menu();
}
