#include "gp32.h"
#include "YmMusic.h"

#include "init1330.h"

#ifdef __ARM_ADS
extern void gp_installSWIIRQ(unsigned int, __irq void (*f)(void));
extern void gp_removeSWIIRQ(unsigned int, __irq void (*f)(void));
#else
extern void gp_installSWIIRQ(unsigned int, void (*f)(void));
extern void gp_removeSWIIRQ(unsigned int, void (*f)(void));
#endif
extern void gp_enableIRQ(void);
extern void gp_disableIRQ(void);
extern int  gp_getPCLK(void);

extern int   ymMusicCompute(void*,s16*,int);
extern void* ymMusicLoadMemory(void*,u32);
extern char* ymMusicGetLastError(void*);
extern void  ymMusicSetLoopMode(void*,int);
extern void  ymMusicPlay(void*);
extern void  ymMusicStop(void*);
extern void  ymMusicUnload(void*);

extern void error(char*);

#define SMP_SIZE 630 // 720 : 16000, 992 : 22050
#define SEGMENTSIZE (SMP_SIZE*4)

static CYmMusic* pMusic = NULL;
static s16* pSndFX0 = NULL; 
static s16* pSndFX = NULL; 
static u32 sndFXlen = 0;
static u32 sndFXloop = 0;

ymMusicInfo_t ymInfos;

void gp_stopMusic();
void gp_stopSoundFX();

#ifdef __ARM_ADS
void __irq MyTIMER4IRQ(void)
#else
static void MyTIMER4IRQ(void) __attribute__ ((interrupt ("IRQ")));

static void MyTIMER4IRQ(void)
#endif
{
    s16* segment; 
	rTCON = BITCLEAR(rTCON,20); // timer4 off
	segment = (s16*)gp_getRingSegment();
	if (pMusic!=NULL)
	    ymMusicCompute((void*)pMusic,segment,SMP_SIZE);
	if (pSndFX!=NULL)
	{
	    u32 i, len = SMP_SIZE>sndFXlen ? sndFXlen : SMP_SIZE;
	  
	    if (pMusic!=NULL) // mix
	    {
	        s32 smp;
            for (i=0;i<len;++i)
            {
                smp = *segment;
                smp = (smp + *(pSndFX++))>>1;
                *(segment++) = smp;
                *(segment++) = smp;
            }
	    }
	    else // only snd fx
	    {
	        u16 smp;
	        for (i=0;i<len;++i)
	        {
	            smp = *(pSndFX++);
	            *(segment++) = smp;
                *(segment++) = smp;
	        }
	        smp = 0;
	        for (;i<SMP_SIZE;++i)
	        {
	            *(segment++) = smp;
	            *(segment++) = smp;
	        }
	    }
	    sndFXlen-=len;
	    if (sndFXlen==0)
	    {
	        if (sndFXloop)
	        {
                pSndFX = pSndFX0;
                sndFXlen = sndFXloop;
	        }
	        else
	            pSndFX = NULL;
	    }
	}
	else if (pMusic==NULL)
	{
	    u32 i;
	    // clear
	    for (i=0;i<SMP_SIZE;++i)
	    {
            *(segment++) = 0;
            *(segment++) = 0;
	    }
	}
	rTCON = BITSET(rTCON,20); // timer4 on
}

static int initsound=1;

void gp_startSound()
{
	int pclk;
	pclk = gp_getPCLK();
	pclk/= 16;
	pclk/= 256;
	pclk/= 48;  // 51Hz timer, tested 33 66 133 Mhz

	if (initsound)
	{
		gp_initSound(YM_SMPRATE,16,SEGMENTSIZE*2); 
		initsound=0; 
	}
	gp_disableIRQ();
	rTCFG0 |= (0xFF<<8);   // Presacler for timer 2,3,4 = 256
	rTCFG1 |= (0x03<<16);  // timer4  1/16
	rTCNTB4 = (long)pclk;
	rTCON  = (0x1<<22) | (0x1<<20); // start timer4, auto reload
    pSndFX0 = NULL;
    sndFXloop = 0;
    sndFXlen = 0;
    pSndFX = NULL;
    pMusic = NULL;
    gp_clearRingbuffer();
	gp_installSWIIRQ(14,MyTIMER4IRQ);
	gp_enableIRQ();
}

//static 
void gp_stopSound()
{
    gp_stopSoundFX();
    gp_stopMusic();
	gp_disableIRQ();
	rTCON = BITCLEAR(rTCON,20); // timer4 off
	gp_removeSWIIRQ(14,MyTIMER4IRQ);
	gp_clearRingbuffer();
    gp_enableIRQ();
}

void gp_playMusic(void *data,u32 size)
{
    CYmMusic *music = (CYmMusic*) ymMusicLoadMemory(data,size);
    if (music != NULL)
    {
        ymMusicSetLoopMode((void*)music,YM_TRUE);
        ymMusicPlay((void*)music);
    }
    else
        error(ymMusicGetLastError(NULL));
    while((rTCON&(1<<20))==0)
#ifdef __ARM_ADS
        __asm { nop }
#else
        asm volatile("nop\n");
#endif
    rTCON = BITCLEAR(rTCON,20); // timer4 off
    pMusic = music;
    rTCON = BITSET(rTCON,20); // timer4 on
}

void gp_playSoundFX(s16 *data,u32 size,u8 loop)
{
    while((rTCON&(1<<20))==0)
#ifdef __ARM_ADS
        __asm { nop }
#else
        asm volatile("nop\n");
#endif
    rTCON = BITCLEAR(rTCON,20); // timer4 off
    sndFXlen = size>>1; // 16 bits mono
    pSndFX = data;
    if (loop)
    {
        sndFXloop = sndFXlen;
        pSndFX0 = pSndFX;
    }
    else
    {
        sndFXloop = 0;
        pSndFX0 = NULL;
    }
    rTCON = BITSET(rTCON,20); // timer4 on
}

char* gp_getMusicTitle()
{
    return pMusic == NULL ? "none" : pMusic->pSongName;
}

char* gp_getMusicAuthor()
{
    return pMusic == NULL ? "none" : pMusic->pSongAuthor;
}

void gp_stopMusic()
{
    CYmMusic *music;
    if(pMusic==NULL)
        return;
    music = pMusic;
    while((rTCON&(1<<20))==0)
#ifdef __ARM_ADS
        __asm { nop }
#else
        asm volatile("nop\n");
#endif
    rTCON = BITCLEAR(rTCON,20); // timer4 off
    pMusic = NULL;
    rTCON = BITSET(rTCON,20); // timer4 on
    ymMusicStop((void*)music);
	ymMusicUnload((void*)music);
}

void gp_stopSoundFX()
{
    while((rTCON&(1<<20))==0)
#ifdef __ARM_ADS
        __asm { nop }
#else
        asm volatile("nop\n");
#endif
    rTCON = BITCLEAR(rTCON,20); // timer4 off
    pSndFX = NULL;
    sndFXlen = 0;
    pSndFX0 = NULL;
    sndFXloop = 0;
    rTCON = BITSET(rTCON,20); // timer4 on
}
