/*
 * GLESonGL implementation
 * Version:  1.0
 *
 * Copyright (C) 2003  David Blythe   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * DAVID BLYTHE BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#ifdef _MSC_VER
#define M_PI	3.14159265358979323846f
#define sinf	sin
#define	cosf	cos
#define sqrtf	sqrt
#endif

#define EXP_TSIZE 128
#define SIN_TSIZE 1024
#define RECIP_TSIZE 8
#define RSQRT_TSIZE 8

main(int argc, char* argv[]) {
    int i;
    int tsize;

    if (argc <= 1) exit(1);
    if (strcmp(argv[1], "-e") == 0) {
	tsize = EXP_TSIZE;
	for(i = 0; i < tsize; i++) {
	    float m = exp(-6.0*i/(tsize-1));
	    unsigned short x = m*0xffff;
	    printf("0x%04x, /* %f\t %f */\n", x, -6.0*i/(tsize-1), m);
	}
    } else if (strcmp(argv[1], "-s") == 0) {
	tsize = SIN_TSIZE;
	for(i = 0; i < tsize; i++) {
	    float a = M_PI*i/2.0/(tsize-1);
	    float m = sin(a);
	    unsigned short x = m*0xffff;
	    printf("0x%04x, /* %f\t %f */\n", x, a, m);
	}
    } else if (strcmp(argv[1], "-r") == 0) {
	tsize = RECIP_TSIZE;
	for(i = 0; i < tsize; i++) {
	    float a = .5f+i/(2.f*tsize);
	    float m = 1.0f/a;
	    unsigned short x = (unsigned)(m*0x10000)>>2;
	    printf("0x%04x, /* %f\t %f */\n", x, a, m);
	}
    } else if (strcmp(argv[1], "-q") == 0) {
	tsize = RSQRT_TSIZE;
	for(i = 0; i < tsize+1; i++) {
	    float a = .5f+(float)i/(2.f*tsize);
	    float m = 1.0f/sqrtf(a);
	    unsigned short x = (unsigned)(m*0x10000)>>1;
	    printf("0x%04x, /* %f\t %f */\n", x, a, m);
	}
    }
    return 0;
}
