/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klArray.cpp,v 1.2 2004/05/03 08:14:49 drgoldie Exp $
 *
\**************************************************************************/

#include "klArray.h"

void
klArray::getVec3(klVec3& nVec3, int nIdx) const
{
	int elementStride = size*getByteSizeFromDataType(type) + stride;
	int offset = nIdx*elementStride;
	const unsigned char* ptr = reinterpret_cast<const unsigned char*>(clientData)+offset;

	switch(type)
	{
	case GL_FIXED:
		nVec3.setFixed(0, *reinterpret_cast<const int*>(ptr));
		nVec3.setFixed(1, *reinterpret_cast<const int*>(ptr+4));
		nVec3.setFixed(2, size>2 ? *reinterpret_cast<const int*>(ptr+8) : 0);
		break;

	case GL_SHORT:
		nVec3.setInt(0, *reinterpret_cast<const short*>(ptr));
		nVec3.setInt(1, *reinterpret_cast<const short*>(ptr+2));
		nVec3.setInt(2, size>2 ? *reinterpret_cast<const short*>(ptr+4) : 0);
		break;

	case GL_INT:
		nVec3.setInt(0, *reinterpret_cast<const int*>(ptr));
		nVec3.setInt(1, *reinterpret_cast<const int*>(ptr+4));
		nVec3.setInt(2, size>2 ? *reinterpret_cast<const int*>(ptr+8) : 0);
		break;

	case GL_FLOAT:
		nVec3.setFloat(0, *reinterpret_cast<const float*>(ptr));
		nVec3.setFloat(1, *reinterpret_cast<const float*>(ptr+4));
		nVec3.setFloat(2, size>2 ? *reinterpret_cast<const float*>(ptr+8) : 0.0f);
		break;

	case GL_DOUBLE:
		nVec3.setDouble(0, *reinterpret_cast<const double*>(ptr));
		nVec3.setDouble(1, *reinterpret_cast<const double*>(ptr+8));
		nVec3.setDouble(2, size>2 ? *reinterpret_cast<const double*>(ptr+16) : 0.0);
		break;
	}
}


void
klArray::getVec4(klVec4& nVec4, int nIdx) const
{
	int elementStride = size*getByteSizeFromDataType(type) + stride;
	int offset = nIdx*elementStride;
	const unsigned char* ptr = reinterpret_cast<const unsigned char*>(clientData)+offset;

	switch(type)
	{
	case GL_FIXED:
		nVec4.setFixed(0, *reinterpret_cast<const int*>(ptr));
		nVec4.setFixed(1, *reinterpret_cast<const int*>(ptr+4));
		nVec4.setFixed(2, size>2 ? *reinterpret_cast<const int*>(ptr+8) : 0);
		nVec4.setFixed(3, size>3 ? *reinterpret_cast<const int*>(ptr+12) : 1<<16);
		break;

	case GL_SHORT:
		nVec4.setInt(0, *reinterpret_cast<const short*>(ptr));
		nVec4.setInt(1, *reinterpret_cast<const short*>(ptr+2));
		nVec4.setInt(2, size>2 ? *reinterpret_cast<const short*>(ptr+4) : 0);
		nVec4.setInt(3, size>3 ? *reinterpret_cast<const short*>(ptr+6) : 1);
		break;

	case GL_INT:
		nVec4.setInt(0, *reinterpret_cast<const int*>(ptr));
		nVec4.setInt(1, *reinterpret_cast<const int*>(ptr+4));
		nVec4.setInt(2, size>2 ? *reinterpret_cast<const int*>(ptr+8) : 0);
		nVec4.setInt(3, size>3 ? *reinterpret_cast<const int*>(ptr+12) : 1);
		break;

	case GL_FLOAT:
		nVec4.setFloat(0, *reinterpret_cast<const float*>(ptr));
		nVec4.setFloat(1, *reinterpret_cast<const float*>(ptr+4));
		nVec4.setFloat(2, size>2 ? *reinterpret_cast<const float*>(ptr+8) : 0.0f);
		nVec4.setFloat(3, size>3 ? *reinterpret_cast<const float*>(ptr+12) : 1.0f);
		break;

	case GL_DOUBLE:
		nVec4.setDouble(0, *reinterpret_cast<const double*>(ptr));
		nVec4.setDouble(1, *reinterpret_cast<const double*>(ptr+8));
		nVec4.setDouble(2, size>2 ? *reinterpret_cast<const double*>(ptr+16) : 0.0);
		nVec4.setDouble(3, size>3 ? *reinterpret_cast<const double*>(ptr+24) : 1.0);
		break;
	}
}


/*klArrayFetcher::klArrayFetcher(const klArray& nArray)
{
	setArray(nArray);
}*/


klArrayFetcher::klArrayFetcher()
{
	array = NULL;
	elementStride = 0;
}


void
klArrayFetcher::setArray(const klArray& nArray)
{
	array = &nArray;
	elementStride = nArray.size*getByteSizeFromDataType(nArray.type) + nArray.stride;

	switch(array->type)
	{
	case GL_UNSIGNED_BYTE:
		getVec3P = &klArrayFetcher::getVec3UnsignedByte;
		getVec4P = &klArrayFetcher::getVec4UnsignedByte;
		break;

	case GL_FIXED:
		getVec3P = &klArrayFetcher::getVec3Fixed;
		getVec4P = &klArrayFetcher::getVec4Fixed;
		break;

	case GL_SHORT:
		getVec3P = &klArrayFetcher::getVec3Short;
		getVec4P = &klArrayFetcher::getVec4Short;
		break;

	case GL_INT:
		getVec3P = &klArrayFetcher::getVec3Int;
		getVec4P = &klArrayFetcher::getVec4Int;
		break;

	case GL_FLOAT:
		getVec3P = &klArrayFetcher::getVec3Float;
		getVec4P = &klArrayFetcher::getVec4Float;
		break;

	case GL_DOUBLE:
		getVec3P = &klArrayFetcher::getVec3Double;
		getVec4P = &klArrayFetcher::getVec4Double;
		break;
	}
}
