/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/Base/klArray.h,v 1.5 2004/05/03 08:14:49 drgoldie Exp $
 *
\**************************************************************************/


#ifndef __KLARRAY_HEADERFILE__
#define __KLARRAY_HEADERFILE__

#include "klBase.h"

// Stores ogl like vertex arrays (color, normal, ...)
//
// Notice: klVec3 & klVec4 must be defined (typedef'd before)
//         getVec3 & getVec4 are quite unefficient. if you need to
//         fetch multiple elements use klArrayFetcher (see below).
//
struct klArray
{
	klArray()  {  clientData = 0;  enabled = false;  }

	const GLvoid	*clientData;
	bool			enabled;
	GLint			size;
	GLenum			type;
	GLsizei			stride;
	bool			clientDataChanged;

	void set(GLint nSize, GLenum nType, GLsizei nStride, const GLvoid *nClientData)
	{
		size = nSize;
		type = nType;
		stride = nStride;
		clientData = nClientData;
		clientDataChanged = true;
	}

	void getVec3(klVec3& nVec3, int nIdx) const;
	void getVec4(klVec4& nVec4, int nIdx) const;
};


// This class lets us fetch data from an array.
// in contrast to klArray::getVecX() this class determines
// as much data as possible at creation time and is therefore
// a lot faster (when fetching multiple elements from the
// same array)
class klArrayFetcher
{
public:
	//klArrayFetcher(const klArray& nArray);
	klArrayFetcher();

	void getVec3(klVec3& nVec3, int nIdx) const;
	void getVec4(klVec4& nVec4, int nIdx) const;

	void setArray(const klArray& nArray);

protected:
	typedef void (klArrayFetcher::* GETVEC3FUNC)(klVec3& nVec3, int nIdx) const;
	typedef void (klArrayFetcher::* GETVEC4FUNC)(klVec4& nVec4, int nIdx) const;

	void getVec3UnsignedByte(klVec3& nVec3, int nIdx) const;
	void getVec3Fixed(klVec3& nVec3, int nIdx) const;
	void getVec3Short(klVec3& nVec3, int nIdx) const;
	void getVec3Int(klVec3& nVec3, int nIdx) const;
	void getVec3Float(klVec3& nVec3, int nIdx) const;
	void getVec3Double(klVec3& nVec3, int nIdx) const;

	void getVec4UnsignedByte(klVec4& nVec4, int nIdx) const;
	void getVec4Fixed(klVec4& nVec4, int nIdx) const;
	void getVec4Short(klVec4& nVec4, int nIdx) const;
	void getVec4Int(klVec4& nVec4, int nIdx) const;
	void getVec4Float(klVec4& nVec4, int nIdx) const;
	void getVec4Double(klVec4& nVec4, int nIdx) const;

	int elementStride;
	const klArray* array;
	GETVEC3FUNC getVec3P;
	GETVEC4FUNC getVec4P;
};

#include "klArray.inl"

#endif //__KLARRAY_HEADERFILE__
