 /**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/Base/klProfiler.cpp,v 1.5 2004/08/22 13:42:13 wasserer Exp $
 *
\**************************************************************************/


#include "klMain.h"
#include "gl_definitions.h"

#include "klProfiler.h"

#ifdef _USE_PROFILING_

int mulCount = 0;
int divCount = 0;
int sqrtCount = 0;
int invSqrtCount = 0;
int trigCount = 0;

#endif

void
klProfiler::reset()
{
	_frame.reset();
	_tnl.reset();
	_lighting.reset();
	_clipping.reset();
	_cliptowindow.reset();
	_rasterprimitive.reset();
	_clear.reset();
	_trianglesetup.reset();
	_drawscanline.reset();
	_pixeltest.reset();
	_pixelwrite.reset();
}


klProfiler::Measurement*
klProfiler::getMes(MES nMes)
{
	switch(nMes)
	{
	case frame:
		return &_frame;
	case tnl:
		return &_tnl;
	case lighting:
		return &_lighting;
	case clipping:
		return &_clipping;
	case cliptowindow:
		return &_cliptowindow;
	case rasterprimitive:
		return &_rasterprimitive;
	case clear:
		return &_clear;
	case trianglesetup:
		return &_trianglesetup;
	case drawscanline:
		return &_drawscanline;
	case pixeltest:
		return &_pixeltest;
	case pixelwrite:
		return &_pixelwrite;
	}

	return NULL;
}


#ifdef _IS_WINDOWS_


void
klProfiler::Measurement::reset()
{
	sum.QuadPart = 0;
}


void
klProfiler::beginSection(Measurement& nM)
{
	QueryPerformanceCounter(&nM.secBegin);
}


void
klProfiler::endSection(Measurement& nM)
{
	QueryPerformanceCounter(&nM.secEnd);
	nM.sum.QuadPart += nM.secEnd.QuadPart - nM.secBegin.QuadPart;
}


float
klProfiler::getFraction(Measurement& nNom, Measurement& nDenom)
{
	long double nom = (long double)nNom.sum.QuadPart,
				denom = (long double)nDenom.sum.QuadPart;

	long double fract = nom/denom;
	return (float)fract;
}


float
klProfiler::getFraction(MES nNom, MES nDenom)
{
	Measurement* nom = getMes(nNom);
	Measurement* denom = getMes(nDenom);

	if(!nom || !denom)
		return 0.0f;

	return getFraction(*nom, *denom);
}


float
klProfiler::getTime(MES nMes)
{
	Measurement* mes = getMes(nMes);
	LARGE_INTEGER freq;

	if(!mes)
		return 0.0f;

	QueryPerformanceFrequency(&freq);

	long double ld_sum = (long double)mes->sum.QuadPart,
				ld_freq = (long double)freq.QuadPart;

	long double dt = ld_sum/ld_freq;
	return (float)dt;
}


#else _IS_WINDOWS_


// supply only stub methods for non-windows systems
//
void
klProfiler::Measurement::reset()
{
}


void
klProfiler::beginSection(Measurement& /*nM*/)
{
}


void
klProfiler::endSection(Measurement& /*nM*/)
{
}


float
klProfiler::getFraction(MES /*nNom*/, MES /*nDenom*/)
{
	return 0.0f;
}


float
klProfiler::getTime(MES /*nMes*/)
{
	return 0.0f;
}


#endif //_IS_WINDOWS_

