 /**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/Base/klProfiler.h,v 1.3 2004/04/19 13:51:54 wasserer Exp $
 *
\**************************************************************************/


#ifndef __KLPROFILER_HEADERFILE__
#define __KLPROFILER_HEADERFILE__

#ifdef _IS_WINDOWS_
  #include <windows.h>
#endif //_IS_WINDOWS_

class klProfiler
{
public:
	enum MES {
		frame = KL_PROFILE_FRAME,
		tnl = KL_PROFILE_TNL,
		lighting = KL_PROFILE_LIGHTING,
		clipping = KL_PROFILE_CLIPPING,
		cliptowindow = KL_PROFILE_CLIPTOWINDOW,
		rasterprimitive = KL_PROFILE_RASTERPRIMITIVE,
		clear = KL_PROFILE_CLEAR,
		trianglesetup = KL_PROFILE_TRIANGLESETUP,
		drawscanline = KL_PROFILE_DRAWSCANLINE,
		pixeltest = KL_PROFILE_PIXELTEST,
		pixelwrite = KL_PROFILE_PIXELWRITE,
	};

	struct Measurement {
#ifdef _IS_WINDOWS_
		LARGE_INTEGER secBegin, secEnd, sum;
#endif

		void reset();
	};

	Measurement _frame, _tnl, _lighting, _clipping,
				_cliptowindow, _rasterprimitive,
				_clear, _trianglesetup, _drawscanline,
				_pixeltest, _pixelwrite;

	void reset();
	void beginSection(Measurement& nM);
	void endSection(Measurement& nM);

	float getFraction(Measurement& nNom, Measurement& nDenom);
	float getFraction(MES nNom, MES nDenom);
	float getTime(MES nMes);

protected:
	Measurement* getMes(MES nMes);
};


//
// we use macros to simplify calls to the profiler
// and to turn profiling on/off
// define _USE_PROFILING_ in the project settings
// to activate profiling. note that this can reduce
// rendering speed a lot (especially if profiling
// per scanline
//
#ifdef _USE_PROFILING_
  #define PROFILE_BEGINSEC(obj, mes)            obj.beginSection(obj._##mes);
  #define PROFILE_ENDSEC(obj, mes)              obj.endSection(obj._##mes);

  #define PROFILEPTR_BEGINSEC(obj, mes)  if(obj) obj->beginSection(obj->_##mes);
  #define PROFILEPTR_ENDSEC(obj, mes)    if(obj) obj->endSection(obj->_##mes);
#else
  #define PROFILE_BEGINSEC(obj, mes);
  #define PROFILE_ENDSEC(obj, mes);

  #define PROFILEPTR_BEGINSEC(obj, mes)
  #define PROFILEPTR_ENDSEC(obj, mes)
#endif //_USE_PROFILING_


#endif //__KLPROFILER_HEADERFILE__
