/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klEGLSurfaceEpocWindow.cpp,v 1.4 2004/08/24 19:38:31 wasserer Exp $
 *
\**************************************************************************/

#include "klMain.h"
#include "klEglNative.h"

klEGLSurfaceEpocWindow::klEGLSurfaceEpocWindow()
{
	colorType = COLOR_TYPE_RGB565;
	depthType = DEPTH_TYPE_INT32;

	window		= 0;
	offScreenBmp= 0;
	display     = 0;
}


klEGLSurfaceEpocWindow::~klEGLSurfaceEpocWindow()
{
	destroy();
}

bool klEGLSurfaceEpocWindow::create(klEGLDisplay* nDisp, NativeWindowType nWnd)	
{
	window = nWnd;
	display = nDisp; 

	TSize size = display->getWsScreen()->SizeInPixels();
	width  = size.iWidth;
	height = size.iHeight;

	offScreenBmp = new CWsBitmap( *display->getWsSession() );
	if( offScreenBmp->Create(size, EColor64K) != KErrNone )
	{
		delete offScreenBmp;
		offScreenBmp = 0;
		return false;
	}

	zBuffer = new DEPTH_TYPE[width*height];

	return true; 
}

bool klEGLSurfaceEpocWindow::destroy()	
{ 
	delete offScreenBmp;

	window		= 0;
	offScreenBmp= 0;

	delete[] zBuffer;
	zBuffer = 0;

	return true; 
}

void* klEGLSurfaceEpocWindow::getColorBuffer() const
{
	return offScreenBmp->DataAddress();
}

bool klEGLSurfaceEpocWindow::flip()
{
	if( !offScreenBmp )
		return false;

	// prepare painting
	TBitmapUtil lock(offScreenBmp);
	lock.Begin(TPoint(0,0)); // Lock bitmap heap

	CWindowGc* windowGc = display->getWindowGc();

	windowGc->Activate(*window);

	TSize windowSize = window->Size();
	TRect rect = TRect(windowSize);
  	window->Invalidate(rect);

	// blit
	window->BeginRedraw(rect);
	windowGc->BitBlt(TPoint(), offScreenBmp);
	window->EndRedraw();

	windowGc->Deactivate();
	lock.End(); // Unlock bitmap heap
	display->getWsSession()->Flush();

	return true;
}
