/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klRSW565_GouraudLineTempl.h,v 1.3 2004/06/15 10:17:33 wasserer Exp $
 *
\**************************************************************************/


//
// this file defines a macro template for gouraud shaded line drawing
//

void LINEDRAWER_GOURAUD(int x1, int y1, const klFloat& z1, unsigned int r1, unsigned int g1, unsigned int b1,
						int x2, int y2, const klFloat& z2, unsigned int r2, unsigned int g2, unsigned int b2)
{
	int	x=x1, y=y1, dx, dy;
	int incx, incy;
	klFloat r,g,b,z, dr,dg,db,dz;

	r.setInt(r1);	dr.setInt(r2-r1);
	g.setInt(g1);	dg.setInt(g2-g1);
	b.setInt(b1);	db.setInt(b2-b1);
	z = z1;			dz = z2-z1;

	if(x2 >= x1)
	{	dx = x2 - x1;	incx = 1;	}
	else
	{	dx = x1 - x2;	incx = -1;	}

	if (y2 >= y1)
	{	dy = y2 - y1;	incy = 1;	}
	else
	{	dy = y1 - y2;	incy = -1;	}

	int	balance, offset = y*width + x,
		incxBuf = incx,
		incyBuf = incy*width;

	if(dx==0 && dy==0)
	{
		SET_PIXEL(offset, z, calcColor(r.getInt(),g.getInt(),b.getInt()));
		return;
	}


	if(r1==r2 && g1==g2 && b1==b2)				// flat shading
	{
		unsigned short col = calcColor(r.getInt(),g.getInt(),b.getInt());

		if (dx >= dy)
		{
			klFloat d = lookupInv[dx];
			dz *= d;

			dy <<= 1;
			balance = dy - dx;
			dx <<= 1;

			while (x != x2)
			{
				SET_PIXEL(offset, z, col);
				if (balance >= 0)
				{
					y += incy;		offset += incyBuf;
					balance -= dx;
				}
				balance += dy;
				x += incx;			offset += incxBuf;
				z += dz;
			}
			SET_PIXEL(offset, z, col);
		}
		else
		{
			klFloat d = lookupInv[dy];
			dz *= d;

			dx <<= 1;
			balance = dx - dy;
			dy <<= 1;

			while (y != y2)
			{
				SET_PIXEL(offset, z, col);
				if (balance >= 0)
				{
					x += incx;		offset += incxBuf;
					balance -= dy;
				}
				balance += dx;
				y += incy;			offset += incyBuf;
				z += dz;
			} 
			SET_PIXEL(offset, z, col);
		}
	}
	else									// gouraud shading
	{
		if (dx >= dy)
		{
			klFloat d = lookupInv[dx];
			dr *= d;	dg *= d;	db *= d;	dz *= d;

			dy <<= 1;
			balance = dy - dx;
			dx <<= 1;

			while (x != x2)
			{
				SET_PIXEL(offset, z, calcColor(r.getInt(),g.getInt(),b.getInt()));
				if (balance >= 0)
				{
					y += incy;		offset += incyBuf;
					balance -= dx;
				}
				balance += dy;
				x += incx;			offset += incxBuf;
				r += dr;			g += dg;			b += db;			z += dz;
			}
			SET_PIXEL(offset, z, calcColor(r.getInt(),g.getInt(),b.getInt()));
		}
		else
		{
			klFloat d = lookupInv[dy];
			dr *= d;	dg *= d;	db *= d;	dz *= d;

			dx <<= 1;
			balance = dx - dy;
			dy <<= 1;

			while (y != y2)
			{
				SET_PIXEL(offset, z, calcColor(r.getInt(),g.getInt(),b.getInt()));
				if (balance >= 0)
				{
					x += incx;		offset += incxBuf;
					balance -= dy;
				}
				balance += dx;
				y += incy;			offset += incyBuf;
				r += dr;			g += dg;			b += db;			z += dz;
			} 
			SET_PIXEL(offset, z, calcColor(r.getInt(),g.getInt(),b.getInt()));
		}
	}
}

#undef SET_PIXEL
#undef LINEDRAWER_GOURAUD
