/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/RasterizerSW565/klRSW565_Internal.h,v 1.8 2004/04/19 07:35:24 drgoldie Exp $
 *
\**************************************************************************/


//
// class klRSW565 {
//

//
// Several internal methods...
//

#define RASTER_SPAN16													\
	RASTER_PIX(0)	RASTER_PIX(1)	RASTER_PIX(2)	RASTER_PIX(3)		\
	RASTER_PIX(4)	RASTER_PIX(5)	RASTER_PIX(6)	RASTER_PIX(7)		\
	RASTER_PIX(8)	RASTER_PIX(9)	RASTER_PIX(10)	RASTER_PIX(11)		\
	RASTER_PIX(12)	RASTER_PIX(13)	RASTER_PIX(14)	RASTER_PIX(15)

#define RASTER_SPAN(num)   switch(num) { \
 case 16: RASTER_ONE \
 case 15: RASTER_ONE \
 case 14: RASTER_ONE \
 case 13: RASTER_ONE \
 case 12: RASTER_ONE \
 case 11: RASTER_ONE \
 case 10: RASTER_ONE \
 case  9: RASTER_ONE \
 case  8: RASTER_ONE \
 case  7: RASTER_ONE \
 case  6: RASTER_ONE \
 case  5: RASTER_ONE \
 case  4: RASTER_ONE \
 case  3: RASTER_ONE \
 case  2: RASTER_ONE \
 case  1: RASTER_ONE \
}


int
getNumPixels() const
{
	return width*height;
}

int
getSizeLR() const
{
	return 2*widthLR*height;
}


void
setPixel(int nOffset, PIXELTYPE nColor)
{
	cBuffer[nOffset] = nColor;
}


static inline PIXELTYPE calcColor(int nR, int nG, int nB)
{
	return (unsigned short)(((nR<<8)&0xF800) | ((nG<<3)&0x07E0) | ((nB>>3)&0x1F));
}


static inline PIXELTYPE calcColor(const klFloat& nR, const klFloat& nG, const klFloat& nB)
{
	return calcColor(nR.getByte(),nG.getByte(),nB.getByte());
}


static inline PIXELTYPE calcColor(const klColor3& nCol)
{
	return calcColor(nCol[0], nCol[1], nCol[2]);
}


static inline PIXELTYPE calcColor(const klColor4& nCol)
{
	return calcColor(nCol[0], nCol[1], nCol[2]);
}


static inline PIXELTYPE calcColor(unsigned int nColor)
{
	return (PIXELTYPE)(((nColor>>8)&RED_MASK) | ((nColor>>5)&GREEN_MASK) | ((nColor>>3)&BLUE_MASK));
}


static inline unsigned short blendPixel16(unsigned short nSrc, unsigned short nDst, int nOpacity)
{
	// shamelessly stolen from Thierry Tremblay's PocketFrog...
	//
	unsigned short RB1 = (unsigned short)(nDst & (RED_MASK | BLUE_MASK));
	unsigned short G1  = (unsigned short)(nDst & (GREEN_MASK ));
	unsigned short RB2 = (unsigned short)(nSrc & (RED_MASK | BLUE_MASK));
	unsigned short G2  = (unsigned short)(nSrc & (GREEN_MASK));
	unsigned short RB = (unsigned short)(RB1 + (((RB2-RB1) * (nOpacity>>3)) >> 5));
	unsigned short G  = (unsigned short)(G1 + (((G2-G1)*(nOpacity>>2))>>6));

	RB &= (RED_MASK | BLUE_MASK);
	G  &= (GREEN_MASK);

	return (unsigned short)(RB | G);
}


static inline unsigned short blendPixel16_6(unsigned short nSrc, unsigned short nDst, int nOpacity)
{
	// shamelessly stolen from Thierry Tremblay's PocketFrog...
	// modified to use 6-bit opacity values
	//
	unsigned short RB1 = (unsigned short)(nDst & (RED_MASK | BLUE_MASK));
	unsigned short G1  = (unsigned short)(nDst & (GREEN_MASK ));
	unsigned short RB2 = (unsigned short)(nSrc & (RED_MASK | BLUE_MASK));
	unsigned short G2  = (unsigned short)(nSrc & (GREEN_MASK));
	unsigned short RB = (unsigned short)(RB1 + (((RB2-RB1) * (nOpacity>>1)) >> 5));
	unsigned short G  = (unsigned short)(G1 + (((G2-G1)*nOpacity)>>6));

	RB &= (RED_MASK | BLUE_MASK);
	G  &= (GREEN_MASK);

	return (unsigned short)(RB | G);
}


static inline unsigned short convertPixel_RGBX32toRGB16(unsigned int nPixel)
{
	return (unsigned short)(((nPixel<<8)&RED_MASK) | ((nPixel>>5)&GREEN_MASK) | ((nPixel>>19)&BLUE_MASK));
}


static inline void convertRGBX32toRGB16(unsigned short* nDst, unsigned char* nSrc, int nNumPixels)
{
	unsigned int *src = reinterpret_cast<unsigned int*>(nSrc);

	while(nNumPixels--)
		*(nDst++) = convertPixel_RGBX32toRGB16(*(src++));
}


static inline void convertRGBX32toRGB16(unsigned int* nDst, unsigned char* nSrc, int nNumPixels)
{
	unsigned int *src = reinterpret_cast<unsigned int*>(nSrc);

	while(nNumPixels--)
		*(nDst++) = convertPixel_RGBX32toRGB16(*(src++));
}


int
getMipMapLevel(const Texture* nTexture, DeviceVertex* nV, int nPrimLen)
{
	if(!nTexture)
		return 0;

	if(nPrimLen>=3)
	{
		int i;

		klFloat v[16][2];

		// create vertices scaled by 1/16
		// (otherwise the area can become to large to fit into fixed point)
		for(i=0; i<nPrimLen; i++)
		{
			v[i][0] = nV[i].pos[0];		v[i][0]>>=4;
			v[i][1] = nV[i].pos[1];		v[i][1]>>=4;
		}

		klFloat screenArea = v[nPrimLen-1][0]*v[0][1] - v[nPrimLen-1][1]*v[0][0];
		for(i=0; i<nPrimLen-1; i++)
			screenArea += v[i][0]*v[i+1][1] - v[i][1]*v[i+1][0];
		klFloat screenAreaI = (screenArea<0 ? -screenArea : screenArea);

		if(screenAreaI==0)		// area zero? so there's nothing to render
			return -1;

		klFloat texArea = nV[nPrimLen-1].tex[0]*nV[0].tex[1] - nV[nPrimLen-1].tex[1]*nV[0].tex[0];
		for(i=0; i<nPrimLen-1; i++)
			texArea += nV[i].tex[0]*nV[i+1].tex[1] - nV[i].tex[1]*nV[i+1].tex[0];
		klFloat texAreaI = (texArea<0 ? -texArea : texArea), w,h;
		w.setInt(nTexture->mmLevels[0]->width>>4);
		h.setInt(nTexture->mmLevels[0]->height>>4);
		texAreaI *= w*h;

		klFloat factor = texAreaI / screenAreaI;
		int mipmapLevel;

		if(factor<5)
			mipmapLevel = 0;
		else if(factor<20)
			mipmapLevel = 1;
		else if(factor<80)
			mipmapLevel = 2;
		else if(factor<300)
			mipmapLevel = 3;
		else if(factor<800)
			mipmapLevel = 4;
		else
			mipmapLevel = 5;

		while(nTexture->mmLevels[mipmapLevel]==NULL && mipmapLevel>=0)
			mipmapLevel--;

		return mipmapLevel;
	}
	else
		return 0;
}


MipMapLevel*
getNextMipMapLevel(const MipMapLevel& nSource)
{
	int sW=nSource.width, sH=nSource.height,
		dW=nSource.width/2, dH=nSource.height/2,
		x,y;
	MipMapLevel* ret;

	if(2*dW!=sW || 2*dH!=sH)
		return NULL;

	ret = new MipMapLevel;
	ret->width = dW;
	ret->height = dH;
	ret->border = nSource.border;
	ret->format = nSource.format;
	ret->type = nSource.type;
	ret->pixelType = nSource.pixelType;
	ret->internalFormat = nSource.internalFormat;


	if(nSource.pixelsRGBA!=NULL)								// is this a texture with RGBA pixels?
	{
		const TEX_PIXELTYPE_RGBA *src=nSource.pixelsRGBA;
		TEX_PIXELTYPE_RGBA		 *dst;

		ret->pixelsRGBA = dst = new TEX_PIXELTYPE_RGBA[dW*dH];

		// scale down the pixel data
		//
		for(y=0; y<dH; y++)
		{
			for(x=0; x<dW; x++)
			{
				TEX_PIXELTYPE_RGBA s0=src[0], s1=src[1], s2=src[sW], s3=src[sW+1];

				int r = (getRed8FromRGB565(s0)        + getRed8FromRGB565(s1)        + getRed8FromRGB565(s2)        + getRed8FromRGB565(s3))>>2,
					g = (getGreen8FromRGB565(s0)      + getGreen8FromRGB565(s1)      + getGreen8FromRGB565(s2)      + getGreen8FromRGB565(s3))>>2,
					b = (getBlue8FromRGB565(s0)       + getBlue8FromRGB565(s1)       + getBlue8FromRGB565(s2)       + getBlue8FromRGB565(s3))>>2,
					a = (getAlpha8FromXARGB88565(s0)  + getAlpha8FromXARGB88565(s1)  + getAlpha8FromXARGB88565(s2)  + getAlpha8FromXARGB88565(s3))>>2;

				unsigned short dstC = calcColor(r,g,b);

				*dst = (a<<16) | dstC;

				src+=2;
				dst++;
			}
			src += sW;
		}
	}
	else
	if(nSource.pixelsRGB!=NULL)									// is this a texture with RGB pixels?
	{
		const TEX_PIXELTYPE_RGB *src=nSource.pixelsRGB;
		TEX_PIXELTYPE_RGB		*dst;

		ret->pixelsRGB = dst = new TEX_PIXELTYPE_RGB[dW*dH];

		// scale down the pixel data
		//
		for(y=0; y<dH; y++)
		{
			for(x=0; x<dW; x++)
			{
				TEX_PIXELTYPE_RGB s0=src[0], s1=src[1], s2=src[sW], s3=src[sW+1];

				int r = (getRed8FromRGB565(s0)        + getRed8FromRGB565(s1)        + getRed8FromRGB565(s2)        + getRed8FromRGB565(s3))>>2,
					g = (getGreen8FromRGB565(s0)      + getGreen8FromRGB565(s1)      + getGreen8FromRGB565(s2)      + getGreen8FromRGB565(s3))>>2,
					b = (getBlue8FromRGB565(s0)       + getBlue8FromRGB565(s1)       + getBlue8FromRGB565(s2)       + getBlue8FromRGB565(s3))>>2;

				*dst = calcColor(r,g,b);

				src+=2;
				dst++;
			}
			src += sW;
		}
	}
	else
	{
		// ERROR: should never come here...
	}
		

	return ret;
}


//
// } class klRSW565
//
