/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/RasterizerSW565/klRSW565_InterpolatorZ.h,v 1.4 2004/04/17 16:57:58 drgoldie Exp $
 *
\**************************************************************************/


//
//class klRSW565 {
//

class InterpolatorZ_Linear
{
public:
	typedef klFloat 	DTYPE;

	InterpolatorZ_Linear() {}

	InterpolatorZ_Linear(klFloat nLeft, klFloat nRight, klFloat nPreStep, klFloat nLenInv)
	{
		dz = (nRight-nLeft)*nLenInv;
		z = nLeft + dz*nPreStep;
	}

	void step()
	{
		z += dz;
	}

	void step7()
	{
		klFloat dz7(dz);
		dz7 <<= 3;
		dz7 -= dz;

		z += dz7;
	}

	void step8()
	{
		klFloat dz8(dz);
		dz8 <<= 3;

		z += dz8;
	}

	void step(int nSteps)
	{
		klFloat dzM(dz);

		switch(nSteps)
		{
		case 0:
			return;
		case 1:
			z += dz;
			return;
		case 2:
			z += dz;
			z += dz;
			return;
		case 3:
			dzM <<= 1;
			z += dzM;
			z += dz;
			return;
		case 4:
			dzM <<= 2;
			z += dzM;
			return;
		case 5:
			dzM <<= 2;
			z += dzM;
			z += dz;
			return;
		case 6:
			dzM <<= 2;
			dzM += dz;
			dzM += dz;
			z += dzM;
			return;
		case 7:
			dzM <<= 3;
			z += dzM;
			z -= dz;
			return;
		case 8:
			dzM <<= 3;
			z += dzM;
			return;
		default:
			z += dz*nSteps;
		}
	}

	DTYPE getZ()
	{
		return z;
	}

	DTYPE getDZ()
	{
		return dz;
	}

	bool isStepPositive()
	{
		return dz>0;
	}

protected:
	DTYPE	z, dz;
};

//
// } class klRSW565
//
