/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/RasterizerSW565/klRSW565_Line.h,v 1.6 2004/06/15 10:17:33 wasserer Exp $
 *
\**************************************************************************/


//
//class klRSW565 {
//

//
// Line drawing functions
//

inline void
drawLine(const DeviceVertex& nV0, const DeviceVertex& nV1)
{
	DeviceVertex v0(nV0), v1(nV1);

	if(!clipline2d(v0,v1))
		return;

	LINEFUNC_GOURAUD drawFunc = getLineFunc_Gouraud(settings.zwrite, settings.ztest);

	(this->*drawFunc)(v0.pos[0].getInt(),v0.pos[1].getInt(),v0.pos[2], v0.col[0].getByte(),v0.col[1].getByte(),v0.col[2].getByte(),
					  v1.pos[0].getInt(),v1.pos[1].getInt(),v1.pos[2], v1.col[0].getByte(),v1.col[1].getByte(),v1.col[2].getByte());
}


// clips a shaded line at the clip rectangle
//
bool
clipline2d(DeviceVertex& nV0, DeviceVertex& nV1)
{
	if(!clipEnabled)
		return true;

	// check for quick rejection cases
	//
	if(nV0.pos[0]<clipX0 && nV1.pos[0]<clipX0)
		return false;
	if(nV0.pos[0]>clipX1 && nV1.pos[0]>clipX1)
		return false;
	if(nV0.pos[1]<clipY0 && nV1.pos[1]<clipY0)
		return false;
	if(nV0.pos[1]>clipY1 && nV1.pos[1]>clipY1)
		return false;


	// clip on left side?
	//
	if(nV0.pos[0]<clipX0 && nV1.pos[0]>=clipX0)			// nV0.pos[0] left outside
	{
		klFloat d = (clipX0-nV0.pos[0]) / (nV1.pos[0]-nV0.pos[0]);

		nV0.pos[0] = clipX0;							nV0.pos[1] += (d*(nV1.pos[1]-nV0.pos[1]));		nV0.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV0.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV0.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV0.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}
	if(nV1.pos[0]<clipX0 && nV0.pos[0]>=clipX0)			// nV1.pos[0] left outside
	{
		klFloat d = (clipX0-nV1.pos[0]) / (nV1.pos[0]-nV0.pos[0]);

		nV1.pos[0] = clipX0;							nV1.pos[1] += (d*(nV1.pos[1]-nV0.pos[1]));		nV1.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV1.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV1.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV1.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}

	// clip on right side?
	//
	if(nV1.pos[0]<clipX1 && nV0.pos[0]>=clipX1)			// nV0.pos[0] right outside
	{
		klFloat d = (clipX1-nV0.pos[0]) / (nV1.pos[0]-nV0.pos[0]);

		nV0.pos[0] = clipX1;							nV0.pos[1] += (d*(nV1.pos[1]-nV0.pos[1]));		nV0.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV0.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV0.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV0.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}
	if(nV0.pos[0]<clipX1 && nV1.pos[0]>=clipX1)			// nV1.pos[0] right outside
	{
		klFloat d = (clipX1-nV1.pos[0]) / (nV1.pos[0]-nV0.pos[0]);

		nV1.pos[0] = clipX1;							nV1.pos[1] += (d*(nV1.pos[1]-nV0.pos[1]));		nV1.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV1.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV1.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV1.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}

	// clip on top side?
	//
	if(nV0.pos[1]<clipY0 && nV1.pos[1]>=clipY0)			// nV0.pos[1] top outside
	{
		klFloat d = (clipY0-nV0.pos[1]) / (nV1.pos[1]-nV0.pos[1]);

		nV0.pos[0] += (d*(nV1.pos[0]-nV0.pos[0]));		nV0.pos[1] = clipY0;						nV0.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV0.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV0.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV0.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}
	if(nV1.pos[1]<clipY0 && nV0.pos[1]>=clipY0)			// nV1.pos[1] top outside
	{
		klFloat d = (clipY0-nV1.pos[1]) / (nV1.pos[1]-nV0.pos[1]);

		nV1.pos[0] += (d*(nV1.pos[0]-nV0.pos[0]));		nV1.pos[1] = clipY0;						nV1.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV1.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV1.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV1.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}

	// clip on bottom side?
	//
	if(nV1.pos[1]<clipY1 && nV0.pos[1]>=clipY1)			// nV0.pos[1] right outside
	{
		klFloat d = (clipY1-nV0.pos[1]) / (nV1.pos[1]-nV0.pos[1]);

		nV0.pos[0] += (d*(nV1.pos[0]-nV0.pos[0]));		nV0.pos[1] = clipY1;						nV0.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV0.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV0.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV0.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}
	if(nV0.pos[1]<clipY1 && nV1.pos[1]>=clipY1)			// nV1.pos[1] right outside
	{
		klFloat d = (clipY1-nV1.pos[1]) / (nV1.pos[1]-nV0.pos[1]);

		nV1.pos[0] += (d*(nV1.pos[0]-nV0.pos[0]));		nV1.pos[1] = clipY1;						nV1.pos[2] += (d*(nV1.pos[2]-nV0.pos[2]));
		nV1.col[0] += (d*(nV1.col[0]-nV0.col[0]));		nV1.col[1] += (d*(nV1.col[1]-nV0.col[1]));		nV1.col[2] += (d*(nV1.col[2]-nV0.col[2]));
	}

	return true;
}


////////////////////////////////////////////////////////////////////////////////////////
//
//   drawLine_Flat draws a flat shaded line without z-buffer
//
void
drawLine_Flat(int x1, int y1, int x2, int y2, PIXELTYPE nColor)
{
	int	x=x1, y=y1;
	int dx, dy;
	int incx, incy;
	int balance;

	if(x2 >= x1)
	{	dx = x2 - x1;	incx = 1;	}
	else
	{	dx = x1 - x2;	incx = -1;	}

	if (y2 >= y1)
	{	dy = y2 - y1;	incy = 1;	}
	else
	{	dy = y1 - y2;	incy = -1;	}

	int	offset = y*width + x,
		incxBuf = incx,
		incyBuf = incy*width;

	if(dx==0 && dy==0)
	{
		setPixel(offset, nColor);
		return;
	}

	if (dx >= dy)
	{
		dy <<= 1;
		balance = dy - dx;
		dx <<= 1;

		while (x != x2)
		{
			setPixel(offset, nColor);
			if (balance >= 0)
			{
				y += incy;		offset += incyBuf;
				balance -= dx;
			}
			balance += dy;
			x += incx;			offset += incxBuf;
		}
		setPixel(offset, nColor);
	}
	else
	{
		dx <<= 1;
		balance = dx - dy;
		dy <<= 1;
		while (y != y2)
		{
			setPixel(offset, nColor);
			if (balance >= 0)
			{
				x += incx;		offset += incxBuf;
				balance -= dy;
			}
			balance += dx;
			y += incy;			offset += incyBuf;
		} 
		setPixel(offset, nColor);
	}
}


////////////////////////////////////////////////////////////////////////////////////////
//
//   drawLine_Gouraud draws a gouraud shaded line with z-buffer
//
void
drawLine_Gouraud(int x1, int y1, int r1, int g1, int b1,
				 int x2, int y2, int r2, int g2, int b2)
{
	int prec = 1024, precs = 10;
	int	x=x1, y=y1, r=r1<<precs, g=g1<<precs, b=b1<<precs;
	int dx, dy, dr,dg,db;
	int incx, incy;
	int balance;

	if(x2 >= x1)
	{	dx = x2 - x1;	incx = 1;	}
	else
	{	dx = x1 - x2;	incx = -1;	}

	if (y2 >= y1)
	{	dy = y2 - y1;	incy = 1;	}
	else
	{	dy = y1 - y2;	incy = -1;	}

	int	offset = y*width + x,
		incxBuf = incx,
		incyBuf = incy*width;

	if(dx==0 && dy==0)
	{
		setPixel(offset, calcColor(r>>precs,g>>precs,b>>precs));
		return;
	}

	if (dx >= dy)
	{
		dr = (prec*(r2-r1))/dx;
		dg = (prec*(g2-g1))/dx;
		db = (prec*(b2-b1))/dx;
		dy <<= 1;
		balance = dy - dx;
		dx <<= 1;

		while (x != x2)
		{
			setPixel(offset, calcColor(r>>precs,g>>precs,b>>precs));
			if (balance >= 0)
			{
				y += incy;		offset += incyBuf;
				balance -= dx;
			}
			balance += dy;
			x += incx;			offset += incxBuf;
			r += dr;	g += dg;	b += db;
		}
		setPixel(offset, calcColor(r>>precs,g>>precs,b>>precs));
	}
	else
	{
		dr = (prec*(r2-r1))/dy;
		dg = (prec*(g2-g1))/dy;
		db = (prec*(b2-b1))/dy;
		dx <<= 1;
		balance = dx - dy;
		dy <<= 1;
		while (y != y2)
		{
			setPixel(offset, calcColor(r>>precs,g>>precs,b>>precs));
			if (balance >= 0)
			{
				x += incx;		offset += incxBuf;
				balance -= dy;
			}
			balance += dx;
			y += incy;			offset += incyBuf;
			r += dr;	g += dg;	b += db;
		} 
		setPixel(offset, calcColor(r>>precs,g>>precs,b>>precs));
	}
}



LINEFUNC_GOURAUD getLineFunc_Gouraud(RASTER_MODE_ZWRITE nZWrite, RASTER_MODE_ZTEST nZTest)
{
	if(nZWrite==RASTER_ZWRITE_OFF)			// no z-write
	{
		switch(nZTest)
		{
		case RASTER_ZTEST_ALWAYS:
			return &klRSW565::drawLine_Gouraud_NoZWrite_Always;

		case RASTER_ZTEST_LESS:
			return &klRSW565::drawLine_Gouraud_NoZWrite_Less;

		case RASTER_ZTEST_LEQUAL:
			return &klRSW565::drawLine_Gouraud_NoZWrite_LessEqual;

		case RASTER_ZTEST_EQUAL:
			return &klRSW565::drawLine_Gouraud_NoZWrite_Equal;

		case RASTER_ZTEST_GEQUAL:
			return &klRSW565::drawLine_Gouraud_NoZWrite_GreaterEqual;

		case RASTER_ZTEST_GREATER:
			return &klRSW565::drawLine_Gouraud_NoZWrite_Greater;

		case RASTER_ZTEST_NOTEQUAL:
			return &klRSW565::drawLine_Gouraud_NoZWrite_NotEqual;

		case RASTER_ZTEST_NEVER:
			return &klRSW565::drawLine_Gouraud_Never;
		}
	}
	else									// z-write
	{
		switch(nZTest)
		{
		case RASTER_ZTEST_ALWAYS:
			return &klRSW565::drawLine_Gouraud_ZWrite_Always;

		case RASTER_ZTEST_LESS:
			return &klRSW565::drawLine_Gouraud_ZWrite_Less;

		case RASTER_ZTEST_LEQUAL:
			return &klRSW565::drawLine_Gouraud_ZWrite_LessEqual;

		case RASTER_ZTEST_EQUAL:
			return &klRSW565::drawLine_Gouraud_ZWrite_Equal;

		case RASTER_ZTEST_GEQUAL:
			return &klRSW565::drawLine_Gouraud_ZWrite_GreaterEqual;

		case RASTER_ZTEST_GREATER:
			return &klRSW565::drawLine_Gouraud_ZWrite_Greater;

		case RASTER_ZTEST_NOTEQUAL:
			return &klRSW565::drawLine_Gouraud_ZWrite_NotEqual;

		case RASTER_ZTEST_NEVER:
			return &klRSW565::drawLine_Gouraud_Never;
		}
	}


	// should never come here...
	//
	return &klRSW565::drawLine_Gouraud_ZWrite_Less;
}


///////////////////////////////////////////////////////////////////
//
//  define pixel access functions for line drawing
//

void
drawPixel_NoZWrite_Less(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ<zBuffer[nOffset])
		cBuffer[nOffset] = nCol;
}

void
drawPixel_NoZWrite_LessEqual(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ<=zBuffer[nOffset])
		cBuffer[nOffset] = nCol;
}

void
drawPixel_NoZWrite_Equal(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ==zBuffer[nOffset])
		cBuffer[nOffset] = nCol;
}

void
drawPixel_NoZWrite_Greater(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ>zBuffer[nOffset])
		cBuffer[nOffset] = nCol;
}

void
drawPixel_NoZWrite_NotEqual(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ!=zBuffer[nOffset])
		cBuffer[nOffset] = nCol;
}

void
drawPixel_NoZWrite_GreaterEqual(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ>=zBuffer[nOffset])
		cBuffer[nOffset] = nCol;
}

void
drawPixel_NoZWrite_Always(int nOffset, const klFloat& /*nZ*/, unsigned short nCol)
{
	cBuffer[nOffset] = nCol;
}


void
drawPixel_ZWrite_Less(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ<zBuffer[nOffset])
	{
		cBuffer[nOffset] = nCol;
		zBuffer[nOffset] = nZ;
	}
}

void
drawPixel_ZWrite_LessEqual(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ<=zBuffer[nOffset])
	{
		cBuffer[nOffset] = nCol;
		zBuffer[nOffset] = nZ;
	}
	else
	{
		nCol = nCol;
	}
}

void
drawPixel_ZWrite_Equal(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ==zBuffer[nOffset])
	{
		cBuffer[nOffset] = nCol;
		zBuffer[nOffset] = nZ;
	}
}

void
drawPixel_ZWrite_Greater(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ>zBuffer[nOffset])
	{
		cBuffer[nOffset] = nCol;
		zBuffer[nOffset] = nZ;
	}
}

void
drawPixel_ZWrite_NotEqual(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ!=zBuffer[nOffset])
	{
		cBuffer[nOffset] = nCol;
		zBuffer[nOffset] = nZ;
	}
}

void
drawPixel_ZWrite_GreaterEqual(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	if(nZ>=zBuffer[nOffset])
	{
		cBuffer[nOffset] = nCol;
		zBuffer[nOffset] = nZ;
	}
}

void
drawPixel_ZWrite_Always(int nOffset, const klFloat& nZ, unsigned short nCol)
{
	cBuffer[nOffset] = nCol;
	zBuffer[nOffset] = nZ;
}


void
drawPixel_Never(int /*nOffset*/, const klFloat& /*nZ*/, unsigned short /*nCol*/)
{
}


///////////////////////////////////////////////////////////////////
//
//  define all kinds of line drawing functions by using the
//  implementation in klRSW565_GouraudLineTempl.h
//

#define SET_PIXEL drawPixel_NoZWrite_Less
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_Less
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_NoZWrite_LessEqual
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_LessEqual
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_NoZWrite_Equal
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_Equal
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_NoZWrite_Greater
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_Greater
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_NoZWrite_NotEqual
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_NotEqual
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_NoZWrite_GreaterEqual
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_GreaterEqual
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_NoZWrite_Always
#define LINEDRAWER_GOURAUD drawLine_Gouraud_NoZWrite_Always
#include "klRSW565_GouraudLineTempl.h"


#define SET_PIXEL drawPixel_ZWrite_Less
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_Less
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_ZWrite_LessEqual
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_LessEqual
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_ZWrite_Equal
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_Equal
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_ZWrite_Greater
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_Greater
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_ZWrite_NotEqual
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_NotEqual
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_ZWrite_GreaterEqual
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_GreaterEqual
#include "klRSW565_GouraudLineTempl.h"

#define SET_PIXEL drawPixel_ZWrite_Always
#define LINEDRAWER_GOURAUD drawLine_Gouraud_ZWrite_Always
#include "klRSW565_GouraudLineTempl.h"


#define SET_PIXEL drawPixel_Never
#define LINEDRAWER_GOURAUD drawLine_Gouraud_Never
#include "klRSW565_GouraudLineTempl.h"

//
// } class klRSW565
//
