/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/RasterizerSW565/klRSW565_Public.cpp,v 1.6 2004/07/29 12:13:29 drgoldie Exp $
 *
\**************************************************************************/

#include "klRSW565.h"

//
// implement the non-inline methods for the 
// public interface of klRasterizer
//

klRSW565::klRSW565()
{
	int i;

	cBuffer = NULL;
	zBuffer = NULL;
//	zBufferLR = NULL;
	width = height = 0;

	surface = NULL;
	display = NULL;

	clearColor = 0;
	clearDepth = 1.0f;

	cLeft = 0.0f;
	cTop = 0.0f;

	fogCol = 0;

	rasterPosX = 0;
	rasterPosY = 0;

	polygonOffset = 0.0f;
	depthMax = depthMaxHalf = 30000.0f;
	depthMaxHalf *= 0.5f;

	//settings.blendMode = RASTER_BLEND_OFF;
	//settings.color = RASTER_COLOR_GOURAUD;
	settings.fog = RASTER_FOG_OFF;
	settings.polyMode = RASTER_FILL;
	settings.ztest = RASTER_ZTEST_ALWAYS;
	settings.zwrite = RASTER_ZWRITE_OFF;
	settings.perspectiveMapping = true;

	clipEnabled = false;

	currentTexture = NULL;
	for(i=0; i<MAX_TEXTURES; i++)
		textures[i] = NULL;


	// define some constants which we need
	// in fixed point format
	constants._100.setInt(100);
	constants._1over100.setFixed(655);		// = 0.01f
	constants._0.setInt(0);
	constants._1.setInt(1);
	constants._255.setInt(255);
	constants._254.setInt(254);
	constants._764.setInt(764);
	constants.white.setInt(255,255,255,255);
	constants.half.setFloat(0.5f);

	lookupInv[0].setInt(0);
	for(i=1; i<MAX_LOOKUPINV; i++)
	{
		lookupInv[i].setInt(i);
		lookupInv[i].inverse();
	}

	profiler = NULL;
}


klRSW565::~klRSW565()
{
	if(!surface && zBuffer)
		delete zBuffer;

//	if(!surface && zBufferLR)
//		delete zBufferLR;
}


void
klRSW565::setSize(int nWidth, int nHeight)
{
	if(width!=nWidth && height!=nHeight)
	{
		width = nWidth;
		height = nHeight;

		widthLR = (width&Z_LOWRES_MASK) ? 1+width/Z_LOWRES_FACT : width/Z_LOWRES_FACT;
			
		screenCenterX = (width>>1);
		screenCenterY = (height>>1);

		deviceScaleX = screenCenterX + 0.01f;		// make the scale factor a little bit larger, so that we reach
		deviceScaleY = screenCenterY + 0.01f;		// the screen's border even due to fixed point errors

		cRight.setInt(width);			cRight -= cLeft;
		cBottom.setInt(height);			cBottom -= cTop;

		if(surface==NULL)
		{
			if(zBuffer)
				delete [] zBuffer;
			if(width!=0 && height!=0)
				zBuffer = new DEPTHTYPE[getNumPixels()];
		}

		//if(zBufferLR)
		//	delete [] zBufferLR;
		//if(widthLR!=0 && height!=0)
		//	zBufferLR = new DEPTHTYPE[getSizeLR()];
	}
}


void
klRSW565::clear(int nFlags)
{
	int i, numPix = getNumPixels(), numPixH = numPix>>1;

	if(nFlags&GL_COLOR_BUFFER_BIT)
	{
		unsigned int	*cb = reinterpret_cast<unsigned int*>(cBuffer),
						col = clearColor | (clearColor<<16);
		
		for(i=0; i<numPixH; i++)
			cb[i] = col;
	}

	if(nFlags&GL_DEPTH_BUFFER_BIT)
	{
		klFloat d = clearDepth*depthMax;
		for(i=0; i<numPix; i++)
			zBuffer[i] = d;

		//numPix = getSizeLR();
		//for(i=0; i<numPix; i++)
		//	zBufferLR[i] = d;
	}
}


void
klRSW565::renderPrimitive(unsigned int nPrimType, const klVertex* nV, int nPrimLen)
{
	int j,j0,j1;

	// everything clipped away ?
	//
	if(nPrimLen==0)
		return;

	// calculate device coordiantes
	//
	for(j=0; j<nPrimLen; j++)
	{
		klFloat	px = screenCenterX + deviceScaleX*nV[j].pos[0],
				py = screenCenterY - deviceScaleY*nV[j].pos[1];

		if(px<cLeft)
			px = cLeft;
		if(px>cRight)
			px = cRight;
		if(py<cTop)
			py = cTop;
		if(py>cBottom)
			py = cBottom;

		vertices[j].pos[0] = px;
		vertices[j].pos[1] = py;
		vertices[j].pos[2] = nV[j].pos[2] + polygonOffset;

		vertices[j].col = nV[j].col;
		vertices[j].fog = nV[j].fogDensity;
	}

	// calculate a non projected z-value for texturing
	// that lies in a good numeric range...
	if(settings.numtexstages>0)
	{
/*
		klFloat maxZ;
		maxZ.setInt(0);

		for(j=0; j<nPrimLen; j++)
		{
			if(nV[j].unprojectedZ>maxZ)
				maxZ = nV[j].unprojectedZ;
			if(-nV[j].unprojectedZ>maxZ)
				maxZ = -nV[j].unprojectedZ;

			vertices[j].tex = nV[j].texCoord;
		}

		maxZ.inverse();
		for(j=0; j<nPrimLen; j++)
			vertices[j].texZ = maxZ + nV[j].unprojectedZ * maxZ;		// z-value for texture gradient (non-projected z)
*/

		for(j=0; j<nPrimLen; j++)
		{
			vertices[j].texZ = nV[j].unprojectedZ;
			vertices[j].tex = nV[j].texCoord;
		}
	}

	RASTER_POLYGONMODE mode = settings.polyMode;


	// some drawing types require mode overriding...
	//
	switch(nPrimType)
	{
	case GL_POINTS:
		mode = RASTER_POINTS;
		break;

	case GL_LINES:
	case GL_LINE_STRIP:
	case GL_LINE_LOOP:
		mode = RASTER_LINES;
		break;
	}


	switch(mode)
	{
	case RASTER_POINTS:
		for(j=0; j<nPrimLen; j++)
			setPixel(vertices[j].pos[0].getInt(), vertices[j].pos[1].getInt(), vertices[j].col);
		break;

	case RASTER_LINES:
		// some draw types do not close the loop
		//
		switch(nPrimType)
		{
		case GL_LINE_STRIP:
		case GL_LINES:
			j=0;  j0=0;  j1=1;
			break;

		default:
			j=-1; j0=0;  j1=nPrimLen-1;
			break;
		}

		for(; j<nPrimLen-1; j++,j0=j,j1=j+1)
			drawLine(vertices[j0], vertices[j1]);

		if(nPrimType==GL_TRIANGLE_FAN)
			for(j=0; j<nPrimLen; j++)
				drawLine(vertices[0], vertices[j]);
		break;

	case RASTER_FILL:
		{
			MipMapLevel* mml = NULL;
			if(currentTexture && settings.doTex())
			{
				int mmlIdx = 0;
				if(currentTexture->mmLevels[1]!=NULL)
				{
					mmlIdx = getMipMapLevel(currentTexture, vertices, nPrimLen);
					if(mmlIdx==-1)		// area zero?
						return;
				}

				mml = currentTexture->mmLevels[mmlIdx];
			}

			for(j=1; j<nPrimLen-1; j++)
				drawTriangle(vertices[0], vertices[j], vertices[j+1], mml);
		}
		break;
	}
}


int
klRSW565::createTexture()
{
	int i;

	for(i=0; i<MAX_TEXTURES; i++)
		if(textures[i] == NULL)
		{
			textures[i] = new Texture;
			return i+1;					// indices should start with 1 in opengl
		}

	return -1;
}


void
klRSW565::destroyTexture(int nHandle)
{
	nHandle--;							// indices start with 1 in opengl
	if(nHandle>=0 && nHandle<MAX_TEXTURES && textures[nHandle])
	{
		if(currentTexture==textures[nHandle])
			currentTexture = NULL;

		delete textures[nHandle];
		textures[nHandle] = NULL;
	}
}


void
klRSW565::updateCurrentTexture(int nLevel, int nInternalFormat, unsigned int nWidth, unsigned int nHeight,
							   int nBorder, GLenum nFormat, GLenum nPixelType, const void* nPixels)

{
	if(!currentTexture || nLevel<0 || nLevel>=Texture::MAX_MIPMAPLEVELS)
		return;

	if(nPixels==NULL)
		return;

	// if this mipmap level already existed we delete and completey recreate it...
	//
	MipMapLevel* mmLevel = currentTexture->mmLevels[nLevel];
	if(mmLevel)
		delete currentTexture->mmLevels[nLevel];
	mmLevel = currentTexture->mmLevels[nLevel] = new MipMapLevel();


	// update mipmap level information
	//
	mmLevel->internalFormat = nInternalFormat;
	mmLevel->width = nWidth;
	mmLevel->height = nHeight;
	mmLevel->border = nBorder;
	mmLevel->format = nFormat;
	mmLevel->pixelType = nPixelType;


	if(nFormat==GL_RGB)
	{
		unsigned int i, size=nWidth*nHeight, cw3=nWidth*3;
		unsigned char	*tmpBuf=new unsigned char[size*3], *src=tmpBuf;

		// fill pixel buffer flipped, since our texturing routine wants it this way...
		//
		//for(i=0; i<nHeight; i++)
		//	memcpy(tmpBuf+i*cw3, (const unsigned char*)nPixels+(nHeight-i-1)*cw3, cw3);

		for(i=0; i<nHeight; i++)
			memcpy(tmpBuf+i*cw3, (const unsigned char*)nPixels+i*cw3, cw3);

		mmLevel->pixelsRGB = new TEX_PIXELTYPE_RGB[size];

		for(i=0; i<size; i++, src+=3)
			mmLevel->pixelsRGB[i] = calcColor(src[0], src[1], src[2]);

		delete [] tmpBuf;
	}
	else
	if(nFormat==GL_RGBA)
	{
		unsigned int i, size=nWidth*nHeight, cw4=nWidth*4;
		unsigned char	*tmpBuf=new unsigned char[size*4];

		// fill pixel buffer flipped, since our texturing routine wants it this way...
		//
		//for(i=0; i<nHeight; i++)
		//	memcpy(tmpBuf+i*cw4, (const unsigned char*)nPixels+(nHeight-i-1)*cw4, cw4);

		for(i=0; i<nHeight; i++)
			memcpy(tmpBuf+i*cw4, (const unsigned char*)nPixels+i*cw4, cw4);

		mmLevel->pixelsRGBA = new TEX_PIXELTYPE_RGBA[size];
		convertRGBX32toRGB16(mmLevel->pixelsRGBA, tmpBuf, size);

		for(i=0; i<size; i++)
			mmLevel->pixelsRGBA[i] |= (tmpBuf[(i<<2)+3]<<16);

		delete [] tmpBuf;
	}
}


// creates mipmap levels for the current texture
void
klRSW565::createMipMaps()
{
	if(!currentTexture)
		return;

	int curLevel=0;

	if(!currentTexture->mmLevels[curLevel])
		return;

	for(;;) {
		MipMapLevel* nextLevel = getNextMipMapLevel(*currentTexture->mmLevels[curLevel]);
		if(!nextLevel)
			return;

		curLevel++;

		if(currentTexture->mmLevels[curLevel])
			delete currentTexture->mmLevels[curLevel];

		currentTexture->mmLevels[curLevel] = nextLevel;
	}

}


bool
klRSW565::setSurface(klEGLSurface* nSurface)
{
	surface = nSurface;

	if(surface==NULL)
	{
		cBuffer = NULL;
		zBuffer = NULL;
		// zBufferLR = NULL;
		setSize(0,0);
		return true;
	}

	if(surface->getColorType()!=klEGLSurface::COLOR_TYPE_RGB565 ||
	   surface->getDepthType()!=klEGLSurface::DEPTH_TYPE_INT32)
	   return false;

	cBuffer = reinterpret_cast<PIXELTYPE*>(surface->getColorBuffer());
	zBuffer = reinterpret_cast<DEPTHTYPE*>(surface->getDepthBuffer());

	setSize(surface->getWidth(), surface->getHeight());
	return true;
}


bool
klRSW565::applySettings(klRasterizer_Settings& nSettings)
{
	if(&settings!=&nSettings)
		settings = nSettings;


	// check wether we have a special rasterization function for this case
	//
	scanLineDoSpecialFunc =	settings.numtexstages<=1 &&
							settings.fog==RASTER_FOG_OFF &&
							!settings.blendEnabled;


	// get the function that finally writes pixels into the frame buffer
	//
	pixelWriteFunc = getPixelWriteFunc(settings);


	// get an depth test and write function
	//
	depthFunc = drawScanLine_GetDepthFunc(settings.zwrite, settings.ztest);


	// calculate edge options that are independent of vertex data
	//
	edgeOptionsBase = 0;
	if(settings.ztest!=RASTER_ZTEST_NEVER && (settings.ztest!=RASTER_ZTEST_ALWAYS || !settings.zwrite==RASTER_ZWRITE_OFF))
		edgeOptionsBase |= OPT_DEPTH;
	if(settings.fog==RASTER_FOG_ON)
		edgeOptionsBase |= OPT_FOG;


	settings = nSettings;
	return true;
}

/**
}  // class klRSW565
*/
