/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/RasterizerSW565/klRSW565_Public.h,v 1.20 2004/07/29 08:54:11 drgoldie Exp $
 *
\**************************************************************************/

//
// implement the inline methods for the 
// public interface of klRasterizer
//

inline void
klRSW565::setColorBuffer(void* nMem)
{
	cBuffer = reinterpret_cast<PIXELTYPE*>(nMem);
	surface = 0;
}


inline void*
klRSW565::getColorBuffer()
{
	if(surface)
		return surface->getColorBuffer();

	return cBuffer;
}

inline void*
klRSW565::getDepthBuffer()
{
	if(surface)
		return surface->getDepthBuffer();

	return zBuffer;
}


inline void
klRSW565::setClearColor(const klColor4& nColor)
{
	clearColor = calcColor(nColor);
}


inline void
klRSW565::setClearDepth(klFloat nDepth)
{
	clearDepth = nDepth;
}


inline void
klRSW565::set2DClippingEnabled(bool nEnable)
{
	clipEnabled = nEnable;
}

inline void
klRSW565::set2DClipping(int nX0, int nY0, int nX1, int nY1)
{
	clipX0 = nX0;
	clipY0 = nY0;
	clipX1 = nX1;
	clipY1 = nY1;
}


inline void
klRSW565::setFogColor(const klColor4& nColor)
{
	fogCol = calcColor(nColor);
}


inline void
klRSW565::setPixel(int nX, int nY, const klColor3& nColor)
{
	klFloat x,y;

	x.setInt(nX);
	y.setInt(nY);

	if(clipEnabled && (x<clipX0 || x>clipX1 || y<clipY0 || y>clipY1))
		return;

	cBuffer[nX + nY*width] = calcColor(nColor);
}


inline void
klRSW565::setPixel(int nX, int nY, const klColor4& nColor)
{
	klFloat x,y;

	x.setInt(nX);
	y.setInt(nY);

	if(clipEnabled && (x<clipX0 || x>clipX1 || y<clipY0 || y>clipY1))
		return;

	cBuffer[nX + nY*width] = calcColor(nColor);
}


inline void
klRSW565::setCurrentTexture(int nHandle)
{
	if(nHandle==0)
		currentTexture = NULL;
	else
	{
		nHandle--;							// indices start with 1 in opengl
		if(nHandle>=0 && nHandle<MAX_TEXTURES && textures[nHandle])
			currentTexture = textures[nHandle];
	}
}


inline const klMipMapLevel_Info*
klRSW565::getMipMapLevelInfo(int nLevel)
{
	if(currentTexture && nLevel<Texture::MAX_MIPMAPLEVELS)
		return currentTexture->mmLevels[nLevel];

	return NULL;
}


// this function does a simple - yet very fast - RGBX8888 to RGB565 conversion
inline void
klRSW565::drawPixels(const void* nPixels)
{
	unsigned int num = width*height/2;

	const unsigned int	*srcI = reinterpret_cast<const unsigned int*>(nPixels);
	unsigned int		*dstI = reinterpret_cast<unsigned int*>(cBuffer);
	register int		b,c;

	while(num--)
	{
		b = srcI[0];
		c = srcI[1];

		// do two pixels at once
		*dstI=    ((b>>8)&0xF800) | ((b>>5)&0x07E0) | ((b>>3)&0x1F) |
				((((c>>8)&0xF800) | ((c>>5)&0x07E0) | ((c>>3)&0x1F))<<16);

		srcI += 2;
		dstI += 1;
	}
}


// can raster pos be negative??
inline void
klRSW565::drawPixels(int nWidth, int nHeight, GLenum /*nFormat*/, GLenum /*nPixelType*/, const void* nPixels)
{
	// shortcut to old full screen version
	// which is ways faster at the moment
	//
	if(nWidth==width && nHeight==height)
	{
		drawPixels(nPixels);
		return;
	}

	const unsigned int	*srcI = reinterpret_cast<const unsigned int*>(nPixels);
	// can no longer to 2 pixels at once as raster position may cause the 2nd pixel to be on other 
	// side of the screen.
	unsigned short			*dstI = reinterpret_cast<unsigned short*>(cBuffer);
	register unsigned int	b;

	int currentW, currentH;

	dstI += width*rasterPosY + rasterPosX;
	for (currentH = 0; currentH < nHeight; currentH++) {
		for (currentW = 0; currentW < nWidth; currentW++) {
			// one pixel at a time
			if ((currentW + rasterPosX) > width) {
				srcI += (nWidth-currentW);  // move forward through the rest of this row, as it won't be displayed.
				break;
			}
			b = srcI[0];
			*dstI=    static_cast<unsigned short>(((b>>8)&0xF800) | ((b>>5)&0x07E0) | ((b>>3)&0x1F));
			srcI += 1;
			dstI += 1;
		}
		if ((currentH + rasterPosY) > height)
			break;
		dstI += (width - currentW);  // currentW rather than width, in case off the edge of screen.
	}
}


inline void
klRSW565::setRasterPos(klFloat nX, klFloat nY)
{
	klFloat x = nX*width + constants.half,
			y = nY*height + constants.half;

	rasterPosX = x.getInt();
	rasterPosY = y.getInt();
}


inline void
klRSW565::setPolygonOffset(const klFloat& nOffset)
{
	polygonOffset = depthMax*nOffset;
}


inline void
klRSW565::setProfiler(klProfiler* nProfiler)
{
	profiler = nProfiler;
}


inline klEGLSurface*
klRSW565::getSurface()
{
	return surface;
}


inline bool
klRSW565::setDisplay(klEGLDisplay* nDisplay)
{
	display = nDisplay;
	return true;
}


inline klEGLDisplay*
klRSW565::getDisplay()
{
	return display;
}

/**
}  // class klRSW565
*/
