/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klEGLSurfaceX86Window.h,v 1.3 2004/07/29 08:54:11 drgoldie Exp $
 *
\**************************************************************************/


#ifndef __KLEGLSURFACE_X86WINDOW_HEADERFILE__
#define __KLEGLSURFACE_X86WINDOW_HEADERFILE__


#include "../klEGLSurface.h"
#include <windows.h>


#ifndef _IS_WIN32_
#pragma message (__FILE__ "() error: this code only works under x86 Windows")
#endif //_IS_WIN32_


/// x86 (desktop) Windows implementation of EGLSurface
/**
 *  This x86 (desktop) Windows implementation of EGLSurface.
 *  It will render 2x zoomed into a window. The window must
 *  therefore be two-times larger than the "real" resolution
 *  of the buffer (e.g. 640x480 for a 320x240 render screen).
 */
class klEGLSurfaceX86Window : public klEGLSurface
{
public:
	klEGLSurfaceX86Window();
	~klEGLSurfaceX86Window();
	
	virtual bool create(klEGLDisplay* nDisp, NativeWindowType nWnd);

	virtual NativeWindowType getNativeWindow();
	virtual klEGLDisplay* getDisplay();

	virtual bool destroy();
	virtual bool flip();

private:
	void setSize(int nWidth, int nHeight);

	BITMAPINFO	bitmapInfo;
	RGBQUAD*	bitmapData;
	HWND		hwnd;

	klEGLDisplay* display;
	bool		halfResSurface;
};


#endif //__KLEGLSURFACE_X86WINDOW_HEADERFILE__
