/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klEGLSurfacePocketHAL.cpp,v 1.8 2004/07/29 08:54:11 drgoldie Exp $
 *
\**************************************************************************/


#ifdef _USE_POCKETHAL_

#include "klMain.h"
#include "../klEGLNative.h"
#include "../klContext.h"

PHAL::Display*	klEGLSurfacePocketHAL::phDisplay = NULL;
bool			klEGLSurfacePocketHAL::phDisplayInitialized = false;

klContext* getCurrentContext();


//
// if you do not want to use PocketHAL please
// remove the _USE_POCKETHAL_ setting from the
// preprocessor definitions in the project settings.
//
// for more information about support for PocketHAL
// have a look into the file klimt/dependencies.txt
//
#include <PocketHal/display.h>
#include <PocketHal/surface.h>
#pragma comment( lib, "PocketHAL_ARM.lib" )



klEGLSurfacePocketHAL::klEGLSurfacePocketHAL()
{
	colorType = COLOR_TYPE_RGB565;
	depthType = DEPTH_TYPE_INT32;

	width = height = 0;
	cBuffer = NULL;
	zBuffer = NULL;

	display = 0;
}


klEGLSurfacePocketHAL::~klEGLSurfacePocketHAL()
{
	destroy();
}


bool
klEGLSurfacePocketHAL::create(klEGLDisplay* nDisplay, NativeWindowType nWnd)
{
	hwnd = nWnd;									// note: we keep this HWND and HDC
	display = nDisplay;

	//
	// TODO: also support non fullscreen windows
	//

	// try to call up PocketHAL
	//
	PHAL::DisplayParameters		phDisplayParams;

	if(/*getCurrentContext()->klGetPortraitMode()*/ false)
		phDisplayParams.m_orientation = PHAL::ORIENTATION_NORMAL;
    else
		phDisplayParams.m_orientation = PHAL::ORIENTATION_ROTATE90CCW;		// use rotated 90 to the left (portrait)

	if(phDisplay==NULL)
		phDisplay = PHAL::Display::Create(hwnd, phDisplayParams);
	phDisplayInitialized = true;

	if(!phDisplay)
		return false;

	if(phDisplay->BeginScene())
		cBuffer = (unsigned short*)phDisplay->GetBackBuffer()->GetPixels();

	setSize(phDisplayParams.m_width, phDisplayParams.m_height);

	return true;
}


NativeWindowType
klEGLSurfacePocketHAL::getNativeWindow()
{
	return hwnd;
}


klEGLDisplay*
klEGLSurfacePocketHAL::getDisplay()
{
	return display;
}


bool
klEGLSurfacePocketHAL::destroy()
{
	bool ret = true;

	if(zBuffer)
		delete zBuffer;
	zBuffer = NULL;

	width = height = 0;

	if(phDisplay && phDisplayInitialized)
	{
		//phDisplay->Shutdown();
		delete phDisplay;
		phDisplay = NULL;
		cBuffer = NULL;
		phDisplayInitialized = false;
		Sleep(200);
	}

	display = 0;
	return ret;
}


void
klEGLSurfacePocketHAL::setSize(int nWidth, int nHeight)
{
	if(width!=nWidth || height!=nHeight)
	{
		width = nWidth;
		height = nHeight;

		if(zBuffer)
			delete zBuffer;
		zBuffer = new unsigned int[width*height];
	}
}


bool
klEGLSurfacePocketHAL::flip()
{
	if(phDisplay)
	{
		phDisplay->Swap();
		if(phDisplay->BeginScene())
			cBuffer = (unsigned short*)phDisplay->GetBackBuffer()->GetPixels();
	}

	return true;
}


#endif //_USE_POCKETHAL_
