/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/klContext_oglArray.h,v 1.6 2004/06/15 10:15:49 wasserer Exp $
 *
\**************************************************************************/



/**
class klContext
{
*/

//
// This file holds all opengl functions that deal with
//
//                  vertex arrays
//

void glEnableClientState(GLenum nState)
{
	switch(nState)
	{
	case GL_VERTEX_ARRAY:
		array.vertex.enabled = true;
		break;

	case GL_NORMAL_ARRAY:
		array.normal.enabled = true;
		break;

	case GL_COLOR_ARRAY:
		array.color.enabled = true;
		break;

	case GL_TEXTURE_COORD_ARRAY:
		array.texCoord.enabled = true;
		break;

	default:
		//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glEnableClientState(): unsupported parameter (%d)", nState);
		break;
	}
}


void glDisableClientState(GLenum nState)
{
	switch(nState)
	{
	case GL_VERTEX_ARRAY:
		array.vertex.enabled = false;
		break;

	case GL_NORMAL_ARRAY:
		array.normal.enabled = false;
		break;

	case GL_COLOR_ARRAY:
		array.color.enabled = false;
		break;

	case GL_TEXTURE_COORD_ARRAY:
		array.texCoord.enabled = false;
		break;

	default:
		//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glDisableClientState(): unsupported parameter (%d)", nState);
		break;
	}

	//arrayCachedDirty = true;
}


void glIndexPointer(GLenum /*type*/, GLsizei /*stride*/, const GLvoid * /*pointer*/)
{
	// silently ignored, klimt does not support indexed colors in general
}

void glColorPointer(GLint size, GLenum type, GLsizei stride, const GLvoid *pointer)
{
	array.color.set(size, type, stride, pointer);
	array.colorFetcher.setArray(array.color);
	vertexCache.makeEmpty();
}


void glNormalPointer(GLenum type, GLsizei stride, const GLvoid *pointer)
{
	array.normal.set(3, type, stride, pointer);
	array.normalFetcher.setArray(array.normal);
	vertexCache.makeEmpty();
}


void glTexCoordPointer(GLint size, GLenum type, GLsizei stride, const GLvoid *pointer)
{
	array.texCoord.set(size, type, stride, pointer);
	array.texCoordFetcher.setArray(array.texCoord);
	vertexCache.makeEmpty();
}


void glVertexPointer(GLint size, GLenum type, GLsizei stride, const GLvoid *pointer)
{
	array.vertex.set(size, type, stride, pointer);
	array.vertexFetcher.setArray(array.vertex);
	vertexCache.makeEmpty();
}


void glDrawElements(GLenum polymode, GLsizei count, GLenum datatype, const GLvoid *indices)
{
	if(!array.vertex.enabled || !array.vertex.clientData)
		return;

	renderArray(polymode, count, datatype, indices);
}


void glDrawArrays(GLenum mode, GLint first, GLsizei count)
{
	array.nonIndexedBase = first;
	renderArray(mode, count, GL_FALSE, NULL);
}


void glArrayElement(GLint index)
{
	applyArrayElement(index);
}


void glInterleavedArrays(GLenum /*format*/, GLsizei /*stride*/, const GLvoid * /*pointer*/)
{
	// silently ignored, not supported yet
}



/**
}  // class klContext
*/
