/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/klContext_oglPixels.h,v 1.7 2004/04/26 10:14:29 drgoldie Exp $
 *
\**************************************************************************/


/**
class klContext
{
*/


//
// This file holds all opengl functions that deal with
//
//                pixel reading/writing
//


void glRasterPos2f(GLfloat x, GLfloat y)
{
	klVec4 pos;

	pos[0] = x;
	pos[1] = y;
	pos[2].setInt(0);
	pos[3].setInt(1);

	glRasterPos(pos);
}


void glRasterPos2i(GLint x, GLint y)
{
	klVec4 pos;

	pos[0].setInt(x);
	pos[1].setInt(y);
	pos[2].setInt(0);
	pos[3].setInt(1);

	glRasterPos(pos);
}


void glRasterPos3f(GLfloat x, GLfloat y, GLfloat z)
{
	klVec4 pos;

	pos[0] = x;
	pos[1] = y;
	pos[2] = z;
	pos[3].setInt(1);

	glRasterPos(pos);
}


void glRasterPos3i(GLint x, GLint y, GLint z)
{
	klVec4 pos;

	pos[0].setInt(x);
	pos[1].setInt(y);
	pos[2].setInt(z);
	pos[3].setInt(1);

	glRasterPos(pos);
}


void glRasterPos4f(GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{
	glRasterPos(klVec4(x,y,z,w));
}


void glRasterPos4fv(const GLfloat *v)
{
	glRasterPos(klVec4(v[0], v[1], v[2], v[3]));
}


void glRasterPos(const klVec4& nRasterPos)
{
	// this method is called by glRasterPosXf and during display list execution
	//
	rasterposition.pos = nRasterPos;


	// the rasterpostion hast to be transformed the same way
	// as a vertex passed with glVertex
	klVertex v;
	int numV=1;

	checkModelProjMatrix();

	v.pos.transform(modelProjMatrix, nRasterPos);
	clipPrimitve(&v, numV);
	if(numV==0)
		return;

	v.pos.divideByW();

	v.pos[0] >>= 1;					// convert from (-1.0 - 1.0) to (-0.5 - 0.5)
	v.pos[1] >>= 1;

	v.pos[0] += constant.half;		// convert from (-0.5 - 0.5) to  (0.0 - 1.0)
	v.pos[1] += constant.half;

	rasterizer->setRasterPos(v.pos[0], v.pos[1]);
}


void glDrawPixels(GLsizei nWidth, GLsizei nHeight, GLenum nFormat, GLenum nType, const GLvoid *nPixels)
{
	// currently we only have a simple full screen draw pixels implementation
	//
	// TODO: glDrawPixels has currently almost no spec conformance
	//

	RECORD_DISPLAYLISTS5(DrawPixels, nWidth, nHeight, nFormat, nType, nPixels)
	if(!displaylist.executeCommands)
		return;

	// Rasterizer's drawPixels() expects RGBX8888 format
	//rasterizer->drawPixels(nPixels);
	rasterizer->drawPixels(nWidth, nHeight, nFormat, nType, nPixels);
}


void glReadPixels(GLint /*x*/, GLint /*y*/, GLsizei /*width*/, GLsizei /*height*/, GLenum /*format*/, GLenum /*type*/, GLvoid* /*pixels*/)
{
	//
	// TODO: implement glReadPixels
	//
}



/**
}  // class klContext
*/
