/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klDisplayList_Commands.cpp,v 1.1 2004/04/19 13:50:51 wasserer Exp $
 *
\**************************************************************************/

#include "klDisplayList_Commands.h"

///////////////////////////////////////////////////////////
//
//                      MatrixMode
//

int
exec_MatrixMode(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_MatrixMode, nData, cmd, size);

	nContext->glMatrixMode(cmd->mode);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      LoadMatrix
//

int
exec_LoadMatrix(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_LoadMatrix, nData, cmd, size);

	nContext->glLoadMatrix(cmd->matrix);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      MultMatrix
//

int
exec_MultMatrix(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_MultMatrix, nData, cmd, size);

	nContext->glMultMatrix(cmd->matrix);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      PushMatrix
//

int
exec_PushMatrix(const void* /*nData*/, klContext* nContext)
{
	GET_SIZE(CMD_PushMatrix, size);

	nContext->glPushMatrix();
	return size;
}


///////////////////////////////////////////////////////////
//
//                      PopMatrix
//

int
exec_PopMatrix(const void* /*nData*/, klContext* nContext)
{
	GET_SIZE(CMD_PopMatrix, size);

	nContext->glPopMatrix();
	return size;
}


///////////////////////////////////////////////////////////
//
//                     LoadIdentity
//

int
exec_LoadIdentity(const void* /*nData*/, klContext* nContext)
{
	GET_SIZE(CMD_LoadIdentity, size);

	nContext->glLoadIdentity();
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Translate
//

int
exec_Translate(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Translate, nData, cmd, size);

	nContext->glTranslate(cmd->x, cmd->y, cmd->z);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Scale
//

int
exec_Scale(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Scale, nData, cmd, size);

	nContext->glScale(cmd->x, cmd->y, cmd->z);
	return size;
}


///////////////////////////////////////////////////////////
//
//                     Rotate
//

int
exec_Rotate(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Rotate, nData, cmd, size);

	nContext->glRotate(cmd->a, cmd->x, cmd->y, cmd->z);
	return size;
}


///////////////////////////////////////////////////////////
//
//                     Ortho
//

int
exec_Ortho(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Ortho, nData, cmd, size);

	nContext->glOrtho(cmd->left,cmd->right, cmd->bottom,cmd->top, cmd->nearval,cmd->farval);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    Frustum
//
int
exec_Frustum(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Frustum, nData, cmd, size);

	nContext->glFrustum(cmd->left,cmd->right, cmd->bottom,cmd->top, cmd->nearval,cmd->farval);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    ClearColor
//

int
exec_ClearColor(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_ClearColor, nData, cmd, size);

	nContext->glClearColor(cmd->color);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    ClearDepth
//

int
exec_ClearDepth(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_ClearDepth, nData, cmd, size);

	nContext->glClearDepth(cmd->depth);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Clear
//

int
exec_Clear(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Clear, nData, cmd, size);

	nContext->glClear(cmd->flags);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    Colormask
//

int
exec_ColorMask(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_ColorMask, nData, cmd, size);

	nContext->glColorMask(cmd->red, cmd->green, cmd->blue, cmd->alpha);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    Enable
//

int
exec_Enable(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Enable, nData, cmd, size);

	nContext->glEnable(cmd->select);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    Disable
//

int
exec_Disable(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Disable, nData, cmd, size);

	nContext->glDisable(cmd->select);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    DepthMask
//

int
exec_DepthMask(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_DepthMask, nData, cmd, size);

	nContext->glDepthMask(cmd->set);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    CullFace
//

int
exec_CullFace(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_CullFace, nData, cmd, size);

	nContext->glCullFace(cmd->cull);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    FrontFace
//

int
exec_FrontFace(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_FrontFace, nData, cmd, size);

	nContext->glFrontFace(cmd->face);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Scissor
//

int
exec_Scissor(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Scissor, nData, cmd, size);

	nContext->glScissor(cmd->x, cmd->y, cmd->width, cmd->height);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Fog
//

int
exec_Fog(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Fog, nData, cmd, size);

	nContext->glFog(cmd->pname, cmd->param);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      FogColor
//

int
exec_FogColor(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_FogColor, nData, cmd, size);

	nContext->glFogColor(cmd->color);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      BlendFunc
//

int
exec_BlendFunc(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_BlendFunc, nData, cmd, size);

	nContext->glBlendFunc(cmd->sfactor, cmd->dfactor);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      DepthFunc
//

int
exec_DepthFunc(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_DepthFunc, nData, cmd, size);

	nContext->glDepthFunc(cmd->param);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Lightv
//

int
exec_Lightv(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Lightv, nData, cmd, size);

	nContext->glLightv(cmd->light, cmd->select, cmd->values);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      LightModel
//

int
exec_LightModel(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_LightModel, nData, cmd, size);

	nContext->glLightModel(cmd->param, cmd->value);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      LightModelv
//

int
exec_LightModelv(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_LightModelv, nData, cmd, size);

	nContext->glLightModelv(cmd->param, cmd->values);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      ShadeModel
//

int
exec_ShadeModel(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_ShadeModel, nData, cmd, size);

	nContext->glShadeModel(cmd->model);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Material
//

int
exec_Material(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Material, nData, cmd, size);

	nContext->glMaterial(cmd->side, cmd->select, cmd->value);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Materialv
//

int
exec_Materialv(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Materialv, nData, cmd, size);

	nContext->glMaterialv(cmd->side, cmd->select, cmd->values);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      ColorMaterial
//

int
exec_ColorMaterial(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_ColorMaterial, nData, cmd, size);

	nContext->glColorMaterial(cmd->side, cmd->mode);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      PolygonMode
//

int
exec_PolygonMode(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_PolygonMode, nData, cmd, size);

	nContext->glPolygonMode(cmd->side, cmd->poly);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      TexEnv
//

int
exec_TexEnv(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_TexEnv, nData, cmd, size);

	nContext->glTexEnvi(0,0, cmd->param);			// first two parameters are fixed and klimt does not check them anyway...
	return size;
}


///////////////////////////////////////////////////////////
//
//                      PushAttrib
//

int
exec_PushAttrib(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_PushAttrib, nData, cmd, size);

	nContext->glPushAttrib(cmd->mask);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      PopAttrib
//

int
exec_PopAttrib(const void* /*nData*/, klContext* nContext)
{
	GET_SIZE(CMD_PopAttrib, size);

	nContext->glPopAttrib();
	return size;
}


///////////////////////////////////////////////////////////
//
//                      BindTexture
//

int
exec_BindTexture(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_BindTexture, nData, cmd, size);

	nContext->glBindTexture(0, cmd->texture);			// nTarget is ignored...
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Begin
//

int
exec_Begin(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Begin, nData, cmd, size);

	nContext->glBegin(cmd->type);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      End
//

int
exec_End(const void* /*nData*/, klContext* nContext)
{
	GET_SIZE(CMD_End, size);

	nContext->glEnd();
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Color
//

int
exec_Color(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Color, nData, cmd, size);

	nContext->glColor(cmd->color);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Normal
//

int
exec_Normal(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Normal, nData, cmd, size);

	nContext->glNormal(cmd->normal);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      TexCoord
//

int
exec_TexCoord(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_TexCoord, nData, cmd, size);

	nContext->glTexCoord(cmd->texCoord);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      Vertex
//

int
exec_Vertex(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_Vertex, nData, cmd, size);

	nContext->glVertex(cmd->vertex);
	return size;
}


///////////////////////////////////////////////////////////
//
//                      TexImage2D
//

int
exec_TexImage2D(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_TexImage2D, nData, cmd, size);

	nContext->getRasterizer()->updateCurrentTexture(cmd->level, cmd->internalformat, cmd->width, cmd->height,
												    cmd->border, cmd->format, cmd->type, cmd->pixels);
	return size;
}


///////////////////////////////////////////////////////////
//
//                    DrawPixels
//

int
exec_DrawPixels(const void* nData, klContext* nContext)
{
	GET_COMMAND(CMD_DrawPixels, nData, cmd, size);

	nContext->glDrawPixels(cmd->width, cmd->height, cmd->format, cmd->type, cmd->rgbxPixels);
	return size;
}

