/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klMain.cpp,v 1.9 2004/07/29 08:54:10 drgoldie Exp $
 *
\**************************************************************************/

#include "Base/klBase.h"

#ifdef _IS_WINDOWS_
	#include <windows.h>
	#include "wgl_definitions.h"

	BOOL APIENTRY DllMain(HANDLE /*hModule*/, DWORD  ul_reason_for_call, LPVOID /*lpReserved*/)
	{
	  switch (ul_reason_for_call)
	  {
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
	}
#endif //_IS_WINDOWS_

#include "klContext.h"

/////////////////////////////////////////////////////////////////////////////
// establish a mechanism for global data access (the current context)

#ifdef __SYMBIAN32__
#ifndef __EXE__
	#define _USE_SYMBIAN_TLS_
#endif // !__EXE__
#endif // __SYMBIAN32__

#ifdef _USE_SYMBIAN_TLS_
	// Symbian .APPs do not allow global data, we use the thread local
	// storage instead.

	// The thread local storage holds only 4 bytes of data
	// per thread and application. This data is not used to point
	// directly to the context, rather it points to a pointerarray
	// which holds the context. This principle allows the application or further
	// libraries to share the TLS. 
	
	// Klimt always occupies the first 32 bits
	// of this global structure. If the structure does not exist, i.e. the
	// thread local storage is not used by the application, Klimt allocates 
	// a block of 16 bytes during context creation and puts a pointer to
	// this block in the TLS. To avoid conflicts, this structure is never
	// deallocated.

	#include <e32std.h>

	inline klContext* getCurrentContext() 
	{
		if( !Dll::Tls() )
			return 0;

		return ((klContext**)Dll::Tls())[0];
	}

	inline void setCurrentContext(klContext* context)
	{
		if( !Dll::Tls() )
		{
			// TLS not used, create 16 byte block
			Dll::SetTls(new char[16]);
		}

		// put pointer into the first 4 bytes of this block
		((klContext**)Dll::Tls())[0] = context;
	}


#else

	// this is the current context
	//
	// TODO: this should be changed to use thread-local storage too,
	//       since that's the correct way for OpenGL...
	//
	klContext *currentContextPtr = 0;

	klContext* getCurrentContext() 
	{
		return currentContextPtr;
	}

	void setCurrentContext(klContext* context) 
	{
		currentContextPtr = context;
	}

#endif

// include implementation of the C-wrapper functions
//
#include "klWrapper.h"

// include implementation of wgl functions
//
#ifdef _IS_WINDOWS_
  #include "klWgl.h"
#endif //

// include implementation of egl functions
//
#include "egl_definitions.h"
#include "klEGL.h"
