/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/klWGL.h,v 1.1 2004/08/01 10:29:33 drgoldie Exp $
 *
\**************************************************************************/


#ifdef __cplusplus
extern "C" {
#endif


//
// in order not to mess with windows functions from <windows.h>
// under x86 we name our methods xxx_kl and have a
// #define statement to correct this behavior in gl.h
//


/// wglCreateContext creates a new OpenGL rendering context
/**
 *   Klimt does not provide thread-specific contexts.
 */
KLIMT_API HGLRC wglCreateContext_kl(HDC hdc)
{
	klRasterizer* rasterizer = new klRasterizer;

#ifdef _IS_WIN32_
	HWND hwnd = WindowFromDC(hdc);
#else 
	// WinCE: this will work as long as there is only one window
	// in this application
	HWND hwnd = GetActiveWindow();
#endif

	if(!hwnd)
		return 0;

	hdc = ::GetDC(hwnd);
	if(!hdc)
		return 0;

	EGLint error;
	klEGLDisplay* display = new klEGLDisplay(hdc, error);
	klEGLSurface* surface = new klEGLSurfaceWindow();

	surface->create(display, hwnd);
	rasterizer->setSurface(surface);

	return (HGLRC)new klContext(rasterizer);
}


/// wglCreateContext destroys an OpenGL rendering context
/**
 *   Klimt does not provide thread-specific contexts.
 */
KLIMT_API BOOL wglDeleteContext_kl(HGLRC hglrc)
{
	klContext* ctx = (klContext*)hglrc;

	if(ctx==NULL)
		return GL_FALSE;

	if(ctx->getRasterizer()!=NULL && ctx->getRasterizer()->getSurface()!=NULL)
	{
		klEGLSurface* surface = ctx->getRasterizer()->getSurface();
		if(surface)
		{
			surface->destroy();
			delete surface;
		}
		getCurrentContext()->getRasterizer()->setSurface(NULL);
	}

	delete ctx;
	return GL_TRUE;
}


/// wglMakeCurrent function makes a specified context the current context
/**
 *   Klimt does not provide thread-specific contexts.
 */
KLIMT_API BOOL wglMakeCurrent_kl(HDC hdc, HGLRC hglrc)
{
	setCurrentContext((klContext*)hglrc);

	if(getCurrentContext()==NULL ||
	   getCurrentContext()->getRasterizer()==NULL ||
	   getCurrentContext()->getRasterizer()->getSurface()==NULL)
		return GL_FALSE;

	klEGLSurface* surface = getCurrentContext()->getRasterizer()->getSurface();

	if(surface)
	{
		// do a quick check if we don't have to recreate the surface...
		//
		//if(surface->getNativeDisplay()==hdc)
		//	return GL_TRUE;

		klEGLDisplay* display = surface->getDisplay();
		surface->destroy();
		delete display;
		EGLint error;
		display = new klEGLDisplay(hdc, error);

#ifdef _IS_WIN32_
		HWND hwnd = WindowFromDC(hdc);
#else 
		// WinCE: this will work as long as there is only one window
		// in this application
		HWND hwnd = GetActiveWindow();
#endif

		if( !surface->create(display, hwnd) )
			return GL_FALSE;
	}

	return getCurrentContext()->getRasterizer()->setSurface(surface) ? GL_TRUE : GL_FALSE;
}


/// wglGetCurrentContext returns a handle to the current rendering context 
/**
 *   Klimt does not provide thread-specific contexts.
 */
KLIMT_API HGLRC wglGetCurrentContext_kl()
{
	return (HGLRC)getCurrentContext();
}


/// wglGetProcAddress returns the address of an OpenGL extension function
/**
 *   Klimt currently always returns NULL
 */
KLIMT_API void* wglGetProcAddress_kl(LPCSTR lpszProc)
{
	return NULL;
}


/// wglGetCurrentDC returns a handle to the device context that is associated rendering context
/**
 *   Works as specified.
 */
KLIMT_API HDC wglGetCurrentDC_kl()
{
	if(getCurrentContext()==NULL ||
	   getCurrentContext()->getRasterizer()==NULL ||
	   getCurrentContext()->getRasterizer()->getSurface()==NULL)
		return 0;

	klEGLSurface* surface = getCurrentContext()->getRasterizer()->getSurface();
	return surface->getDisplay()->getNativeDisplay();
}


/// SwapBuffers exchanges the front and back buffers
/**
 *   Works as specified.
 */
KLIMT_API BOOL SwapBuffers_kl(HDC hdc)
{
	if(getCurrentContext()==NULL ||
	   getCurrentContext()->getRasterizer()==NULL ||
	   getCurrentContext()->getRasterizer()->getSurface()==NULL)
		return EGL_FALSE;

	klEGLSurface* surface = getCurrentContext()->getRasterizer()->getSurface();

	if(!surface->flip())
		return EGL_FALSE;

	// need to set the surface again since it might have changed
	// some attributes (e.g. color buffer)
	return getCurrentContext()->getRasterizer()->setSurface(surface) ? GL_TRUE : GL_FALSE;
}


#ifdef __cplusplus
}
#endif
