/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/klWrapper.h,v 1.19 2004/07/11 19:51:48 wasserer Exp $
 *
\**************************************************************************/


#ifdef __cplusplus
extern "C" {
#endif


//
// C-Wrapper functions for the C++ API
//


/// glGetError currently always returns GL_NO_ERROR
KLIMT_API GLenum glGetError()
{
	return GL_NO_ERROR;
}

// wrapper methods implemented in klContext_oglMatrix.h
//


/// glMatrixMode specifies which matrix is the current matrix
/**
 *  Works with GL_MODELVIEW, GL_PROJECTION, GL_TEXTURE
 */
KLIMT_API void glMatrixMode(GLenum nMode)
{
	getCurrentContext()->glMatrixMode(nMode);
}


/// glLoadMatrixf replaces the current matrix with an arbitrary matrix
/**
 *   Works as specified.
 */
KLIMT_API void glLoadMatrixf(const GLfloat *nMatrix)
{
	getCurrentContext()->glLoadMatrixf(nMatrix);
}


/// glLoadMatrixd replaces the current matrix with an arbitrary matrix
/**
 *   Works as specified
 */
KLIMT_API void glLoadMatrixd(const GLdouble *nMatrix)
{
	getCurrentContext()->glLoadMatrixd(nMatrix);
}


/// glLoadMatrixx replaces the current matrix with an arbitrary matrix
/**
 *   Works as specified
 */
KLIMT_API void glLoadMatrixx(const GLfixed *nMatrix)
{
	getCurrentContext()->glLoadMatrixx(nMatrix);
}


/// glMultMatrixf multiplies the current matrix by an arbitrary matrix
/**
 *   Works as specified
 */
KLIMT_API void glMultMatrixf(const GLfloat *nMatrix)
{
	getCurrentContext()->glMultMatrixf(nMatrix);
}


/// glMultMatrixd multiplies the current matrix by an arbitrary matrix
/**
 *   Works as specified
 */
KLIMT_API void glMultMatrixd(const GLdouble *nMatrix)
{
	getCurrentContext()->glMultMatrixd(nMatrix);
}


/// glMultMatrixx multiplies the current matrix by an arbitrary matrix
/**
 *   Works as specified
 */
KLIMT_API void glMultMatrixx(const GLfixed *nMatrix)
{
	getCurrentContext()->glMultMatrixx(nMatrix);
}


/// glPushMatrix pushes the current matrix onto the stack
/**
 *   Works as specified
 */
KLIMT_API void glPushMatrix()
{
	getCurrentContext()->glPushMatrix();
}


/// glPushMatrix pop the current matrix from the stack
/**
 *   Works as specified
 */
KLIMT_API void glPopMatrix()
{
	getCurrentContext()->glPopMatrix();
}


/// glLoadIdentity replaces the current matrix with the identity matrix
/**
 *   Works as specified
 */
KLIMT_API void glLoadIdentity()
{
	getCurrentContext()->glLoadIdentity();
}


/// glTranslatef multiplies the current matrix by a translation matrix
/**
 *   Works as specified
 */
KLIMT_API void glTranslatef(GLfloat nX, GLfloat nY, GLfloat nZ)
{
	getCurrentContext()->glTranslatef(nX, nY, nZ);
}


/// glTranslated multiplies the current matrix by a translation matrix
/**
 *   Works as specified
 */
KLIMT_API void glTranslated(GLdouble nX, GLdouble nY, GLdouble nZ)
{
	getCurrentContext()->glTranslated(nX, nY, nZ);
}


/// glTranslatefv multiplies the current matrix by a translation matrix
/**
 *   Works as specified
 */
KLIMT_API void glTranslatefv(const GLfloat* nTrans)
{
	getCurrentContext()->glTranslatefv(nTrans);
}


/// glTranslatex multiplies the current matrix by a translation matrix
/**
 *   Works as specified
 */
KLIMT_API void glTranslatex(GLfixed nX, GLfixed nY, GLfixed nZ)
{
	getCurrentContext()->glTranslatex(nX, nY, nZ);
}


/// glScalef multiplies the current matrix by a general scaling matrix
/**
 *   Works as specified
 */
KLIMT_API void glScalef(GLfloat nX, GLfloat nY, GLfloat nZ)
{
	getCurrentContext()->glScalef(nX, nY, nZ);
}


/// glScaled multiplies the current matrix by a general scaling matrix
/**
 *   Works as specified
 */
KLIMT_API void glScaled(GLdouble nX, GLdouble nY, GLdouble nZ)
{
	getCurrentContext()->glScaled(nX, nY, nZ);
}


/// glScalex multiplies the current matrix by a general scaling matrix
/**
 *   Works as specified
 */
KLIMT_API void glScalex(GLfixed nX, GLfixed nY, GLfixed nZ)
{
	getCurrentContext()->glScalex(nX, nY, nZ);
}


/// glRotatef multiplies the current matrix by a rotation matrix
/**
 *   Works as specified
 */
KLIMT_API void glRotatef(GLfloat nAngle, GLfloat nX, GLfloat nY, GLfloat nZ)
{
	getCurrentContext()->glRotatef(nAngle, nX, nY, nZ);
}


/// glRotated multiplies the current matrix by a rotation matrix
/**
 *   Works as specified
 */
KLIMT_API void glRotated(GLdouble nAngle, GLdouble nX, GLdouble nY, GLdouble nZ)
{
	getCurrentContext()->glRotated(nAngle, nX, nY, nZ);
}


/// glRotatex multiplies the current matrix by a rotation matrix
/**
 *   Works as specified
 */
KLIMT_API void glRotatex(GLfixed nAngle, GLfixed nX, GLfixed nY, GLfixed nZ)
{
	getCurrentContext()->glRotatex(nAngle, nX, nY, nZ);
}


/// glOrtho multiplies the current matrix by an orthographic matrix
/**
 *   Works as specified
 */
KLIMT_API void glOrtho(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble nearval, GLdouble farval)
{
	getCurrentContext()->glOrtho(left, right, bottom, top, nearval, farval);
}


/// glOrthox multiplies the current matrix by an orthographic matrix
/**
 *   Works as specified
 */
KLIMT_API void glOrthox(GLfixed left, GLfixed right, GLfixed bottom, GLfixed top, GLfixed nearval, GLfixed farval)
{
	getCurrentContext()->glOrthox(left, right, bottom, top, nearval, farval);
}


/// glFrustum multiplies the current matrix by a perspective matrix
/**
 *   Works as specified
 */
KLIMT_API void glFrustum(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble znear, GLdouble zfar)
{
	getCurrentContext()->glFrustum(left, right, bottom, top, znear, zfar);
}


/// glFrustumx multiplies the current matrix by a perspective matrix
/**
 *   Works as specified
 */
KLIMT_API void glFrustumx(GLfixed left, GLfixed right, GLfixed bottom, GLfixed top, GLfixed znear, GLfixed zfar)
{
	getCurrentContext()->glFrustumx(left, right, bottom, top, znear, zfar);
}


// wrapper methods implemented in klContext_oglAttributeStack.h
//


/// glPushAttrib pushes the attribute stack
/**
 *   Works as specified - but only for those features
 *   that supported by Klimt. E.g. GL_ACCUM_BUFFER_BIT
 *   won't effect anything since the accumulation buffer
 *   is not supported anyhow...
 */
KLIMT_API void glPushAttrib(GLbitfield nMask)
{
	getCurrentContext()->glPushAttrib(nMask);
}


/// glPopAttrib pops the attribute stack
/**
 *   Works as specified - but only for those features
 *   that supported by Klimt. E.g. GL_ACCUM_BUFFER_BIT
 *   won't effect anything since the accumulation buffer
 *   is not supported anyhow...
 */
KLIMT_API void glPopAttrib()
{
	getCurrentContext()->glPopAttrib();
}


/// glPushClientAttrib saves groups of client-state variables on the client-attribute stack
/**
 *   Works as specified - but only for those features
 *   that supported by Klimt. E.g. GL_CLIENT_PIXEL_STORE_BIT
 *   won't effect anything...
 */
KLIMT_API void glPushClientAttrib(GLbitfield mask)
{
	getCurrentContext()->glPushClientAttrib(mask);
}


/// glPopClientAttrib restores groups of client-state variables from the client-attribute stack
/**
 *   Works as specified - but only for those features
 *   that supported by Klimt. E.g. GL_CLIENT_PIXEL_STORE_BIT
 *   won't effect anything...
 */
KLIMT_API void glPopClientAttrib()
{
	getCurrentContext()->glPopClientAttrib();
}



// wrapper methods implemented in klContext_oglPrimitve.h
//


/// glDisableClientState enables arrays respectively
/**
 *   GL_INDEX_ARRAY and GL_EDGE_FLAG_ARRAY are not supported
 */
KLIMT_API void glEnableClientState(GLenum nState)
{
	getCurrentContext()->glEnableClientState(nState);
}


/// glDisableClientState disables arrays respectively
/**
 *   GL_INDEX_ARRAY and GL_EDGE_FLAG_ARRAY are not supported
 */
KLIMT_API void glDisableClientState(GLenum nState)
{
	getCurrentContext()->glDisableClientState(nState);
}


/// glIndexPointer defines an array of color indexes
/**
 *   Silently ignored - Klimt does not support indexed colors in general
 */

KLIMT_API void glIndexPointer(GLenum type, GLsizei stride, const GLvoid *pointer)
{
	getCurrentContext()->glIndexPointer(type, stride, pointer);
}


/// glColorPointer specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glColorPointer(GLint size, GLenum type, GLsizei stride, const GLvoid *pointer)
{
	getCurrentContext()->glColorPointer(size, type, stride, pointer);
}


/// glNormalPointer specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glNormalPointer(GLenum type, GLsizei stride, const GLvoid *pointer)
{
	getCurrentContext()->glNormalPointer(type, stride, pointer);
}


/// glTexCoordPointer specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glTexCoordPointer(GLint size, GLenum type, GLsizei stride, const GLvoid *pointer)
{
	getCurrentContext()->glTexCoordPointer(size, type, stride, pointer);
}


/// glVertexPointer specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertexPointer(GLint size, GLenum type, GLsizei stride, const GLvoid *pointer)
{
	getCurrentContext()->glVertexPointer(size, type, stride, pointer);
}


/// glDrawElements renders primitives from array data
/**
 *   Works as specified
 */
KLIMT_API void glDrawElements(GLenum polymode, GLsizei count, GLenum datatype, const GLvoid *indices)
{
	getCurrentContext()->glDrawElements(polymode, count, datatype, indices);
}


/// glDrawArrays specifies multiple primitives to render
/**
 *   Silently ignored
 */
KLIMT_API void glDrawArrays(GLenum mode, GLint first, GLsizei count)
{
	getCurrentContext()->glDrawArrays(mode, first, count);
}


/// glArrayElement specifies the array elements used to render a vertex
/**
 *   Silently ignored
 */
KLIMT_API void glArrayElement(GLint index)
{
	getCurrentContext()->glArrayElement(index);
}


/// glInterleavedArrays simultaneously specifies and enables several interleaved arrays in a larger aggregate array
/**
 *   Silently ignored
 */
KLIMT_API void glInterleavedArrays(GLenum format, GLsizei stride, const GLvoid *pointer)
{
	getCurrentContext()->glInterleavedArrays(format, stride, pointer);
}


// wrapper methods implemented in klContext_oglSetup.h
//


/// glClearColor specifies clear values for the color buffers
/**
 *   The alpha value is ignored
 */
KLIMT_API void glClearColor(GLclampf nR, GLclampf nG, GLclampf nB, GLclampf nA)
{
	getCurrentContext()->glClearColor(nR, nG, nB, nA);
}


/// glClearColorx specifies clear values for the color buffer
/**
 *   The alpha value is ignored
 */
KLIMT_API void glClearColorx(GLfixed nR, GLfixed nG, GLfixed nB, GLfixed nA)
{
	getCurrentContext()->glClearColorx(nR, nG, nB, nA);
}


/// glClearDepth specifies clear values for the depth buffer
/**
 *   Works as specified
 */
KLIMT_API void glClearDepth(GLclampd nDepth)
{
	getCurrentContext()->glClearDepth(nDepth);
}


/// glClearDepthx specifies clear values for the color buffers
/**
 *   Works as specified
 */
KLIMT_API void glClearDepthx(GLfixed nDepth)
{
	getCurrentContext()->glClearDepthx(nDepth);
}


/// glClear clears buffers to preset values
/**
 *   Only GL_COLOR_BUFFER_BIT and GL_DEPTH_BUFFER_BIT are used.
 */
KLIMT_API void glClear(GLbitfield nFlags)
{
	getCurrentContext()->glClear(nFlags);
}


/// glPolygonMode selects a polygon rasterization mode
/**
 *   Works as specified
 */
KLIMT_API void glPolygonMode(GLenum nSide, GLenum nPoly)
{
	getCurrentContext()->glPolygonMode(nSide, nPoly);
}


/// glEnable enables capabilities
/**
 *   The following list of parameters is currently supported:
 *   - GL_BLEND
 *   - GL_COLOR_MATERIAL
 *   - GL_CULL_FACE
 *   - GL_DEPTH_TEST
 *   - GL_FOG
 *   - GL_LIGHTING
 *   - GL_LIGHT0-7
 *   - GL_NORMALIZE
 *   - GL_SCISSOR_TEST
 *   - GL_TEXTURE_2D
 */
KLIMT_API void glEnable(GLenum nSelect)
{
	getCurrentContext()->glEnable(nSelect);
}


/// glDisable disables capabilities
/**
 *  @see glEnable for a list of supported parameters
 */
KLIMT_API void glDisable(GLenum nSelect)
{
	getCurrentContext()->glDisable(nSelect);
}


/// glIsEnabled tests whether a capability is enabled
/**
 *  @see glEnable for a list of supported parameters
 */
KLIMT_API GLboolean glIsEnabled(GLenum nSelect)
{
	return getCurrentContext()->glIsEnabled(nSelect);
}


/// glDepthMask enables or disables writing into the depth buffer
/**
 *   Works as specified
 */
KLIMT_API void glDepthMask(GLboolean nSet)
{
	getCurrentContext()->glDepthMask(nSet);
}


/// glCullFace specifies whether front- or back-facing facets can be culled
/**
 *   Works as specified
 */
KLIMT_API void glCullFace(GLenum nCull)
{
	getCurrentContext()->glCullFace(nCull);
}


/// glFrontFace defines front- and back-facing polygons
/**
 *   Works as specified
 */
KLIMT_API void glFrontFace(GLenum nFace)
{
	getCurrentContext()->glFrontFace(nFace);
}


/// glGetBooleanv returns the value or values of a selected parameter
/**
 *   Only GL_RGBA_MODE is supported
 */
KLIMT_API void glGetBooleanv(GLenum nType, GLboolean *nParams)
{
	getCurrentContext()->glGetBooleanv(nType, nParams);
}


/// glGetFloatv returns the value or values of a selected parameter
/**
 *   Supported parameters for nType:
 *   - GL_PROJECTION_MATRIX
 *   - GL_MODELVIEW_MATRIX
 *   - GL_LIGHT_MODEL_AMBIENT
 *   - GL_CURRENT_COLOR
 *   - GL_LINE_WIDTH_RANGE
 *   - GL_POINT_SIZE_RANGE
 */
KLIMT_API void glGetFloatv(GLenum nType, GLfloat* nParams)
{
	getCurrentContext()->glGetFloatv(nType, nParams);
}


/// glGetIntegerv returns the value or values of a selected parameter
/**
 *   Supported parameters for nType:
 *   - GL_FRONT_FACE
 *   - GL_CULL_FACE_MODE
 *   - GL_LIGHT_MODEL_AMBIENT
 *   - GL_MAX_LIGHTS
 *   - GL_RED_BITS
 *   - GL_GREEN_BITS
 *   - GL_BLUE_BITS
 *   - GL_ALPHA_BITS
 *   - GL_MAX_VIEWPORT_DIMS
 *   - GL_MAX_CLIP_PLANES
 *   - GL_MAX_TEXTURE_SIZE
 */
KLIMT_API void glGetIntegerv(GLenum nType, GLint* nParams)
{
	getCurrentContext()->glGetIntegerv(nType, nParams);
}


/// glGetString returns the value or values of a selected parameter
/**
 *   Supported parameters:
 *   - GL_VENDOR
 *   - GL_RENDERER
 *   - GL_VERSION
 *   - GL_EXTENSIONS
 */
KLIMT_API const GLubyte* glGetString(GLenum nName)
{
	return getCurrentContext()->glGetString(nName);
}


/// glColorMask enables and disables writing of frame-buffer color components
/**
 *   The alpha value is ignored.
 */
KLIMT_API void glColorMask(GLboolean nRed, GLboolean nGreen, GLboolean nBlue, GLboolean nAlpha)
{
	getCurrentContext()->glColorMask(nRed, nGreen, nBlue, nAlpha);
}


/// glScissor defines the scissor box
/**
 *   Works as specified
 */
KLIMT_API void glScissor(GLint x, GLint y, GLsizei width, GLsizei height)
{
	getCurrentContext()->glScissor(x, y, width, height);
}


/// glFogi specifies fog parameters
/**
*   Only linear fog is supported.
*/
KLIMT_API void glFogi(GLenum pname, GLint param)
{
	getCurrentContext()->glFogi(pname, param);
}


/// glFogfv specifies fog parameters
/**
*    Only linear fog is supported.
*/
KLIMT_API void glFogf(GLenum pname, GLfloat param)
{
	getCurrentContext()->glFogf(pname, param);
}


/// glFogfv specifies fog parameters
/**
*    Only linear fog is supported.
*/
KLIMT_API void glFogfv(GLenum pname, const GLfloat *params)
{
	getCurrentContext()->glFogfv(pname, params);
}


/// glFogiv specifies fog parameters
/**
*    Only linear fog is supported.
*/
KLIMT_API void glFogiv(GLenum pname, const GLint *params)
{
	getCurrentContext()->glFogiv(pname, params);
}


/// glBlendFunc specifies pixel arithmetic
/**
 *   glBlendFunc works with all parameters except destination alpha
 *   (there is no alpha buffer support at the moment in Klimt)
 *
 */
KLIMT_API void glBlendFunc(GLenum sfactor, GLenum dfactor)
{
	getCurrentContext()->glBlendFunc(sfactor, dfactor);
}


/// glDepthFunc specifies the value used for depth-buffer comparisons
/**
 *   Works as specified
 */
KLIMT_API void glDepthFunc(GLenum nFunc)
{
	getCurrentContext()->glDepthFunc(nFunc);
}


/// glFlush forces execution of functions in finite time
/**
 *   Works as specified
 */
KLIMT_API void glFlush()
{
	getCurrentContext()->glFlush();
}


/// glDrawBuffer specifies which color buffers are to be drawn into
/**
 *   Works as specified
 */
KLIMT_API void glDrawBuffer(GLenum mode)
{
	getCurrentContext()->glDrawBuffer(mode);
}


/// glReadBuffer selects a color buffer source for pixels
/**
 *   Works as specified
 */
KLIMT_API void glReadBuffer(GLenum mode)
{
	getCurrentContext()->glReadBuffer(mode);
}


/// glHint specifies implementation-specific hints
/**
 *   Currently takes only one hint: KL_WBUFFERING
 *     Set GL_TRUE or GL_FALSE to enable/disable w-buffering
 *     instead of z-buffering.
 */
KLIMT_API void glHint(GLenum target, GLenum mode)
{
	getCurrentContext()->glHint(target, mode);
}


// wrapper methods implemented in klContext_oglLight.h
//


/// glLightfv sets light-source parameters
/**
 *   Spot lights and attenuation are not supported
 */
KLIMT_API void glLightfv(GLenum nLight, GLenum nSelect, const GLfloat* nValues)
{
	getCurrentContext()->glLightfv(nLight, nSelect, nValues);
}


/// glLightxv sets light-source parameters
/**
 *   Spot lights and attenuation are not supported
 */
KLIMT_API void glLightxv(GLenum nLight, GLenum nSelect, const GLfixed* nValues)
{
	getCurrentContext()->glLightxv(nLight, nSelect, nValues);
}


/// glLightfv sets light-source parameters
/**
 *   Spot lights and attenuation are not supported.
 *   These paramters always return zero values.
 */
KLIMT_API void glGetLightfv(GLenum nLight, GLenum nPname, GLfloat *nParams)
{
	getCurrentContext()->glGetLightfv(nLight, nPname, nParams);
}


/// glLightModeli sets the lighting model parameters
/**
 *   GL_LIGHT_MODEL_TWO_SIDE is not supported
 */
KLIMT_API void glLightModeli(GLenum nParam, GLint nValue)
{
	getCurrentContext()->glLightModeli(nParam, nValue);
}


/// glLightModelf sets the lighting model parameters
/**
 *   GL_LIGHT_MODEL_TWO_SIDE is not supported
 */
KLIMT_API void glLightModelf(GLenum nParam, GLfloat nValue)
{
	getCurrentContext()->glLightModelf(nParam, nValue);
}


/// glLightModeld sets the lighting model parameters
/**
 *   GL_LIGHT_MODEL_TWO_SIDE is not supported
 */
KLIMT_API void glLightModeld(GLenum nParam, GLdouble nValue)
{
	getCurrentContext()->glLightModeld(nParam, nValue);
}


/// glLightModelx sets the lighting model parameters
/**
*   GL_LIGHT_MODEL_TWO_SIDE is not supported
*/
KLIMT_API void glLightModelx(GLenum nParam, GLfixed nValue)
{
	getCurrentContext()->glLightModelx(nParam, nValue);
}


/// glLightModelfv sets the lighting model parameters
/**
 *   Only GL_LIGHT_MODEL_AMBIENT is supported
 */
KLIMT_API void glLightModelfv(GLenum nParam, const GLfloat* nValues)
{
	getCurrentContext()->glLightModelfv(nParam, nValues);
}


/// glLightModeldv sets the lighting model parameters
/**
 *   Only GL_LIGHT_MODEL_AMBIENT is supported
 */
KLIMT_API void glLightModeldv(GLenum nParam, const GLdouble* nValues)
{
	getCurrentContext()->glLightModeldv(nParam, nValues);
}


/// glLightModelxv sets the lighting model parameters
/**
*   Only GL_LIGHT_MODEL_AMBIENT is supported
*/
KLIMT_API void glLightModelxv(GLenum nParam, const GLfixed* nValues)
{
	getCurrentContext()->glLightModelxv(nParam, nValues);
}


/// glShadeModel selects flat or smooth shading
/**
 *   Works as specified
 */
KLIMT_API void glShadeModel(GLenum nModel)
{
	getCurrentContext()->glShadeModel(nModel);
}


// wrapper methods implemented in klContext_oglMaterial.h
//


/// glMaterialf specifies material parameters for the lighting model
/**
 *   Works as specified
 */
KLIMT_API void glMaterialf(GLenum nSide, GLenum nSelect, GLfloat nValue)
{
	getCurrentContext()->glMaterialf(nSide, nSelect, nValue);
}


/// glMateriald specifies material parameters for the lighting model
/**
 *   Works as specified
 */
KLIMT_API void glMateriald(GLenum nSide, GLenum nSelect, GLdouble nValue)
{
	getCurrentContext()->glMateriald(nSide, nSelect, nValue);
}


/// glMaterialx specifies material parameters for the lighting model
/**
 *   Works as specified
 */
KLIMT_API void glMaterialx(GLenum nSide, GLenum nSelect, GLfixed nValue)
{
	getCurrentContext()->glMaterialx(nSide, nSelect, nValue);
}


/// glMaterialfv specifies material parameters for the lighting model
/**
 *   GL_COLOR_INDEXES is not supported
 */
KLIMT_API void glMaterialfv(GLenum nSide, GLenum nSelect, const GLfloat* nValues)
{
	getCurrentContext()->glMaterialfv(nSide, nSelect, nValues);
}


/// glMaterialdv specifies material parameters for the lighting model
/**
 *   GL_COLOR_INDEXES is not supported
 */
KLIMT_API void glMaterialdv(GLenum nSide, GLenum nSelect, const GLdouble* nValues)
{
	getCurrentContext()->glMaterialdv(nSide, nSelect, nValues);
}


/// glMaterialiv specifies material parameters for the lighting model
/**
 *   GL_COLOR_INDEXES is not supported
 */
KLIMT_API void glMaterialiv(GLenum nSide, GLenum nSelect, const GLint* nValues)
{
	getCurrentContext()->glMaterialiv(nSide, nSelect, nValues);
}


/// glMaterialxv specifies material parameters for the lighting model
/**
*   GL_COLOR_INDEXES is not supported
*/
KLIMT_API void glMaterialxv(GLenum nSide, GLenum nSelect, const GLfixed* nValues)
{
	getCurrentContext()->glMaterialxv(nSide, nSelect, nValues);
}


/// glColorMaterial causes a material color to track the current color
/**
 *   Works as specified
 */
KLIMT_API void glColorMaterial(GLenum nSide, GLenum nMode)
{
	getCurrentContext()->glColorMaterial(nSide, nMode);
}


/// glGetMaterialfv returns material parameters
/**
 *   GL_COLOR_INDEXES is not supported
 */
KLIMT_API void glGetMaterialfv(GLenum nSide, GLenum nSelect, GLfloat* nValues)
{
	getCurrentContext()->glGetMaterialfv(nSide, nSelect, nValues);
}


/// glGetMaterialiv returns material parameters
/**
 *   GL_COLOR_INDEXES is not supported
 */
KLIMT_API void glGetMaterialiv(GLenum nSide, GLenum nSelect, GLint* nValues)
{
	getCurrentContext()->glGetMaterialiv(nSide, nSelect, nValues);
}


/// glGetMaterialxv returns material parameters
/**
 *   GL_COLOR_INDEXES is not supported
 */
KLIMT_API void glGetMaterialxv(GLenum nSide, GLenum nSelect, GLfixed* nValues)
{
	getCurrentContext()->glGetMaterialxv(nSide, nSelect, nValues);
}


// wrapper methods implemented in klContext_oglPixels.h
//


/// glRasterPos2f specifies the raster position for pixel operations
/**
 *   Works as specified
 */
KLIMT_API void glRasterPos2f(GLfloat x, GLfloat y)
{
	getCurrentContext()->glRasterPos2f(x, y);
}


/// glRasterPos2i specifies the raster position for pixel operations
/**
 *   Works as specified
 */
KLIMT_API void glRasterPos2i(GLint x, GLint y)
{
	getCurrentContext()->glRasterPos2i(x, y);
}


/// glRasterPos3f specifies the raster position for pixel operations
/**
 *   Works as specified
 */
KLIMT_API void glRasterPos3f(GLfloat x, GLfloat y, GLfloat z)
{
	getCurrentContext()->glRasterPos3f(x, y, z);
}


/// glRasterPos3i specifies the raster position for pixel operations
/**
 *   Works as specified
 */
KLIMT_API void glRasterPos3i(GLint x, GLint y, GLint z)
{
	getCurrentContext()->glRasterPos3i(x, y, z);
}


/// glRasterPos4f specifies the raster position for pixel operations
/**
 *   Works as specified
 */
KLIMT_API void glRasterPos4f(GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{
	getCurrentContext()->glRasterPos4f(x, y, z, w);
}


/// glRasterPos4fv specifies the raster position for pixel operations
/**
 *   Works as specified
 */
KLIMT_API void glRasterPos4fv(const GLfloat *v)
{
	getCurrentContext()->glRasterPos4fv(v);
}


/// glDrawPixels writes a block of pixels to the framebuffer
/**
 *   - Only GL_RGBA for nFormat (alpha value ignored) and
 *     GL_UNSIGNED_BYTE for nType are supported.
 *   - nWidth currently has to be full screen width.
 *   - glDrawPixels is currently only draws
 *     full screen images.
 */
KLIMT_API void glDrawPixels(GLsizei nWidth, GLsizei nHeight, GLenum nFormat, GLenum nType, const GLvoid *nPixels)
{
	getCurrentContext()->glDrawPixels(nWidth, nHeight, nFormat, nType, nPixels);
}


/// glReadPixels is silently ignored
KLIMT_API void glReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels)
{
	getCurrentContext()->glReadPixels(x, y, width, height, format, type, pixels);
}


// wrapper methods implemented in klContext_oglBeginEnd.h
//

/// glBegin delimits the vertices of a primitive or a group of like primitives
/**
 *   Works with all primitve types as specified
 */
KLIMT_API void glBegin(GLenum nType)
{
	getCurrentContext()->glBegin(nType);
}


/// glEnd delimits the vertices of a primitive or a group of like primitives
/**
 *   Works with all primitve types as specified
 */
KLIMT_API void glEnd()
{
	getCurrentContext()->glEnd();
}


/// glColor3f sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3f(GLfloat nR, GLfloat nG, GLfloat nB)
{
	getCurrentContext()->glColor3f(nR, nG, nB);
}


/// glColor3fv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3fv(const GLfloat *nV)
{
	getCurrentContext()->glColor3fv(nV);
}


/// glColor3i sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3i(GLint nR, GLint nG, GLint nB)
{
	getCurrentContext()->glColor3i(nR, nG, nB);
}


/// glColor3iv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3iv(const GLint* nV)
{
	getCurrentContext()->glColor3iv(nV);
}


/// glColor3ub sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3ub(GLubyte nR, GLubyte nG, GLubyte nB)
{
	getCurrentContext()->glColor3ub(nR, nG, nB);
}


/// glColor3ubv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3ubv(const GLubyte *nV)
{
	getCurrentContext()->glColor3ubv(nV);
}


/// glColor3x sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3x(GLfixed nR, GLfixed nG, GLfixed nB)
{
	getCurrentContext()->glColor3x(nR, nG, nB);
}


/// glColor3xv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor3xv(const GLfixed *nV)
{
	getCurrentContext()->glColor3xv(nV);
}


/// glColor4f sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4f(GLfloat nR, GLfloat nG, GLfloat nB, GLfloat nA)
{
	getCurrentContext()->glColor4f(nR, nG, nB, nA);
}


/// glColor4fv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4fv(const GLfloat *nV)
{
	getCurrentContext()->glColor4fv(nV);
}


/// glColor4i sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4i(GLint nR, GLint nG, GLint nB, GLint nA)
{
	getCurrentContext()->glColor4i(nR, nG, nB, nA);
}


/// glColor4iv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4iv(const GLint* nV)
{
	getCurrentContext()->glColor4iv(nV);
}


/// glColor4ub sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4ub(GLubyte nR, GLubyte nG, GLubyte nB, GLubyte nA)
{
	getCurrentContext()->glColor4ub(nR, nG, nB, nA);
}


/// glColor4ubv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4ubv(const GLubyte *nV)
{
	getCurrentContext()->glColor4ubv(nV);
}


/// glColor4x sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4x(GLfixed nR, GLfixed nG, GLfixed nB, GLfixed nA)
{
	getCurrentContext()->glColor4x(nR, nG, nB, nA);
}


/// glColor4xv sets the current color
/**
 *   Works as specified
 */
KLIMT_API void glColor4xv(const GLfixed *nV)
{
	getCurrentContext()->glColor4xv(nV);
}


/// glNormal3f sets the current normal vector
/**
 *   Works as specified
 */
KLIMT_API void glNormal3f(GLfloat nX, GLfloat nY, GLfloat nZ)
{
	getCurrentContext()->glNormal3f(nX, nY, nZ);
}


/// glNormal3fv sets the current normal vector
/**
 *   Works as specified
 */
KLIMT_API void glNormal3fv(const GLfloat *nV)
{
	getCurrentContext()->glNormal3fv(nV);
}


/// glNormal3d sets the current normal vector
/**
 *   Works as specified
 */
KLIMT_API void glNormal3d(GLdouble nX, GLdouble nY, GLdouble nZ)
{
	getCurrentContext()->glNormal3d(nX, nY, nZ);
}


/// glNormal3dv sets the current normal vector
/**
 *   Works as specified
 */
KLIMT_API void glNormal3dv(const GLdouble *nV)
{
	getCurrentContext()->glNormal3dv(nV);
}


/// glNormal3x sets the current normal vector
/**
 *   Works as specified
 */
KLIMT_API void glNormal3x(GLfixed nX, GLfixed nY, GLfixed nZ)
{
	getCurrentContext()->glNormal3x(nX, nY, nZ);
}


/// glNormal3xv sets the current normal vector
/**
 *   Works as specified
 */
KLIMT_API void glNormal3xv(const GLfixed *nV)
{
	getCurrentContext()->glNormal3xv(nV);
}


/// glTexCoord2f sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord2f(GLfloat nX, GLfloat nY)
{
	getCurrentContext()->glTexCoord2f(nX, nY);
}


/// glTexCoord2fv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord2fv(const GLfloat *nV)
{
	getCurrentContext()->glTexCoord2fv(nV);
}


/// glTexCoord3f sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord3f(GLfloat nX, GLfloat nY, GLfloat nZ)
{
	getCurrentContext()->glTexCoord3f(nX, nY, nZ);
}


/// glTexCoord3fv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord3fv(const GLfloat *nV)
{
	getCurrentContext()->glTexCoord3fv(nV);
}


/// glTexCoord4f sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord4f(GLfloat nX, GLfloat nY, GLfloat nZ, GLfloat nW)
{
	getCurrentContext()->glTexCoord4f(nX, nY, nZ, nW);
}


/// glTexCoord4fv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord4fv(const GLfloat *nV)
{
	getCurrentContext()->glTexCoord4fv(nV);
}


/// glTexCoord2d sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord2d(GLdouble nX, GLdouble nY)
{
	getCurrentContext()->glTexCoord2d(nX, nY);
}


/// glTexCoord2dv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord2dv(const GLdouble *nV)
{
	getCurrentContext()->glTexCoord2dv(nV);
}


/// glTexCoord3d sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord3d(GLdouble nX, GLdouble nY, GLdouble nZ)
{
	getCurrentContext()->glTexCoord3d(nX, nY, nZ);
}


/// glTexCoord3dv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord3dv(const GLdouble *nV)
{
	getCurrentContext()->glTexCoord3dv(nV);
}


/// glTexCoord4d sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord4d(GLdouble nX, GLdouble nY, GLdouble nZ, GLdouble nW)
{
	getCurrentContext()->glTexCoord4d(nX, nY, nZ, nW);
}


/// glTexCoord4dv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord4dv(const GLdouble *nV)
{
	getCurrentContext()->glTexCoord4dv(nV);
}


/// glTexCoord2x sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord2x(GLfixed nX, GLfixed nY)
{
	getCurrentContext()->glTexCoord2x(nX, nY);
}


/// glTexCoord2xv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord2xv(const GLfixed *nV)
{
	getCurrentContext()->glTexCoord2xv(nV);
}


/// glTexCoord3x sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord3x(GLfixed nX, GLfixed nY, GLfixed nZ)
{
	getCurrentContext()->glTexCoord3x(nX, nY, nZ);
}


/// glTexCoord3xv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord3xv(const GLfixed *nV)
{
	getCurrentContext()->glTexCoord3xv(nV);
}


/// glTexCoord4x sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord4x(GLfixed nX, GLfixed nY, GLfixed nZ, GLfixed nW)
{
	getCurrentContext()->glTexCoord4x(nX, nY, nZ, nW);
}


/// glTexCoord4xv sets the current texture coordinates
/**
 *   Works as specified
 */
KLIMT_API void glTexCoord4xv(const GLfixed *nV)
{
	getCurrentContext()->glTexCoord4xv(nV);
}


/// glVertex2s specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex2s(GLshort x, GLshort y)
{
	getCurrentContext()->glVertex2s(x, y);
}


/// glVertex2i specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex2i(GLint x, GLint y)
{
	getCurrentContext()->glVertex2i(x, y);
}


/// glVertex2f specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex2f(GLfloat x, GLfloat y)
{
	getCurrentContext()->glVertex2f(x, y);
}


/// glVertex3x specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3x(GLfixed nX, GLfixed nY, GLfixed nZ)
{
	getCurrentContext()->glVertex3x(nX, nY, nZ);
}


/// glVertex3xv specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3xv(const GLfixed* nVertex)
{
	getCurrentContext()->glVertex3xv(nVertex);
}


/// glVertex3f specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3f(GLfloat nX, GLfloat nY, GLfloat nZ)
{
	getCurrentContext()->glVertex3f(nX, nY, nZ);
}


/// glVertex3fv specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3fv(const GLfloat *nVertex)
{
	getCurrentContext()->glVertex3fv(nVertex);
}


/// glVertex3d specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3d(GLdouble nX, GLdouble nY, GLdouble nZ)
{
	getCurrentContext()->glVertex3d(nX, nY, nZ);
}


/// glVertex3dv specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3dv(const GLdouble *nVertex)
{
	getCurrentContext()->glVertex3dv(nVertex);
}


/// glVertex3i specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3i(GLint nX, GLint nY, GLint nZ)
{
	getCurrentContext()->glVertex3i(nX, nY, nZ);
}


/// glVertex3iv specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex3iv(const GLint *nVertex)
{
	getCurrentContext()->glVertex3iv(nVertex);
}


/// glVertex4fv specifies a vertex
/**
 *   Works as specified
 */
KLIMT_API void glVertex4fv(const GLfloat *nVertex)
{
	getCurrentContext()->glVertex4fv(nVertex);
}


// wrapper methods implemented in klContext_oglTexture.h
//


/// glGenTextures generates texture names
/**
 *   Works as specified
 */
KLIMT_API void glGenTextures(GLsizei n, GLuint *textures)
{
	getCurrentContext()->glGenTextures(n, textures);
}


/// glDeleteTextures deletes named textures
/**
 *   Works as specified
 */
KLIMT_API void glDeleteTextures(GLsizei n, const GLuint *textures)
{
	getCurrentContext()->glDeleteTextures(n, textures);
}


/// glBindTexture enables the creation of a named texture that is bound to a texture target
/**
 *   nTarget must be GL_TEXTURE_2D.
 */
KLIMT_API void glBindTexture(GLenum nTarget, GLuint nTexture)
{
	getCurrentContext()->glBindTexture(nTarget, nTexture);
}


/// glGetTexLevelParameteriv returns texture parameter values for a specific level of detail
/**
 *   target must be GL_TEXTURE_2D.\n
 *   level is ignored.\n
 *   pname is allowed to be
 *   - GL_TEXTURE_WIDTH
 *   - GL_TEXTURE_HEIGHT
 *   - GL_TEXTURE_INTERNAL_FORMAT
 */
KLIMT_API void glGetTexLevelParameteriv(GLenum target, GLint level, GLenum pname, GLint *params)
{
	getCurrentContext()->glGetTexLevelParameteriv(target, level, pname, params);
}


/// glTexImage2D specifies a two-dimensional texture image
/**
 *   glTexImage2D works with the following restrictions:
 *   - glTexImage2D accepts only RGB24 or RGBX32 images.
 *   - nTarget must be GL_TEXTURE_2D.
 *   - nInternalFormat must 3 or 4.
 *   - nBorder must is ignored.
 *   - nFormat must be GL_RGB or GL_RGBA.
 *   - nPixelType must be GL_UNSIGNED_BYTE
 */
KLIMT_API void glTexImage2D(GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height,
							GLint border, GLenum format, GLenum type, const GLvoid *pixels)
{
	getCurrentContext()->glTexImage2D(target, level, internalformat, width, height, border, format, type, pixels);
}


/// glCopyTexImage2D copies pixels from the framebuffer into a two-dimensional texture image
/**
*   Currently silently ignored
*/
KLIMT_API void glCopyTexImage2D(GLenum target, GLint level, GLenum internalFormat, GLint x, GLint y,
								GLsizei width, GLsizei height, GLint border)
{
	getCurrentContext()->glCopyTexImage2D(target, level, internalFormat, x, y, width, height, border);
}


/// glCopyTexSubImage2D copies a sub-image of a two-dimensional texture image from the framebuffer
/**
*   Currently silently ignored
*/
KLIMT_API void glCopyTexSubImage2D(GLenum target, GLint level, GLint xoffset, GLint yoffset,
								   GLint x, GLint y, GLsizei width, GLsizei height)
{
	getCurrentContext()->glCopyTexSubImage2D(target, level, xoffset, yoffset, x, y, width, height);
}


/// glTexEnvi sets texture environment parameters
/**
*   Only GL_MODULATE and GL_DECAL are supported
*/
KLIMT_API void glTexEnvi(GLenum target, GLenum pname, GLint param)
{
	getCurrentContext()->glTexEnvi(target, pname, param);
}


/// glTexEnvf sets texture environment parameters
/**
*   Only GL_MODULATE and GL_DECAL are supported
*/
KLIMT_API void glTexEnvf(GLenum target, GLenum pname, GLfloat param)
{
	getCurrentContext()->glTexEnvf(target, pname, param);
}


/// glTexEnvx sets texture environment parameters
/**
*   Only GL_MODULATE and GL_DECAL are supported
*/
KLIMT_API void glTexEnvx(GLenum target, GLenum pname, GLfixed param)
{
	getCurrentContext()->glTexEnvx(target, pname, param);
}


// wrapper methods implemented in klContext_oglDisplayList.h
//


/// glCallLists executes a list of display lists
/**
 *   Works as specified
 */
KLIMT_API void glCallLists(GLsizei n, GLenum type, const GLvoid *lists)
{
	getCurrentContext()->glCallLists(n, type, lists);
}


/// glCallList executes a display list
/**
 *   Works as specified
 */
KLIMT_API void glCallList(GLuint list)
{
	getCurrentContext()->glCallList(list);
}


/// glGenLists generates a contiguous set of empty display lists
/**
 *   Works as specified
 */
KLIMT_API GLuint glGenLists(GLsizei range)
{
	return getCurrentContext()->glGenLists(range);
}


/// glDeleteLists deletes a contiguous group of display lists
/**
 *   Works as specified
 */
KLIMT_API void glDeleteLists(GLuint list, GLsizei range)
{
	getCurrentContext()->glDeleteLists(list, range);
}


/// glNewList creates or replaces a display list
/**
 *   Works as specified
 */
KLIMT_API void glNewList(GLuint list, GLenum mode)
{
	getCurrentContext()->glNewList(list, mode);
}


/// glEndList closes display list compiling
/**
 *   Works as specified
 */
KLIMT_API void glEndList()
{
	getCurrentContext()->glEndList();
}


/// glListBase sets the display list base for glCallLists
KLIMT_API void glListBase(GLuint base)
{
	getCurrentContext()->glListBase(base);
}


/// gllsList tests for display list existence
KLIMT_API GLboolean glIsList(GLuint list)
{
	return getCurrentContext()->glIsList(list);
}



// Klimt "extensions"
//


/// klCreateDefaultContext creates a render context
/**
 *   Use this method in case you don't want to use
 *   wgl or egl methods.
 */
KLIMT_API void klCreateDefaultContext()
{
	if(getCurrentContext())
		delete getCurrentContext();
	setCurrentContext(new klContext(new klRasterizer));
}

KLIMT_API void klDestroyContext()
{
	delete getCurrentContext();
}


/// klSetBufferSize sets the size of the color and depth buffer
KLIMT_API void klSetBufferSize(int nWidth, int nHeight)
{
	getCurrentContext()->klSetBufferSize(nWidth, nHeight);
}


/// klSetColorBuffer sets the color buffer Klimt renders to
/**
 *   The color buffer must be large enough to hold all
 *   pixels as specified in klSetBufferSize.\n
 *   This works only with the software rasterizer.
 */
KLIMT_API void klSetColorBuffer(void* nMem)
{
	getCurrentContext()->klSetColorBuffer(nMem);
}


/// klSetColorBuffer sets the color buffer Klimt renders to
/**
 *   This works only with the software rasterizer.
 */
KLIMT_API void* klGetColorBuffer()
{
	return getCurrentContext()->klGetColorBuffer();
}

KLIMT_API void* klGetDepthBuffer()
{
	return getCurrentContext()->klGetDepthBuffer();
}


KLIMT_API void klTweak(GLenum nSelect, GLboolean nEnable)
{
	getCurrentContext()->klTweak(nSelect, nEnable);
}

KLIMT_API void klPrepareVertexArrays(int nNumVertices, int nNumNormals, int nNumColors, int nNumTexCoords,
									 void* nVertices, void* nColors, void* nTexCoords)
{
	getCurrentContext()->klPrepareVertexArrays(nNumVertices, nNumNormals, nNumColors, nNumTexCoords,
										  nVertices, nColors, nTexCoords);
}


KLIMT_API void klRenderPreparedIndexedArray(GLenum nPrimType, GLsizei nCount,
										    const unsigned int* nVertexIndices, const unsigned int* nColorIndices, const unsigned int* nTexCoordsIndices,
											const void* nVertices, const void* nColors, const void* nTexCoords)
{
	getCurrentContext()->klRenderPreparedIndexedArray(nPrimType, nCount, nVertexIndices, nColorIndices, nTexCoordsIndices,
												 nVertices, nColors, nTexCoords);
}


KLIMT_API void klCreateMipMaps()
{
	getCurrentContext()->klCreateMipMaps();
}


KLIMT_API void klSetPolygonOffset(GLfloat nOffset)
{
	getCurrentContext()->klSetPolygonOffset(nOffset);
}


KLIMT_API void klSetPortraitMode(GLboolean nEnable)
{
	getCurrentContext()->klSetPortraitMode(nEnable);
}


KLIMT_API void klProfilerReset()
{
	getCurrentContext()->klProfilerReset();
}


KLIMT_API void klProfilerBeginFrame()
{
	getCurrentContext()->klProfilerBeginFrame();
}


KLIMT_API void klProfilerEndFrame()
{
	getCurrentContext()->klProfilerEndFrame();
}


KLIMT_API GLfloat klProfilerGetRatio(GLenum nNom, GLenum nDenom)
{
	return getCurrentContext()->klProfilerGetRatio(nNom, nDenom);
}

KLIMT_API GLfloat klProfilerGetTime(GLenum nMesurement)
{
	return getCurrentContext()->klProfilerGetTime(nMesurement);
}



/////////////////////////////////////////////////////////////////////////////
//
//       the following API functions are currently not implemented
//
//    The empty wrapper functions allow to link against them, although they
//    won't do anything...
//



/// glTexParameteri is silently ignored
KLIMT_API void glTexParameteri(GLenum /*nTarget*/, GLenum /*nPName*/, GLint /*nParam*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glTexParameteri(): this method call is ignored");
}


/// glTexParameterf is silently ignored
KLIMT_API void glTexParameterf(GLenum /*nTarget*/, GLenum /*nPName*/, GLfloat /*nParam*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glTexParameterf(): this method call is ignored");
}


/// glTexEnvfv is silently ignored
KLIMT_API void glTexEnvfv(GLenum /*target*/, GLenum /*pname*/, const GLfloat /* *params*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glTexEnvfv(): this method call is ignored");
}


/// glTexGenf is silently ignored
KLIMT_API void glTexGenf(GLenum /*coord*/, GLenum /*pname*/, GLfloat /*param*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glTexGenf(): this method call is ignored");
}


/// glTexGeni is silently ignored
KLIMT_API void glTexGeni(GLenum /*coord*/, GLenum /*pname*/, GLint /*param*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glTexGeni(): this method call is ignored");
}


/// glTexGenfv is silently ignored
KLIMT_API void glTexGenfv(GLenum /*coord*/, GLenum /*pname*/, const GLfloat /* *params*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glTexGenfv(): this method call is ignored");
}


/// glAccum is silently ignored
KLIMT_API void glAccum(GLenum /*nOp*/, GLfloat /*nValue*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glAccum(): this method call is ignored");
}


/// glLightf is silently ignored
KLIMT_API void glLightf(GLenum /*nLight*/, GLenum /*nParamName*/, GLfloat /*nParam*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED2, "glLightf(): this method call is ignored");
}


/// glLightx is silently ignored
KLIMT_API void glLightx(GLenum /*nLight*/, GLenum /*nParamName*/, GLfixed /*nParam*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED2, "glLightx(): this method call is ignored");
}


/// glPointSize is silently ignored
KLIMT_API void glPointSize(GLfloat /*nSize*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glPointSize(): this method call is ignored");
}


/// glLineWidth is silently ignored
KLIMT_API void glLineWidth(GLfloat /*nWidth*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glLineWidth(): this method call is ignored");
}


/// glLineStipple is silently ignored
KLIMT_API void glLineStipple(GLint /*factor*/, GLushort /*pattern*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glLineStipple(): this method call is ignored");
}


/// glPolygonStipple is silently ignored
KLIMT_API void glPolygonStipple(const GLubyte /* *mask*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED, "glPolygonStipple(): this method call is ignored");
}


/// glPixelStorei is silently ignored
KLIMT_API void glPixelStorei(GLenum /*pname*/, GLint /*param*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glPixelStorei(): this method call is not implemented");
}


/// glBitmap is silently ignored
KLIMT_API void glBitmap(GLsizei /*width*/, GLsizei /*height*/, GLfloat /*xorig*/, GLfloat /*yorig*/, GLfloat /*xmove*/, GLfloat /*ymove*/, const GLubyte /* *bitmap*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glBitmap(): this method call is not implemented");
}


/// glClearIndex is silently ignored
KLIMT_API void glClearIndex(GLfloat /*c*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glClearIndex(): this method call is not implemented");
}


/// glIndexi is silently ignored
KLIMT_API void glIndexi(GLint /*c*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IMPORTANT, "glIndexi(): this method call is not implemented");
}


/// glViewport is silently ignored
KLIMT_API void glViewport(GLint /*x*/, GLint /*y*/, GLsizei /*width*/, GLsizei /*height*/)
{
	//report(reportMode, reportSelector, KL_REPORT_IGNORED2, "glViewport(): this method call is not implemented");
}


/// glClipPlane is silently ignored
KLIMT_API void glClipPlane(GLenum /*plane*/, const GLdouble /* *equation*/)
{
	//report(reportMode, reportSelector, KL_REPORT_UNIMPORTANT, "glClipPlane(): this method call is not implemented");
}


/// glAlphaFunc is silently ignored
KLIMT_API void glAlphaFunc(GLenum /*func*/, GLclampf /*ref*/)
{
	//report(reportMode, reportSelector, KL_REPORT_UNIMPORTANT, "glAlphaFunc(): this method call is not implemented");
}


/// glEvalMesh1 is silently ignored
KLIMT_API void glEvalMesh1(GLenum /*mode*/, GLint /*i1*/, GLint /*i2*/)
{

}


/// glEvalMesh2 is silently ignored
KLIMT_API void glEvalMesh2(GLenum /*mode*/, GLint /*i1*/, GLint /*i2*/, GLint /*j1*/, GLint /*j2*/)
{

}


/// glMapGrid1d is silently ignored
KLIMT_API void glMapGrid1d(GLint /*un*/, GLdouble /*u1*/, GLdouble /*u2*/)
{

}


/// glMapGrid1f is silently ignored
KLIMT_API void glMapGrid1f(GLint /*un*/, GLfloat /*u1*/, GLfloat /*u2*/)
{

}


/// glMapGrid2d is silently ignored
KLIMT_API void glMapGrid2d(GLint /*un*/, GLdouble /*u1*/, GLdouble /*u2*/, GLint /*vn*/, GLdouble /*v1*/, GLdouble /*v2*/)
{

}


/// glMapGrid2f is silently ignored
KLIMT_API void glMapGrid2f(GLint /*un*/, GLfloat /*u1*/, GLfloat /*u2*/, GLint /*vn*/, GLfloat /*v1*/, GLfloat /*v2*/)
{

}


/// glMap1d is silently ignored
KLIMT_API void glMap1d(GLenum /*target*/, GLdouble /*u1*/, GLdouble /*u2*/, GLint /*stride*/, GLint /*order*/, const GLdouble * /*points*/)
{

}


/// glMap1f is silently ignored
KLIMT_API void glMap1f(GLenum /*target*/, GLfloat /*u1*/, GLfloat /*u2*/, GLint /*stride*/, GLint /*order*/, const GLfloat * /*points*/)
{

}


/// glMap2d is silently ignored
KLIMT_API void glMap2d(GLenum /*target*/, GLdouble /*u1*/, GLdouble /*u2*/, GLint /*ustride*/, GLint /*uorder*/, GLdouble /*v1*/,
					   GLdouble /*v2*/, GLint /*vstride*/, GLint /*vorder*/, const GLdouble * /*points*/)
{

}


/// glMap2f is silently ignored
KLIMT_API void glMap2f(GLenum /*target*/, GLfloat /*u1*/, GLfloat /*u2*/, GLint /*ustride*/, GLint /*uorder*/, GLfloat /*v1*/,
					   GLfloat /*v2*/, GLint /*vstride*/, GLint /*vorder*/, const GLfloat * /*points*/)
{

}


#ifdef __cplusplus
}
#endif
