
// Example ported by maulo/Necrostudios
// for the Klimt for GP32 project

// GP32 specific includes
#include "gp32.h"
//#include "cpp_prototypes.h"

// As you can see, there is nothing standard here...
//#include <stdlib.h>
//#include <stdio.h>
//#include <string.h>

// We include our own headers
#include "gl/gl.h"
#include "gl/glu.h"

// Nope, there is no glut either
//#include "glut.h"

float f1 = 45.0f;
float f2 = 65.0f;

// These are the pointers to our screen framebuffers. We have two of them
// because of the double-buffer thing
u16 *screen;
u16 *screen2;

// "Contador": Thing that counts
int contador = 0;
char conta[3];

// Color correction flag
int color_correction = 1;

// For the FPS count
long r;
int fps;
char text_fps[3];

void idle();

// I know, I know. This is a dirty hack from hell, but I'll be using this
// until I find a cleaner solution or Klimt introduces a RGB555 raster.
// If someone has a better or faster idea, plese tell me. I'd be very
// happy to solve this issue in a stylish way :-D
// The problem is: By now, Klimt works only with RGB565 color type and
// our beloved GP supports only RGB555 color type. So, we must convert our
// framebuffer to RGB555 format just before showing it on the screen. If
// we don't, funny things will happen with the colors, just try it :-)
void torgb555(u16 s[]) {
    int i;
    unsigned short l,r;

    for ( i = 0 ; i < 320*240 ; i++) {
	l = s[i] & 0xFFC0;  // Left
	r = s[i] & 0x001F;  // Right, losing MSB

	r = r << 1;

	s[i] = l | r;
    }
}

// Function to toggle the color correction flag
void toggleColorCorrection() {
    
    if (color_correction == 0) color_correction = 1;
    else color_correction = 0;
}

int main(int argc, char *argv[])
{
    //int done;
    
    screen = (u16*) FRAMEBUFFER1;
    screen2 = (u16*) FRAMEBUFFER2;
    
    gp_setCpuspeed(133);  // Faster is unstable in some GPs
    gp_initFramebuffer(screen2,16,85);
    gp_initFramebuffer(screen,16,85);

    klCreateDefaultContext();

    // Here we have another curiosity. We must (by now) initialize the
    // screen in portrait mode. I haven't already been able to make
    // it work in natural landscape mode.
    klSetBufferSize(240, 320);

    // For double-buffering issues, we choose "screen2" as our initial
    // buffer to draw ...
    klSetColorBuffer(screen2);
   
    // ... but we set "screen" as the initial framebuffer
    gp_setFramebuffer(screen,1);

    // We set up our Real Time Clock
    gp_initRTC();
    
    // General GL setup
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f,(float)240/(float)320,0.1f,100.0f);

    glMatrixMode(GL_MODELVIEW);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    glClearDepth(1.0);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
    
    //done = 0;

    r = 0;
    fps = 0;
    
    // The main loop. It basically calls idle(), converts the buffer to
    // RGB555 (if desired) and swaps buffers.
    while(1){
	    idle();

	    // The FPS stuff
	    r = gp_getRTC();

	    // If less than a second has passed, increase the number of frames
	    if ( r < 64 ) {
		fps++;
	    } else {
	    // We reset the count
		gp_clearRTC();
		sprintf(text_fps, "%d", fps);
		fps = 0;
	    }
	    
	    
	    if ( contador == 0 ) {
		contador = 1;
		if (color_correction == 1) {
			torgb555(screen2);
			gp_drawString(300,20,2,"CC",0xFFC0,screen2);
		}
		gp_drawString(300,10,3,text_fps,0xFFC0,screen2);
		gp_setFramebuffer(screen2,1);
		klSetColorBuffer(screen);
	    } else {
		contador = 0;
		if (color_correction == 1) {
			torgb555(screen);
			gp_drawString(300,20,2,"CC",0xFFC0,screen);
		}
		gp_drawString(300,10,3,text_fps,0xFFC0,screen);
		gp_setFramebuffer(screen,1);
		klSetColorBuffer(screen2); 
	    }
	    
	    if ( gp_getButton()&BUTTON_A ) gp_Reset();
	    if ( gp_getButton()&BUTTON_B ) toggleColorCorrection();
	}
	
    // In order to be polite :-)
    return 0;

}

// This is the main drawing function called in every iteration.
// It draws the cube in its correct position
void idle()
{

	float l0Amb[] = { 0.8f, 0.8f, 0.8f, 1.0f };
	float lPos[4] = { 0.0f, 0.0f, 1.0f, 0.0f };
	
	//glClearColor(0.0f, 0.5f, 1.0f,0.0f);
	glClearColor(0.0f, 0.0f, 0.0f,0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  // Clear The Screen And The Depth Buffer
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glLightfv(GL_LIGHT0, GL_POSITION, lPos);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, l0Amb);
	
	glLoadIdentity();  // Reset The View
	glTranslatef(0.0f,-0.0f,-6.0f);

	glRotatef(f1,1.0f,0.0f,0.0f);
	glRotatef(f2,0.0f,1.0f,0.0f);

	glDepthFunc(GL_LESS);
	glDisable(GL_DEPTH_TEST);  // Disable z-test
	//glDepthMask(GL_FALSE);
	glDepthMask(0);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glCullFace(GL_BACK);
	glShadeModel(GL_FLAT);
	glEnable(GL_CULL_FACE);

	glColor3f(1.0f, 1.0f, 1.0f);
	//glutSolidCube(2.0);
	klutSolidCube(1.0);

	// You can make the cube seem to rotate faster or slower 
	// by changing these values
	f1+=0.9f;
	f2+=1.5f;
	
}

